/*
	File:		GraphicsHALTemplateEZ.c

	Contains:	This file contains the items needed to implement the Graphics HAL
				for an arbitrary TemplateEZ.
				
	The 'TemplateEZ' graphic hardware is characterized as follows:
	
		Cosmo - Graphics Controller
		This is the arbitrary name of the graphics controller ASIC.
		It arbitrarily has an extremely simple programing model.  It can support the following:

			 640x480 at 60 Hz (Common variant of VGA timing)
			 640x480 at 67 Hz (Standard Macintosh 'high resolution' display)
			 640x870 at 75 Hz (Portrait Display)
			 832x624 at 75 Hz (16" RGB Display)
			
		All displays can be supported at 8, 16, or 32 bits per pixel.
		There is no support for hardware cursor, or putting the ASICs into a low power state.
					
		Irazu - CLUT/DAC
		This is the arbitrary name of the CLUT/DAC, hereafter merely referred to as a CLUT.
		It is a simple 'triple 8x256' CLUT.
		
		
		The conversion of this TemplateEZ to a HAL for a specific hardware implementation can be
		started by doing a CASE SENSITIVE replacement of:
		
			"TemplateEZ" with the "NameOfYourGraphicsArchitectureGoesHere"
			"templateEZ" with the "nameOfYourGraphicsArchitectureGoesHere"
			
			"Cosmo" with the "NameOfYourGraphicsControllerGoesHere"
			"cosmo" with the "nameOfYourGraphicsControllerGoesHere"
			
			"Irazu" with the "NameOfYourCLUTGoesHere"
			"irazu" with the "nameOfYourCLUTGoesHere"


	Written by:	Sean Williams

	Copyright:	 1994-1995 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <2>	 7/17/95	SW		In GraphicsHAL GetModeTiming, now setting 'timingForamt =
		 							kDeclROMtables' prior to any error checking.  Also, in
									GraphicsHAL SetSync(), 'sitBitValidField' is now input only.
									Converted over to of xxIntxx instead of long/short/etc.
									Corrected a typo in GraphicsHAL GetVideoParams where a 
									'<< 1' was typed instead of a '<< 2'
		 <1>	 4/15/95	SW		First Checked In
			
*/

#include "GraphicsPriv.h"
#include "GraphicsHAL.h"
#include "GraphicsCoreUtils.h"		// for GraphicsUtilMapSenseCodesToDisplayCode()
#include "GraphicsOSS.h"

#include <PCI.h>
#include <Devices.h>
#include <DriverServices.h>
#include <Errors.h>
#include <Kernel.h>					// for DelayForHardware()
#include <Types.h>
#include <Video.h>



//
// CosmoRegisters
//	This structure models the graphic controller in the TemplateEZ graphics architecture.
//	It is a simple model:
//
//		graphicsMode		This register is programmed with arbitrary values to drive one
//							one of the supported rasters.
//							 640x480 at 60 Hz (Common variant of VGA timing)
//							 640x480 at 67 Hz (Standard Macintosh 'high resolution' display)
//							 640x870 at 75 Hz (Portrait Display)
//							 832x624 at 75 Hz (16" RGB Display)
//
//		bitDepth			This register is programmed with arbitrary values to configure
//							the controller for the bit depth.
//
//		monitorSense		This register is programmed with arbitrary values to drive/read the
//							state of the sense lines.
//
//		interruptControl	This register is programmed with arbitray values to enable/disable
//							and clear VBL interrupts.
//
typedef struct CosmoRegisters CosmoRegisters;
struct CosmoRegisters
{
	HWRegister8Bit *graphicsMode;
	HWRegister8Bit *bitDepth;
	HWRegister8Bit *monitorSense;
	HWRegister8Bit *interruptControl;
};



//
// IrazuRegisters
//	Irazu is a simple 'triple 8x256' CLUT, with an auto-incrementing address register.
//	It is a simple model:
//
//		address		This register has control which address of the CLUT is accessed.
//						After 3 consecutive access to the 'data' register, this auto-increments.
//
//		data			This register access the red, green, and blue channels of the CLUT via
//						consecutive reads/writes.  Following the access to the blue channel,
//						the address register auto-increments.
//
//		control		This register is programmed with arbitrary values to configure the
//						CLUT for the appropriate bit depth.
//
typedef struct IrazuRegisters IrazuRegisters;
struct IrazuRegisters
{
	HWRegister8Bit *address;
	HWRegister8Bit *data;
	HWRegister8Bit *control;
};



//
// TemplateEZHALData
//	This structure contains the necessary items for the TemplateEZ HAL to maintain its
//	state information.
//	
//	This is COMPLETELY private to the TemplateEZ implementation of the Graphics HAL.
//
typedef struct TemplateEZHALData TemplateEZHALData;
struct TemplateEZHALData
{
	RegEntryID regEntryID;		// The RegEntryID for this device
	void *hardwareBaseAddress;	// Base address for the hardware registers
	void *vramBaseAddress;		// Base address for the frame buffer VRAM
	AbsoluteTime ns260;			// 260 ns is the absolute time between CLUT accesses
	DisplayCode displayCode;	// Class of the connected display
	UInt8 graphicsMode;			// Value of Cosmo's 'graphicsMode' register for current DisplayModeID
	Boolean syncsDisabled;		// 'true' when V,H, and C syncs are disabled
	CosmoRegisters cosmo;		// Structure of pointers to various Cosmo registers
	IrazuRegisters irazu;		// Structure of pointers to various Irazu registers
};




//
// HALReplacementDriverInfo
//	In the event that the driver is being superseded, the HAL will try and save this information so
//	that the driver's HAL which replaces it can attempt to come up in the same state, if possible.
//	This will help to minimize the visual artifacts that a users sees.
//
//	However, TemplateEZ's behavior is so straight forward, that very little state information needs
//	to be maintained.
//
typedef struct HALReplacementDriverInfo HALReplacementDriverInfo;
struct HALReplacementDriverInfo
{
	UInt8 graphicsMode;			// Value of Cosmo's 'graphicsMode' register for current DisplayModeID
};

TemplateEZHALData *GraphicsHALGetHALData(void);

// Naming conventions for functions:
//		 GraphicsHALxxx	- functions which all HALs must implement.  These have external scope.
//		 {TemplateEZ}xxx	- functions which are strictly private to a specific HAL.


// TemplateEZ specific forward declarations.

static GDXErr TemplateEZMapDepthModeToCLUTAttributes(DepthMode depthMode, UInt32 *startAddress, UInt32 *entryOffset);

static void 	TemplateEZHandleVBLInterrupts(void *vblRefCon);
static 			ExtendedSenseCode TemplateEZGetExtendedSenseCode(void);
static void 	TemplateEZResetSenseLines(void);
static void 	TemplateEZDriveSenseLines(SenseLine senseLine);
static UInt32 	TemplateEZReadSenseLines(void);
static void 	TemplateEZWaitForVBL();
static void 	TemplateEZDelayForDuration(Duration duration);


TemplateEZHALData	gTemplateEZHALData;		// Persistant TemplateEZ specific data storage


// The version of TemplateEZ Graphics driver
enum
{
	kMajorRev = 1, 
	kMinorAndBugRev = 0,
	kStage = 0x20,			// 0x80 = Final, 0x60 = Beta, 0x40 = Alpha, 0x20 = Development
	kNonRelRev = 0

};



//
// DriverDescription
// 	This structure describes the native driver.  Since this is a Graphics driver, the driver should
//	be loaded by the DisplayMgr's expert loader.  This is indicated by the 'driverOSRuntimeInfo'
//	with the 'kDriverIsUnderExpertControl' flag set.
DriverDescription TheDriverDescription =
{
	// driverDescSignature
		kTheDescriptionSignature,						// Signature of DriverDescription

	// driverDescVersion
		kInitialDriverDescriptor,						// Version of this data structure

	// driverType
	{
		"\ptemplateEZ",									// device name must match in Devicetree
		kMajorRev, kMinorAndBugRev, kStage, kNonRelRev,	// Major, Minor, Stage, Rev
	},

	
	// driverOSRuntimeInfo								// OS Runtime Requirements of Driver
		{
			kDriverIsOpenedUponLoad + kDriverIsUnderExpertControl,	// Runtime Options
			"\p.Display_Video_Apple_TemplateEZ",
		},

		
	// driverServices									// Apple Service API Membership

		// nServices
		1,												// Number of Services Supported						

		// serviceCategory								// Service Category Name
		kServiceCategoryNdrvDriver,						// We support the 'ndrv' category

		// serviceType									// Type within Category
		kNdrvTypeIsVideo,								// Type1

		// serviceVersion								// Version of service 
		1,0,0,0											// majorRev, minorAndBugRev, stage, nonRelRev

};



//=====================================================================================================
//
// GraphicsHALGetHALData()
//	Return the pointer to the global Core data.  Yes...you guessed it...it just returning
//	the address of the global, but use the accessor function anyway to isolate yourself
//	in the event CFM coolness is ever lost.
//
//=====================================================================================================
TemplateEZHALData  *GraphicsHALGetHALData(void)
{
	return (&gTemplateEZHALData);
}



//=====================================================================================================
//
// GraphicsHALInitPrivateData()
//	Allocate and initialize the HAL private data
//
//		-> regEntryID	The NameRegistry ID for this device.
//
//		<> replacingDriver
//		On input, this indicates whether the Core got a 'kInitializeCommand' or a 'kReplaceCommand'.
//		These commands are similar, but with subtle differences.  A 'kInitializeCommand' is issued
//		if no version of this driver has been previously loaded, whereas a 'kReplaceCommand' is
//		issued if a previous version of the driver has been loaded, but subsequently superseded.
//
//		If 'false', then a 'kInitializeCommand' had been recieved by the Core, and the HAL should
//		do a full hardware initialization.
//
//		If 'true', then a 'kReplaceCommand' had been received by the Core, and the HAL can attempt
//		to configure itself to its state prior to it being superceded.
//
//		On output, this allows the HAL to override the Core's default behavior if it chooses to do so.
//
//		If 'false' then the HAL is signinaling the Core that it is unable to re-configure itself to
//		its state prior to being superseded, and the core will continue as if a 'kInitializeCommand'
//		had occurred.
//		
//		If 'true', then the HAL was able to re-configure itself in the event of being replaced,
//		and the Core will proceeded accordingly.
//
//=====================================================================================================
GDXErr GraphicsHALInitPrivateData(const RegEntryID *regEntryID, Boolean *replacingDriver)
{
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	GDXErr err = kGDXErrUnableToAllocateHALData;	// Assume failure

	PCIAssignedAddress assignedAddresses[2];		// There should be two "phys-addr size" pairs 
	UInt32 applAddress[2];							// There should be two logical addresses
	UInt32 baseRegister0Index;						// Entry in APPL,address for Base Register 0
	UInt32 baseRegister1Index;						// Entry in APPL,address for Base Register 1
	UInt16 commandRegister;							// Enable TemplateEZ's memory if everything ok
	OSErr osErr;

	Nanoseconds nanoseconds;						// Convert 260ns into Absolute Time
	void *hardwareBaseAddress;
	UInt32 halPreferences;							// Used for retrieving HAL prefs from NVRAM.
	
	// Save the regEntryID,
	RegistryEntryIDCopy(regEntryID, &templateEZHALData->regEntryID);


	// TemplateEZ is on a PCI card.  TemplateEZ contains the Cosmo graphics controller, and the 
	// Irazu CLUT.  In this sample code, it is assumed that the emplateEZ PCI card uses 2 
	// Base Registers.  This illustrates how the "assigned-address" property must be examined to
	// determine which "AAPL,address" 32 bit entry corresponds with each Base Register.
	//
	// 		Base Register 0 (offset 0x10 from the configuration registers) is the base address for
	// 		the hardware registers.
	//	
	// 		Base Register 1 (offset 0x14 from the configuration registers) is the base address for
	//		the VRAM. 
	//
	// TemplateEZ's Open Firmware startup code created a "reg" property that described
	// 		1) the configuration registers
	// 		2) Base Register 0
	// 		3) Base Register 1
	// Open Firmware creates an "assigned-address" property that describes Base Register 0 and
	// Base Register 1


	// Get the kPCIAssignedAddressProperty.  (Note that a multiplier of 2x is used since since
	// there should be 2 assigned-addresses.
	
	err = GraphicsOSSGetProperty(&templateEZHALData->regEntryID, kPCIAssignedAddressProperty,
			assignedAddresses, sizeof(PCIAssignedAddress)*2);

	// If there is an error, that means Open Firmware was unable to allocate the memory space that was
	// requested.  Hence, TemplateEZ is unusable at this time and the driver should quit.
	if (err)
		goto ErrorExit;
		
	// Since the entries of the "assigned-address" property does not necessarily match the order of the
	// entries in the "reg" property, examine the rrrrrrrr bits in each physHi to determine which 
	// "phys-addr size" pair matches each Base Register.
	// (Base Register 0 (offset at 0x10 from the configuration registers) has rrrrrrrr = 0x10 etc...)
	
	if ( 0x10 == assignedAddresses[0].registerNumber )
	{
		baseRegister0Index = 0;
		baseRegister1Index = 1;
	}
	else
	{
		baseRegister0Index = 1;
		baseRegister1Index = 0;
	}
		
	// The order of the "phys-addr size" pairs in the "assigned-address" property match the order
	// of the logical addresses in the "AAPL,address" property.  Go find the logical addresses.
	
	err = GraphicsOSSGetProperty(&templateEZHALData->regEntryID, "AAPL,address", applAddress, 8);
	
	if (err)							// Should NEVER be an error if gotten this far
		goto ErrorExit;					
		
	// Have successfully found all base addresses.
	// Enable TemplateEZ's memory space.  Always do a read-modify-write when hitting configuration
	// registers.
	
	osErr = ExpMgrConfigReadWord(&templateEZHALData->regEntryID, (LogicalAddress) 0x4, &commandRegister);

	if (osErr)							// Shouldn't be an error
		goto ErrorExit;					

	commandRegister|= 1 << 1;			// Enable the Memory space which is Bit 1

	osErr = ExpMgrConfigWriteWord(&templateEZHALData->regEntryID, (LogicalAddress) 0x4, commandRegister);
	if (err)							// Shouldn't be an error
		goto ErrorExit;					

	
	// Note: More examples
	// 1) A PCI card only uses 1 BaseRegister to access hardware and VRAM.  If the card's
	// Open Firmware startup code created a "reg" property for just the 1 Base Register that it uses,
	// the "AAPL,address" property will contain the single logical address.  So if the "AAPL,address"
	// property exists and is the right size (4), then the PCI card knows Open Firmware was able to
	// allocate the
	// requested memory.  Such a card would only have to do the following to find the logical address.
	//
	//	err = GraphicsOSSGetProperty(&templateEZHALData->regEntryID, "AAPL,address", &hwBaseAddress, 4);
	//	if (err)							// It is possible to error out if Open Firmware couldn't 
	//		goto ErrorExit;					// allocate the requested memory space.
	//
	// 2) A PCI card that doesn't have Open Firmware startup code (BECAUSE THE CARD IS IN DEVELOPMENT)
	// will have a "reg" property automatically created for it by Open Firmware, as described in
	// chapter 4 of Designing PCI Cards & Drivers.
	// The PCI Graphics card should implement the Expansion ROM Base Register (offset 0x30 from the
	// configuration registers) because it will eventually be a startup device. Open Firmware will
	// automatically add a "phys-addr size" pair to the "assigned-address" property for the Expansion
	// ROM Base Register. Likewise, an additional entry will be created in the "AAPL,address" property.
	// The driver, during development, needs to allocate additional space in the buffers used to fetch
	// each property.  The entries of the "assigned-address" property need to be exaimined to determine the order
	// as shown above.
	// (All PCI graphics cards should be a startup device and need the Expansion ROM Base Register.
	// YOU ARE LAME IF YOU DON'T HAVE ONE)



	// Successfully obtained base address, so intilize the various pointers to the hardware
	// registers.

	templateEZHALData->hardwareBaseAddress = (void *) applAddress[baseRegister0Index];
	templateEZHALData->vramBaseAddress = (void *) applAddress[baseRegister1Index];

	hardwareBaseAddress = templateEZHALData->hardwareBaseAddress;
	
	templateEZHALData->cosmo.graphicsMode = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0x00);
	templateEZHALData->cosmo.bitDepth = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0x10);
	templateEZHALData->cosmo.monitorSense = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0x20);
	templateEZHALData->cosmo.interruptControl = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0x30);

	templateEZHALData->irazu.address = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0xF000);
	templateEZHALData->irazu.data = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0xF010);
	templateEZHALData->irazu.control = (HWRegister8Bit *) ((UInt32) hardwareBaseAddress + 0xF020);

	// Irazu must wait at least 4 pixel clocks following writing to the blue channel in order to
	// allow the data to be latched correctly.  Worst case (for the VGA display) this
	// corresponds to roughly 260 ns. Since this is delay is done many times, 
	// convert 260 ns to absolute time and save in the HAL data.

	nanoseconds.hi = 0;
	nanoseconds.lo = 260;
	templateEZHALData->ns260 = NanosecondsToAbsolute(nanoseconds);


	if (*replacingDriver)
	{
		// Important Implementation Note:  The exact behavior of what should happen during driver
		// replacement is largely dependent of the implementation of the driver, and the reason for
		// it being replaced.
		// Under the ideal circumstances, a driver should be able to be replaced without having to
		// shut down the raster or reprogram the hardware.  This will prevent visible flashes on the
		// display during the replacement process.
		// 
		// When a driver is being updated to incorporate new features (as opposed to bug fixes), then
		// replacement can usually occur without flashes.  This is because by default, the
		// Driver Loader Library (DLL) does not let QuickDraw know that a new driver got loaded, so
		// QuickDraw doesn't issue commands to gray the screen, etc. which result in 'flashes.'
		//
		// However, this default behavior is only acceptable if the new driver comes up in the same
		// bit depth and has the same 'base address' and 'rowbytes' as the driver it replaced.  If 
		// this is not the case, then the DLL needs to be informed that a full initialization of all
		// the QuickDraw variables is required.  The DLL can be informed by setting uncommenting
		// the following lines:
		// 
		// UInt32 needFullInit  = 1;
		// err = GraphicsOSSSaveProperty(&sixty6HALData->regEntryID, "needFullInit",
		//			&needFullInit, sizeof(needFullInit), kOSSPropertyVolatile);



		// This driver is replacing a previous version.  Therefore, to try and minimize the visual
		// effects of this driver being loaded, retrieve whatever state information the superseded
		// driver left behind.
		//
		// Note:  For TemplateEZ, this is largely just a place holder.  Your real hardware probably
		// has more extensive requirements.
		
		HALReplacementDriverInfo replacementDriverInfo;
		
		err = GraphicsOSSGetProperty(&templateEZHALData->regEntryID, "HALReplacementInfo",
				&replacementDriverInfo, sizeof(HALReplacementDriverInfo));

		if (!err)
		{
			templateEZHALData->graphicsMode = replacementDriverInfo.graphicsMode;
			
			// Have now grabbed all the old state information.  The new driver knows what bugs the old
			// driver had.  Examine the state information to see if it is necessary to do a full hw 
			// initialization.  For example, one displayModeID was known to be bad...the hw was
			// programmed incorrectly for that resolution.
			// if (brokenDisplayModeID == templateEZHALData->displayModeID)
			//		*replacingDriver = false;
			
		}
		else
		{
			// An error occurred while attempting to retrieve the previous state information, so
			// report that a full initialization is required.
			
			*replacingDriver = false;
			err = kGDXErrNoError;
		}
	}

	// Always try to delete the HALReplacementDriverInfo in the NameRegistry, so it doesn't cause
	// confustion later on.
	(void) GraphicsOSSDeleteProperty(&templateEZHALData->regEntryID, "HALReplacementInfo");


	// TemplateEZ might want to look at the HAL specfic data stored in NVRAM.
	// The last 4 bytes (each device is allowed 8 bytes of NVRAM) is private to each HAL.
	// GraphicsOSSGetHALPref() automatically fetches the last 4 bytes for the HAL.
	
	err = GraphicsOSSGetHALPref(&templateEZHALData->regEntryID, &halPreferences);
	if (!err)
	{
		// Examine the halPreferences stored in NVRAM.  The HAL portion of the NVRAM is private to the
		// HAL (the Core never examines or messes with it) and can have any meaning defined by the HAL.
		// If the HAL wishes to save data, it can use the routine GraphicsOSSSetHALPref().
		// NOTE: the HAL should do a read-modify-write if it wishes to change the data.  This allows
		// disk based drivers that replace the ROM based drivers to define new HAL data.  Ensures that
		// the ROM based driver (always run at startup before the disk based driver) doesn't smash
		// the data that the disk based driver saved.
	}

	err = kGDXErrNoError;								// Everything ok
	
	
ErrorExit:

	return (err);
	
}



//=====================================================================================================
//
// GraphicsHALKillPrivateData()
//	Dispose of the HAL's private data
//
//=====================================================================================================
void GraphicsHALKillPrivateData()
{
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();

	RegistryEntryIDDispose(&templateEZHALData->regEntryID);

}



//=====================================================================================================
//
// GraphicsHALOpen()
//	It is possible for the driver to be opened and closed many times.
//	The HAL private data has been setup, so just initialize the hardware into
//	the state that is should be in on startup.  This should be the complete initialization
//	to get the HW into the desired state for the amount of VRAM that is in the system and
//	any other hardware specfic stuff that the HAL cares about. The sense lines, for example, should
//	be able to be read and tweaked to determine the type of the connected monitor.
//	No programming to set up the raster for a givin 'DisplayModeID' or 'DepthMode' is necessary at
//	this point.
//
//		-> spaceID			The AddressSpaceID for this device.
//		-> replacingDriver	'true' if the HAL should behave as if the driver is being replaced,
//							'false' otherwise.
//
//=====================================================================================================
GDXErr GraphicsHALOpen(const AddressSpaceID spaceID, Boolean replacingDriver)
{
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo = &templateEZHALData->cosmo;
	GDXErr err = kGDXErrNoError;

	// No real hardware initialization needs to occur, since things should (according to the
	// arbitrary ERSs) come in in a polite state.  Additionally, things were left in a polite
	// state in the event of a previous GraphicsHALClose().
	
	templateEZHALData->syncsDisabled = false;	



ErrorExit:

	return(err);
}


//=====================================================================================================
//
// GraphicsHALClose(const AddressSpaceID spaceID)
//	Upon close, there are no major requirements, since the majority of the work will be handled
//	elsewhere.
//	
//
//=====================================================================================================
GDXErr GraphicsHALClose(const AddressSpaceID spaceID)
{
	GDXErr err = kGDXErrNoError;
		
ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALTerminate()
//
//		-> superseded
//		'true' if current driver is going to be superseded by another driver, 'false' otherwise.
//		If 'true', the current driver can choose to save any state that the replacement driver
//		may need..if it wants to keep raster going.
//
//		'false' no driver is going to replace it.  In that event, it should stop the raster and
//		leave hardware in a 'polite' state.
//
//=====================================================================================================
GDXErr GraphicsHALTerminate(Boolean superseded)
{
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	GDXErr err = kGDXErrNoError;										// Assume success
	CosmoRegisters *cosmo = &templateEZHALData->cosmo;


	if (superseded)
	{
		// Driver is being superseded, so save state information for the replacement driver.
		
		HALReplacementDriverInfo replacementDriverInfo;

		replacementDriverInfo.graphicsMode = templateEZHALData->graphicsMode;

		err = GraphicsOSSSaveProperty(&templateEZHALData->regEntryID, "HALReplacementInfo",
				&replacementDriverInfo, sizeof(HALReplacementDriverInfo), kOSSPropertyVolatile);

		if (err)
			*cosmo->graphicsMode = 0xFF;	// Causes Comso's state machine to stop and blanks display
	}
	else
	{
		// Driver is going away for good, so blank turn off state machines.
		*cosmo->graphicsMode = 0xFF;		// Causes Comso's state machine to stop and blanks display
	}
	
	return err;
}



//=====================================================================================================
//
// GraphicsHALGetVBLInterruptRoutines()
//	The OSS encapsulates how interrupts are handled by the system.  This routine supplies
//	that OSS with the HAL's interrupt routines that follow the OSS conventions.  Hopefully,
//	if the OS changes, only the OSS will need to change.
//
//		<- installVBLInterrupts	
//		'true' 	if the HAL's interrupt scheme can match the OSS's scheme. i.e. the HAL lets the OSS
//		handle most of the interrupt functions.
//		'false'	if the HAL's interrupt scheme is radically different than the OSS's scheme.  The
//		HAL is responsible for knowing how the OS handles interrupts.  Obviously, this is the 
//		escape mechanism for a poor OSS design.  The HAL needs a radically different interrupt handling
//		design so the OSS will not handle any interrupt services.  If this is false, all other
//		paramters are ignored
//
//		<- chainDefault
//		If 'halVBLEnabler' or 'halVBLDisabler' = NULL, this is ignored by the OSS for the respective
//		function since the default enabler/disabler supplied by the OS is used.
//		If 'chainDefault = true', if the halVBLEnabler or halVBLDisabler != NULL, the OSS will call
//		the default OS enabler/disbler after the HAL's enabler/disabler is called
//		If 'chainDefault = false', if the halVBLEnabler or halVBLDisabler != NULL, the OSS will NOT call
//		the default OS enabler/disbler after the HAL's enabler/disabler is called.  The HAL assumes
//		the responsibility for enabling/disabling the interrupt source.  (Dangerous!)
//
//		<- halVBLHandler
//		The HAL's VBL handler which should clear the internal interrupt source.
//
//		<- halVBLEnabler
//		If 'halVBLEnabler = NULL', the default OS enabler will be called and the HAL can ignore things.
//		If 'halVBLEnabler != NULL' and 'hainDefault = true', the HAL needs to enable the internal
//		interrupt source and the OSS calls the default OS enable routine to enable external interrupts.
//		If 'halVBLEnabler != NULL' and 'chainDefault = false', the HAL needs to enable the internal
//		and external interrupt source.  (Dangerous!)
//
//		<- halVBLDisabler
//		If 'halVBLDisabler = NULL', the default OS enabler will be called and the HAL can ignore things.
//		If 'halVBLDisabler != NULL' and 'chainDefault = true', the HAL can choose to disable the internal
//		interrupt source and the OSS calls the default OS disable routine to disable external interrupts.
//		If 'halVBLDisabler != NULL' and 'chainDefault = false', the HAL can choose to disable the internal
//		and must disable the external interrupt source.  (Dangerous!)
//
//		<- vblRefCon
//		If the HAL needs some data for the interrupt routines, allocate some structure that
//		is pointed to by vblRefCon.
//
//
//	TemplateEZ always has the internal interrupt source enabled so it knows when vbls occur.  Hence,
//	halVBLEnabler and halVBLDisabler = NULL.  The default OS enable/disable routines are used to
//	control what interrupts the OS sees.
//
//=====================================================================================================
GDXErr GraphicsHALGetVBLInterruptRoutines(Boolean *installVBLInterrupts, Boolean *chainDefault,
		VBLHandler **halVBLHandler, VBLEnabler **halVBLEnabler, VBLDisabler **halVBLDisabler,
		void **vblRefCon)
{
	// TemplateEZ is always going to have its internal interrupt source enabled for VBLs.  Hence,
	// 'halVBLEnabler' and 'halVBLDisabler' are NULL.  The internal interrupt will be 
	// allowed/stopped to propagate by opening/closing the external interrupt 'gateway' via the
	// default OS enabler/disabler.
	
	*installVBLInterrupts = true;		// This HAL supports the OSS's interrupt scheme.
	*chainDefault = false;				// Ignored by OSS since HAL's enabler/disablers are NULL
	*halVBLHandler = TemplateEZHandleVBLInterrupts; 
	*halVBLEnabler = NULL;
	*halVBLDisabler = NULL; 
	*vblRefCon = NULL;					// No private refCon needed.

	return kGDXErrNoError;
}



//=====================================================================================================
//
// GraphicsHALGrayCLUT()
//	This routine sets all the CLUT entries to 50% gray (with gamma correction).
//	This is useful so that the pixel depth can be subsequently changed without 
//	introducing screen anonmalies.
//	The 50% gray value will be obtained by using the midpoint value of the supplied
//	gamma table.
//
//	NOTE:  this assumes that the gamma correction data size is 1 byte,
//	as stated in the core.
//
//=====================================================================================================
GDXErr GraphicsHALGrayCLUT(const GammaTbl *gamma)
{

	// Gray all 256 entries...it doesn't matter what the current 'depthMode' is.
	enum { kClutSize = 256 };

	GDXErr err = kGDXErrUnknownError;					// Assume failure

	UInt8 *midPointRed;									// Midpoint of red correction data
	UInt8 *midPointGreen;								//     "    "  green    "       "
	UInt8 *midPointBlue;								//     "    "  blue     "       "	
	SInt16 channelCount = gamma->gChanCnt;	
	SInt16 entriesPerChannel = gamma->gDataCnt;
	
	IrazuRegisters *irazu;
	Boolean vblInterruptsEnabled;

	UInt32 i;											// Loop iterator


	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	AbsoluteTime ns260 = templateEZHALData->ns260;			// 260ns in absolute time
	
	irazu = &templateEZHALData->irazu;
	
	// Get the midpoint of the red correction data.  This is found by starting at the begining of the
	// correction data, which can be found at '&gFormulaData[0]', adding the 'gFormulaSize'.
	// Then go halfway into the table as determined by 'entriesPerChannel / 2'
	
	midPointRed = (UInt8 *) ((UInt32) &gamma->gFormulaData[0] + gamma->gFormulaSize +
			(entriesPerChannel / 2));

	// If there is only 1 channel of correction data, it means the same correction is applied to the
	// red, green, and blue channels.  If there are 3 channels then each color has its own correction
	// data.
	
	if (1 == channelCount)
	{
		midPointGreen = midPointRed;
		midPointBlue = midPointRed;
	} else {
		midPointGreen = midPointRed + entriesPerChannel;
		midPointBlue = midPointRed + (entriesPerChannel * 2);
	}

	// Turn off VBL interrupts so the CLUT can be adjusted during VBL time, without having to worry
	// about the VBL interrupt handler stealing time.
		
	vblInterruptsEnabled = GraphicsOSSSetVBLInterrupt(false);						
	TemplateEZWaitForVBL();
	
	*irazu->address = 0;						// Start at CLUT entry 0
	SynchronizeIO();

	for (i = 0 ; i < kClutSize ; i++)
	{
	
		*irazu->data = *midPointRed;			// Red
		SynchronizeIO();
		
		*irazu->data = *midPointGreen;			// Green
		SynchronizeIO();
		
		*irazu->data = *midPointBlue;			// Blue
		SynchronizeIO();
		
		(void) DelayForHardware(ns260);			// Wait for data to latch (4 dot clocks)
	}
	
	if (vblInterruptsEnabled)
		(void) GraphicsOSSSetVBLInterrupt(true);
		
	err = kGDXErrNoError;						// Everything okay
	
ErrorExit:

	return (err);
	
}



//=====================================================================================================
//
// GraphicsHALSetCLUT()
//	This routine will program the CLUT with the specified array of 'ColorSpecs'.
//	Two such arrays are provided, the original, and a second that has been luminance mapped (if
//	appropriate) and gamma corrected. It is up to the HAL implementation to decide which array should
//	be applied to the hardware. Most hardware will use the corrected version.
//
//	It is important to note that the positions of the entries refers to logical positions, not physical
//	ones. In 4-bits-per-pixel mode, for example, the entry positions could range from 0, 1, 3,, 15,
//	even though the physical positions may not have this number sequence.
//
//	No range checking is required, because the caller has allready done so.	
//
//		-> originalCSTable		
//		This is a pointer to the array of ColorSpecs provided by the caller. This is only provided in
//		the event that the hardware should not use the correctedCSTable. If any adjuments need to made
//		to it, then they should be done to a copy. Dont throw away the const!
//
//		-> correctedCSTable
//		This is essentially a copy of originalCSTable, except that it has been luminance mapped (if
//		appropriate) and gamma corrected. Most hardware will use this information to set the CLUT. 
//		Though it is unlikely that you will need to change this information, it is not marked as
//		'const' in case you need to build a special version from the originalCSTable.  In that event,
//		you can alter the array as you see fit.  (Note:  regardless of the size of the originalCSTable,
//		correctedCSTable points to an array of ColorSpecs with 256 entries.
//
//		-> startPosition
//		(0 based) Starting point in the array of ColorSpecs
//		
//		-> numberOfEntries
//		(0 based) This is the number of entries to be set.
//
//		-> sequential
//		If 'false', then the 'value' field of the ColorSpec should be inspected to see what logical
//		position should be set. If 'true', then the array index indicates what logical position should
//		be set.
//
//		-> depthMode
//		The relative bit depth. This is provided so that the HAL can decide how to map the logical
//		entry positions to the physical entry postions.
//
//=====================================================================================================
GDXErr GraphicsHALSetCLUT(const ColorSpec *originalCSTable, ColorSpec *correctedCSTable,
		SInt16 startPosition, SInt16 numberOfEntries, Boolean sequential, DepthMode depthMode)
{
	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	AbsoluteTime ns260 = templateEZHALData->ns260;				// 260ns in absolute time
	IrazuRegisters *irazu;
	Boolean vblInterruptsEnabled;

	UInt32 startAddress;
	UInt32 entryOffset;
	UInt32 physicalAddress;
	UInt32 logicalAddress;
	
	UInt32 i;									// Loop control variable
	
	GDXErr err = kGDXErrUnknownError;			// Assume failure.

	irazu = &templateEZHALData->irazu;
	
	err = TemplateEZMapDepthModeToCLUTAttributes(depthMode, &startAddress, &entryOffset);
	if (err)
		goto ErrorExit;

	// Turn off VBL interrupts so the CLUT can be adjusted during VBL time, without having to worry
	// about the VBL interrupt handler stealing time.
		
	vblInterruptsEnabled = GraphicsOSSSetVBLInterrupt(false);						
	TemplateEZWaitForVBL();

	// Program the CLUT entries.  For our hardware, use the correctedCSTable
	for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
	{
		if (sequential)
			logicalAddress = i;
		else
			logicalAddress = correctedCSTable[i].value;

		physicalAddress = startAddress + (logicalAddress * entryOffset);
		
		*irazu->address = physicalAddress;
		SynchronizeIO();
		
		*irazu->data = correctedCSTable[i].rgb.red;			// Red
		SynchronizeIO();
		
		*irazu->data = correctedCSTable[i].rgb.green;		// Green
		SynchronizeIO();
		
		*irazu->data = correctedCSTable[i].rgb.blue;		// Blue
		SynchronizeIO();

		(void) DelayForHardware(ns260);						// Wait for data to latch (4 dot clocks)

	}
		
	if (vblInterruptsEnabled)
		(void) GraphicsOSSSetVBLInterrupt(true);
		
	err = kGDXErrNoError;									// Everything okay
	
ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetCLUT()
//	This routine will fill out the specified array of ColorSpecs with the contents of the CLUT.
//	The 'RGBColor' structure in each 'ColorSpec' uses 16-bits for each channel (red, green, and blue),
//	whereas most CLUTs only use 8-bits. Therefore, when filling in the 'RGBColor' structure, the most
//	significant byte for each channel should be filled with the 8-bits extracted from its respective
//	channel in the CLUT. Moreover, to maintain the same behavior as the reference model, the 8-bits
//	from the CLUT should also be written to the least significant byte for each RGBColor.
//	
//	It is important to note that the positions of the entries refer to logical positions, not physical
//	ones. At 4 bpp, for example, the entry positions could range from 0, 1, 2,, 15, even though the
//	physical positions may not have this number sequence.
//
//	No range checking is required, because the caller has already done so.
//
//		<-> csTable
//		Pointer to array of 'ColorSpecs' provided by the caller to be filled with the CLUT contents.
//
//		-> startPosition	(0 based) Starting point in the array to fill.
//		-> numberOfEntries	(0 based) The number of entries to be get.
//
//		-> sequential
//		If 'false', then the value field of the 'ColorSpec' should be inspected to see what logical
//		position should be retrieved. If 'true', then the array index indicates what logical position
//		should be read.
//
//		-> depthMode
//		The relative bit depth. This is provided so that the HAL can decide how to map the logical
//		entry positions to the physical entry postions.
//
//=====================================================================================================
GDXErr GraphicsHALGetCLUT(ColorSpec *csTable, SInt16 startPosition, SInt16 numberOfEntries,
		Boolean sequential, DepthMode depthMode)
{	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	AbsoluteTime ns260 = templateEZHALData->ns260;				// 260ns in absolute time
	IrazuRegisters *irazu;

	UInt32 startAddress;
	UInt32 entryOffset;
	UInt32 physicalAddress;
	UInt32 logicalAddress;
	
	UInt32 i;											// Loop control variable
	
	GDXErr err = kGDXErrUnknownError;					// Assume failure.

	irazu = &templateEZHALData->irazu;
	
	err = TemplateEZMapDepthModeToCLUTAttributes(depthMode, &startAddress, &entryOffset);
	if (err)
		goto ErrorExit;

	// Read the CLUT entries.
	for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
	{
		if (sequential)
			logicalAddress = i;
		else
			logicalAddress = csTable[i].value;

		physicalAddress = startAddress + (logicalAddress * entryOffset);
		
		*irazu->address = physicalAddress;
		SynchronizeIO();
	
		(void) DelayForHardware(ns260);							// Let data settle (4 dot clocks)
		
		csTable[i].rgb.red = *irazu->data;						// Get lower byte for Red		(xxrr)
		csTable[i].rgb.red |= (csTable[i].rgb.red << 8);		// Copy it to most sig. byte 	(rrrr)
		SynchronizeIO();
		
		csTable[i].rgb.green = *irazu->data;					// Get lower byte for Green		(xxgg)
		csTable[i].rgb.green |= (csTable[i].rgb.green << 8);	// Copy it to most sig. byte 	(gggg)
		SynchronizeIO();
		
		csTable[i].rgb.blue = *irazu->data;						// Get lower byte for Blue		(xxbb)
		csTable[i].rgb.blue |= (csTable[i].rgb.blue << 8);		// Copy it to most sig. byte 	(bbbb)
		SynchronizeIO();

	}
	
	err = kGDXErrNoError;										// Everything okay
	
ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetPages()
//	This routine reports the number of graphics pages supported for the specified 'DisplayModeID' at
//	the specified 'DepthMode'.
//	No attempt should be made to determine whether or not a display capable of being driven with
//	a raster of type 'DisplayModeID' is physically connected.
//
//		-> displaymodeID	The DisplayModeID for which the page count is desired.
//		-> depthMode		The relative bit depth for which the page count is desired.
//
//		<- pageCount
//		# of pages supported at the specified 'DisplayModeID' and 'DepthMode'.  In the event of an
//		error, 'pageCount' is undefined.  This is a counting number, so it is NOT zero based.
//
//=====================================================================================================
GDXErr GraphicsHALGetPages(DisplayModeID displayModeID, DepthMode depthMode, SInt16 *pageCount)
{
	
	GDXErr err = kGDXErrUnknownError;							// Assume failure.
	Boolean	modePossible = false;
	
	// For the TemplateEZ graphics hardware, the 'pageCount' is always 1 for any supported 
	// 'DisplayModeID'  Therefore, the validity of the  the specified 'DisplayModeID' and 'DepthMode'
	// will be tested by calling GraphicsHALModePossible(), passing in '0' for the page #.
	
	err = GraphicsHALModePossible(displayModeID, depthMode, 0, &modePossible);
	if (err || !modePossible)
	{
		// Opps...caller specified an invalid 'DisplayModeID' or 'DepthMode'
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	// TemplateEZ graphics architecture only supports 1 page.
	
	*pageCount = 1;
	
	err = kGDXErrNoError;									// Everything okay

ErrorExit:

	return (err);
}




//=====================================================================================================
//
// GraphicsHALGetBaseAddress()
//	This returns the base address of a specified page in the current mode. 
//	This allows video pages to be written to even when not displayed
//
//			->	page			Page number ( 0 based ).  Return the base address for this page.
//			<-	baseAddress		Base address of desired page.
//
//=====================================================================================================
GDXErr GraphicsHALGetBaseAddress(SInt16 page, char **baseAddress)
{


	GDXErr err = kGDXErrUnknownError;			// Assume failure.
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	
	// TemplateEZ only supports 1 page.  Hence, only page = 0 is valid.
	// All other values will return an error.

	if ( 0 != page )
	{
		// Caller asked for baseAddress of invalid page
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	*baseAddress = (char *) (templateEZHALData->vramBaseAddress);
		
	err = kGDXErrNoError;									// Everything okay
		

ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetSync()
//	Multipurpose call to 
//	1) Report the capabilities of the frame buffer in controlling the sync lines and if HW can
//	"sync" on Red, Green or Blue
//	2) Report the current status of the sync lines and if HW is "syncing" on Red, Green or Blue
//	
//	If the display supported the VESA Device Power Management Standard (DPMS), it would respond
//	to HSync and VSync in the following manner:
//	The VESA Standards is:
//	
//			State       	Vert Sync		Hor Sync		Video
//			-----			--------		---------		------
//			Active         	Pulses      	Pulses      	Active
//			Standby			Pulses   		No Pulses      	Blanked
//			Idle			No Pulses      	Pulses			Blanked
//			Off			 	No Pulses		No Pulses		Blanked
//
//
//			->	getHardwareSyncCapability	'true' if reporting HW capability
//											'false' if reporting current status
//
//	For this routine, the relevant fields of the 'VDSyncInfoRec' structure are as follows:
//			<-	sync		Report HW capability or current state.
//
//				If 'getHardwareSyncCapability = true' then report the cabability of the HW
//				When reporting the capability of the HW, set the appropriate bits of csMode:
//				kDisableHorizontalSyncBit		Set if HW can disable Horizontal Sync
//				kDisableVerticalSyncBit			Set if HW can disable Vertical Sync
//				kDisableCompositeSyncBit		Set if HW can disable Composite Sync
//				kSyncOnRedEnableBit				Set if HW can sync on Red
//				kSyncOnGreenEnableBit			Set if HW can sync on Green
//				kSyncOnBlueEnableBit			Set if HW can sync on Blue
//				kNoSeparateSyncControlBit		Set if HW CANNOT enable/disable H,V,C sync independently
//												Means that HW ONLY supports the VESA "OFF" state
//
//				If 'getHardwareSyncCapability  = false' then report the current state of sync lines and 
//				if HW is "syncing" on Red, Green or Blue.
//				Reporting the "current state of the sync lines" means: "Report the State of the Display"
//				To report the current state of the display, the 'kDisableHorizontalSyncBit' and
//				the 'kDisableVerticalSyncBit' have the following interpretation:
//
//				State       	kDisableVerticalSyncBit		kDisableHorizontalSyncBit		Video
//				-----			-----------------------		-------------------------		------
//				Active       		  	0      						0      					Active
//				Standby					0   						1		   			   	Blanked
//				Idle					1     			 			0						Blanked
//				Off			 			1							1						Blanked
//
//				To report if HW is "syncing" on Red, Green or Blue:
//				kSyncOnRedEnableBit				Set if HW is "syncing" on Blue
//				kSyncOnGreenEnableBit			Set if HW is "syncing" on Green
//				kSyncOnBlueEnableBit			Set if HW is "syncing" on Blue
//
//=====================================================================================================
GDXErr GraphicsHALGetSync(Boolean getHardwareSyncCapability, VDSyncInfoRec *sync)
{
	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	GDXErr err = kGDXErrUnknownError;								// Assume failure.


	if (getHardwareSyncCapability)
	{
		// Report capability of the hardware.  Cosmo CANNOT control H,V, or C sync independently 
		// and is UNABLE to sync on Red/Green/Blue
		sync->csMode = 0 |
						  1 << kDisableHorizontalSyncBit |
						  1 << kDisableVerticalSyncBit |
						  1 << kDisableCompositeSyncBit |
						  1 << kNoSeparateSyncControlBit;
		sync->csFlags = 0;
	} else {		
		// Report the current status of the various syncs
		sync->csMode = 0;
		sync->csFlags = 0;
		
		if (templateEZHALData->syncsDisabled)
			sync->csMode = kDPMSSyncMask;
	}

	err = kGDXErrNoError;											// Everything okay
		
ErrorExit:

	return (err);
}



//=====================================================================================================
// GraphicsHALSetSync()
//	If the display supported the VESA Device Power Management Standard (DPMS), it would respond
//	to HSync and VSync in the following manner:
//	The VESA Standards are:
//	
//	State       	Vert Sync		Hor Sync		Video
//	-----			--------		---------		------
//	Active         	Pulses      	Pulses      	Active
//	Standby			Pulses   		No Pulses      	Blanked
//	Idle			No Pulses      	Pulses			Blanked
//	Off			 	No Pulses		No Pulses		Blanked
//
//
//	For this routine, the relevant fields of the 'VDSyncInfoRec' structure are as follows:
//			->	syncBitField		bit field of the sync bits that need to be disabled/enabled
//
//				kDisableHorizontalSyncBit		Set if HW should disable Horizontal Sync (No Pulses)
//				kDisableVerticalSyncBit			Set if HW should disable Vertical Sync (No Pulses)
//				kDisableCompositeSyncBit		Set if HW should disable Composite Sync (No Pulses)
//				kSyncOnRedEnableBit				Set if HW should sync on Red
//				kSyncOnGreenEnableBit			Set if HW should sync on Green
//				kSyncOnBlueEnableBit			Set if HW should sync on Blue
//
//
//			->	syncBitFieldValid		Mask of the bits that are valid in the csMode bit field
//
//=====================================================================================================
GDXErr GraphicsHALSetSync(UInt8 syncBitField, UInt8 syncBitFieldValid)
{

		
	HWRegister8Bit graphicsMode;						// For programing Cosmo with correct c,v,h Syncs
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo = &templateEZHALData->cosmo;

	GDXErr err = kGDXErrUnknownError;								// Assume failure
	
	
	if (0 == syncBitFieldValid)					// No valid bits are set, just exit with no error.
	{
		err = kGDXErrNoError;
		goto ErrorExit;
	}

	// Cosmo is UNABLE to sync on Red/Green/Blue and reported that on the GetSync status call.
	// If any of the kEnableSyncOnBlue, kEnableSyncOnGreen or kEnableSyncOnRed bits in csFlags
	// are set than return an error.
	// The only valid values for syncBitFieldValid are 0 (do nothing) and 0x07
	// Since Cosmo ONLY supports turning off H, V, and C as a group, return an error if asked
	// to act on an invalid combination.  The valid combinations are all syncs ON, or all OFF.
	

	if (		(kDPMSSyncMask != syncBitFieldValid) || 
				( (0 != (syncBitField & kDPMSSyncMask)) && (7 != (syncBitField & kDPMSSyncMask))) )
	{
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}

		
	if (syncBitField & kDPMSSyncMask) {
		// Turn off the syncs
		templateEZHALData->syncsDisabled = true;
		*cosmo->graphicsMode = 0x9f;							// Blank the video and stop all syncs		
	} else {
		// Turn on the syncs
		templateEZHALData->syncsDisabled = false;
		*cosmo->graphicsMode = templateEZHALData->graphicsMode;	// Restore register to proper value
	}
	
		
	err = kGDXErrNoError;										// Everything okay
		
ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetModeTiming()
//	This is used to to gather scan timing information.  Look at 'displayModeID' and return the 
//	appropriate info.  The 'displayModeID' must be valid for the display type attached.
//
//			->	displayModeID	Describes the dispaly resolution and scan timing
//
//			<-	timingFormat	Format of the info in 'timingData' field, only 'kDeclROMtables' is valid.
//
//			<-	timingFlags		Whether the display mode with these scan timings is required or optional.
//								If a 'displayModeID' is not thought to be valid for given display, set
//								timingFlags to 0 (invalid and unsafe)...the DisplayMgr will ask
//								display modules if the mode is valid.
//
//=====================================================================================================
GDXErr GraphicsHALGetModeTiming(DisplayModeID displayModeID, UInt32 *timingFormat, UInt32 *timingFlags)
{
	
	// Define a new type which is a table of supported 'DisplayCodes' and their corresponding
	// applicable 'DisplayModeIDs' and their 'timingFlags'
	
	typedef struct DisplayModeTimingTable DisplayModeTimingTable;
	struct DisplayModeTimingTable
	{
		DisplayCode displayCode;
		DisplayModeID displayModeID;
		UInt32 timingFlags;
	};
	
	enum
	{	
		kMaxTimingTableEntries = 11,
		kValid = (1 << kModeValid),
		kValidAndSafe = (1 << kModeValid | 1 << kModeSafe),
		kValidAndSafeAndDefault = ( (1 << kModeValid) | (1 << kModeSafe) | (1 << kModeDefault) )
	};
	
	DisplayModeTimingTable theDisplayModeTimingTable[kMaxTimingTableEntries] =
	{
		{kDisplayCodeVGA, kDisplay640x480At60HzVGA, kValidAndSafeAndDefault},
		{kDisplayCodeStandard, kDisplay640x480At67Hz, kValidAndSafeAndDefault},
		{kDisplayCodePortraitMono, kDisplay640x870At75Hz, kValidAndSafeAndDefault},
		{kDisplayCodePortrait, kDisplay640x870At75Hz, kValidAndSafeAndDefault},
		{kDisplayCode16Inch, kDisplay832x624At75Hz, kValidAndSafeAndDefault},
	
		{kDisplayCodeMultiScanBand1, kDisplay640x480At67Hz, kValidAndSafeAndDefault},
		{kDisplayCodeMultiScanBand1, kDisplay832x624At75Hz, kValidAndSafe},

		{kDisplayCodeMultiScanBand2, kDisplay640x480At67Hz, kValidAndSafe},
		{kDisplayCodeMultiScanBand2, kDisplay832x624At75Hz, kValidAndSafeAndDefault},
		
		{kDisplayCodeMultiScanBand3, kDisplay640x480At67Hz, kValidAndSafe},
		{kDisplayCodeMultiScanBand3, kDisplay832x624At75Hz, kValidAndSafeAndDefault},
	};
	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	
	DisplayCode displayCode = templateEZHALData->displayCode;	// Class of connected display
	
	UInt32 i;												// Loop iterator
	Boolean modePossible = false;							// Used for paranoid error checking

	
	GDXErr err = kGDXErrNoError;					// Never fails.  'timingFlags = 0' if don't think
													// display supports the 'displayModeID'

	*timingFlags = 0;				// Default to invalid and unsafe.  If HAL doesn't know if a monitor
									// supports a 'displayModeID', this will allow the Display Mgr to
									// ask a display module.


	*timingFormat = kDeclROMtables;	// Default to kDeclROMtables -- the only valid timingFormat

	// Prior to doing anything else, do some paranoid error checking, and make sure that Cosmo
	// can drive the indicated 'displayModeID.'  This can be accomplished by calling the
	// GraphicsHALModePossible() routine.
	
	err = GraphicsHALModePossible(displayModeID, kDepthMode1, 0, &modePossible);
	if (err || !modePossible)
	{
		// Opps...caller specified an invalid 'DisplayModeID'
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	
	// Scan the the 'DisplayModeTimingTable' to see if the connected monitor supports the 'displayModeID'
		
	for (i = 0; i < kMaxTimingTableEntries; i++)
	{
		if ((theDisplayModeTimingTable[i].displayCode == displayCode) && 
				(theDisplayModeTimingTable[i].displayModeID == displayModeID))
		{
				*timingFlags = theDisplayModeTimingTable[i].timingFlags;
				break;
	
		}
	}
		
ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetNextResolution()
//	This call will take a 'previousDisplayModeID' and return the next supported display mode.
//	The Core takes care of most of the work.  The HAL just needs to look at the 
//	'previousDisplayModeID' and return the next supported d'isplayModeID' and the max depthMode
//	that the HW supports.
//	The Core has already checked to see if previousDisplayModeID = -1, in which case it has
//	returned the current resolution's information.
//	The HAL returns all DisplayModeIDs that it supports.  Pay No attention to the monitor behind
//	the sense lines.
//
//			->	previousDisplayModeID		
//			'previousDiplayModeID = kDisplayModeIDCurrent' never happens.  Core will handle that case. 
//			If 'previousDiplayModeID = kDisplayModeIDFindFirstResolution', get the first supported 
//			resolution for the monitor.
//			Otherwise, 'previousDiplayModeID' contains the previous displayModeID (hence its name)
//			from the previous call.
//
//			<-	displayModeID	ID of the next display mode following 'csPreviousDiplayModeID'
//			Set to 'kDisplayModeIDNoMoreResolutions' once all supported display modes have been reported.
//
//			<-	maxDepthMode	Maximum relative bit depth for the 'displayModeID'
//
//=====================================================================================================
GDXErr GraphicsHALGetNextResolution(DisplayModeID previousDisplayModeID,
		DisplayModeID *displayModeID, DepthMode *maxDepthMode)
{
	GDXErr err = kGDXErrUnknownError;								// Assume failure	
			
	
enum { kMaxResolutions = 4 };

	DisplayModeID theResolutionTable[kMaxResolutions] =
	{
		kDisplay640x480At60HzVGA,
		kDisplay640x480At67Hz,
		kDisplay640x870At75Hz,
		kDisplay832x624At75Hz,
	};
	
	
	
	UInt32 i;									// Loop iterator
	UInt32 displayModeIDIndex;					// Index for the returned DisplayModeID
	
	// If 'kDisplayModeIDFindFirstResolution == previousDisplayModeID', return the first resolution 
	// that the HW supports.

	if (kDisplayModeIDFindFirstResolution == previousDisplayModeID)
	{
		// Index to first supported resolution
		displayModeIDIndex = 0;		
	} else {
		// Find the previousDisplayModeID in 'theResolutionTable'.
		i = 0;
		while ( (i < kMaxResolutions) &&  (theResolutionTable[i] != previousDisplayModeID) )
			i++;

		// Check if i has exceeded kMaxResolutions.  This means the 'previousDisplayModeID' was not valid.
	
		if (kMaxResolutions == i)
		{
			err = kGDXErrInvalidParameters;
			goto ErrorExit;
		}
		
		displayModeIDIndex = ++i;		// Index to next supported resolution. (or no more)
	}
	
	// DisplayModeIndex now points to the next resolution.
	// Make sure 'displayModeIndex != kMaxResolutions'... if it does, all resolutions have been reported.

	if (displayModeIDIndex < kMaxResolutions)
	{
		*displayModeID = theResolutionTable[displayModeIDIndex];
		err = GraphicsHALGetMaxDepthMode(*displayModeID, maxDepthMode);
	} 
	else
	{
		*displayModeID = kDisplayModeIDNoMoreResolutions;		// No more supported resolutions
		err = kGDXErrNoError;									// Everything okay
	}
	
	
ErrorExit:
	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetVideoParams()
//	The HAL only needs to return the absolute bits per pixel and the rowBytes for a
//	given depthMode.  The rowBytes, on input, contains the number of horizontal pixels
//	for the displayModeID.
//
//		-> depthMode	The relative bit depth for which the info is desired.
//		<- bitsPerPixel	Absolute bit depth for the given depthMode.
//		<> rowBytes		On input, rowbytes contains the horizontal pixels for the dispalyModeID.
//						On output, width between successive rows of video memory for the given depthMode.
//
//=====================================================================================================
GDXErr GraphicsHALGetVideoParams(DisplayModeID displayModeID, DepthMode depthMode,
		UInt32 *bitsPerPixel, SInt16 *rowBytes)
{
	GDXErr err = kGDXErrUnknownError;				// Assume failure	

	err = GraphicsHALMapDepthModeToBPP(depthMode, bitsPerPixel);
	if (err)
		goto ErrorExit;
	
	switch (*bitsPerPixel)
	{
		case 8:
			 *rowBytes = *rowBytes;
			break;
		case 16:
			*rowBytes = (*rowBytes << 1);
			break;
		case 32:
			*rowBytes = (*rowBytes << 2);
			break;
		default:
			err = kGDXErrInvalidParameters;				// Invalid depth...shouldn't happen
	}


ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetMaxDepthMode()
// This takes a 'displayModeID' and returns the maximum depthMode that is supported by the
// hardware for that 'displayModeID'  NO check is made to determine if the 'displayModeID' is
// valid for the connected monitor.  The HAL should return an error if the 'displayModeID' is
// not supported or there is not enough VRAM to support the 'displayModeID'
//
//			->	displayModeID		Get the information for this display mode
//			<-	maxDepthMode		Maximum relative bit depth for the DisplayModeID
//
//
//=====================================================================================================
GDXErr GraphicsHALGetMaxDepthMode(DisplayModeID displayModeID, DepthMode *maxDepthMode)
{

	UInt32 i;											// Loop control variable
	
	// Define a new type which maps the 'DisplayModeIDs' to their corresponding
	// maximum 'DepthMode'.
	
	typedef struct DisplayModeTable DisplayModeTable;
	struct DisplayModeTable
	{
		DisplayModeID displayModeID;
		DepthMode maximumDepthMode;
	};
	
	enum {kTableSize = 4};
	
	// The TemplateEZ architecture can support up to 'kDepthMode3' for all supported displays.
	
	DisplayModeTable displayModeTable[kTableSize] =
	{
		{kDisplay640x480At60HzVGA, 	kDepthMode3},
		{kDisplay640x480At67Hz, 	kDepthMode3},
		{kDisplay640x870At75Hz, 	kDepthMode3},
		{kDisplay832x624At75Hz, 	kDepthMode3}
	};
	
	GDXErr err = kGDXErrDisplayModeIDUnsupported;			// Assume invalid 'DisplayModeID' provided.
			
	// Scan the displayModeTable to see to find the maximum depthmode
		
	for (i = 0 ; i < kTableSize ; i++)
	{
		if (displayModeTable[i].displayModeID == displayModeID)
		{
			*maxDepthMode = displayModeTable[i].maximumDepthMode;
			err = kGDXErrNoError;
			break;
		}
	}

ErrorExit:

	return (err);

}



//=====================================================================================================
//
// GraphicsHALMapDepthModeToBPP()
//	This routine maps a relative pixel depth (DepthMode) to an absolute pixel depth (bits per pixel)
//
//		-> depthMode		The relative pixel depth 
//
//		<- bitsPerPixel		Corresponding abosolute pixel depth.
//
//=====================================================================================================
GDXErr GraphicsHALMapDepthModeToBPP(DepthMode depthMode, UInt32 *bitsPerPixel)
{
	GDXErr err = kGDXErrUnknownError;			// Assume failure.
	
	switch (depthMode)
	{
		case kDepthMode1:
			*bitsPerPixel = 8;
			break;
		case kDepthMode2:
			*bitsPerPixel = 16;
			break;
		case kDepthMode3:
			*bitsPerPixel = 32;
			break;
		default:								// Someone passed in bogus depthMode...say it is 8
			*bitsPerPixel = 8;
			err = kGDXErrUnableToMapDepthModeToBPP;
			goto ErrorExit;
	}
	

	err = kGDXErrNoError;

ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALModePossible()
//	This routine checks to see if the frame buffer is cabable of driving the given 'displayModeID' at
//	the indicated 'depthMode' and 'page'. This DOES NOT check to see that the 'displayModeID' is valid
//	for the display type that is physically connected to the frame buffer.
//
//	IMPORTANT NOTE: The GDXErr return value DOES NOT indicate whether the mode is possible or not. 
//	It only signifies whether or not the value returned in 'modePossible' was correctly determined.  In
//	the event of an error, 'modePossible' does not contain valid information.
//
//		-> displaymodeID		The desired DisplayModeID. 
//		-> depthMode			The desired relative bit depth.
//		-> page					The desired page.
//
//		<- modePossible
//		This will be 'true' if the frame buffer can support the desired items, 'false' otherwise. In 
//		the event of an error, 'modePossible' is undefined.
//
//=====================================================================================================
GDXErr GraphicsHALModePossible(DisplayModeID displayModeID, DepthMode depthMode, SInt16 page,
		Boolean *modePossible)
{
	
	DepthMode maxDepthMode;						// The max depthMode for the DisplayModeID

	GDXErr err = kGDXErrNoError;				// If displayModeID, depthMode or page isn't supported
												// just return modePossible = false

	*modePossible = false;						// Assume mode not possible
	
	// Check to see if the requested page is valid.  For TemplateEZ hardware, page 0
	// is the only valid page supported.
	
	if (0 != page)
		goto ErrorExit;
		
	// Find the maximum depthMode for the displayModeID
	// If the displayModeID is not supportted by hardware, GetMaxDepthMode() returns
	// kGDXErrDisplayModeIDUnsupported 
	
	err =  GraphicsHALGetMaxDepthMode(displayModeID, &maxDepthMode);

	if (err)
		goto ErrorExit;

	if (maxDepthMode >= depthMode)
		*modePossible = true;

ErrorExit:

	return (kGDXErrNoError);
}



//=====================================================================================================
//
// GraphicsHALDetermineDisplayCode()
//	This routine is called whenever it is necessary to determine the type of display that is
//	connected to the frame buffer controller.
//	When this routine is called, the following actions should occur:
//
//		 Perform required steps to determine what display is connected (e.g., read sense lines)
//		 Update the HAL's state information regarding the type of display connected (if HAL
//		  implementation maintains that state information)
//
//	In the event that the HAL is does not the specific type of display attached, it should set
//	'*displayCode = kDisplayCodeUnknown'
//
//			<- displayCode	DisplayCode for the attached display.
//
//=====================================================================================================
GDXErr GraphicsHALDetermineDisplayCode(DisplayCode *displayCode)
{
	RawSenseCode rawSenseCode;						// Result from reading sense lines after reset
	ExtendedSenseCode extendedSenseCode;			// Result from applying extended sense algorithm 

	GDXErr err = kGDXErrNoError;					// Assume sucess

	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	
	// The rawSenseCode is obtained by making sure the sense lines are reset (i.e., none of the lines
	// are being actively driven by the frame buffer controller) and then reading the state of the
	// lines.

	TemplateEZResetSenseLines();						
	rawSenseCode = TemplateEZReadSenseLines();	

	extendedSenseCode = TemplateEZGetExtendedSenseCode();
	
	// Since the TemplateEZ architecture uses 'standard' sense codes, call the utility function
	// that maps the raw and extendended sense codes to a 'displayCode.'
	
	err = GraphicsUtilMapSenseCodesToDisplayCode(rawSenseCode, extendedSenseCode, displayCode);
	if (err)
		goto ErrorExit;

	templateEZHALData->displayCode = *displayCode;		// Update HAL information

ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetSenseCodes()
//	This routine is called whenever the state of the sense codes need to be reported.
//	NOTE:  This should only report the the sense code information.  No attempt should be made to
//	determine what type of display is attached here.  Moreover, the sense codes should be
//	determined EVERY time this call is made, and not make use of any previously saved values.
//
//		<- rawSenseCode
//		For 'standard' sense code hardware, this value is found by instructing the frame buffer
//		controller NOT to actively drive any of the monitor sense lines, and then reading the
//		state of the the monitor sense lines 2, 1, and 0.  (2 is the MSB, 0 the LSB)
//
//		<- extendedSenseCode
//		For 'standard' sense code hardware, the extended sense code algorithm is as follows:
//		(Note:  as described here, sense line 'A' corresponds to '2', 'B' to '1', and 'C' to '0')
//			 Drive sense line 'A' low and read the values of 'B' and 'C'.  
//			 Drive sense line 'B' low and read the values of 'A' and 'C'.
//			 Drive sense line 'C' low and read the values of 'A' and 'B'.
//
//		In this way, a six-bit number of the form BC/AC/AB is generated. 
//
//		<- standardInterpretation
//		If 'standard' sense code hardware is implemented (or the values are coerced to appear
//		'standard' then set this to 'true'.  Otherwise, set it to 'false', and the interpretation for
//		'rawSenseCode' and 'extendedSenseCode' will be considered private.
//
//=====================================================================================================
GDXErr GraphicsHALGetSenseCodes(RawSenseCode *rawSenseCode, ExtendedSenseCode *extendedSenseCode,
		Boolean *standardInterpretation)
{
	GDXErr err = kGDXErrNoError;							// Assume sucess
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	
	// The rawSenseCode is obtained by making sure the sense lines are reset (i.e., none of the lines
	// are being actively driven by the frame buffer controller) and then reading the state of the
	// lines.
	
	TemplateEZResetSenseLines();						
	*rawSenseCode = TemplateEZReadSenseLines();	

	*extendedSenseCode = TemplateEZGetExtendedSenseCode();
		
	*standardInterpretation = true;						// TemplateEZ has 'standard' sense lines
			
ErrorExit:
	return (err);
}



//=====================================================================================================
//
// GraphicsHALGetDefaultDisplayModeID()
//	The 'displayCode' is passed in and the HAL returns the default 'displayModeID' and the
//	'depthMode'.  This routine gets called when a new monitor is connected to the computer.  The
//	HAL knows how much VRAM is available and whether it can handle a given 'displayModeID' for a 
//	monitor.
//	For example, the 'kIndexedMultiScanBand3' has a default 'displayModeID' of kDisplay1024x768At7500Hz.
//	If there is not enough VRAM available, the HAL is unable to switch into that resolution, hence
//	the HAL will switch into the next best resolution.
//	The HAL will also return the max 'depthMode' that can be supported for the resolution.  This
//	enables a switch from a 13" monitor at millions of colors to be switched with a 21" monitor
//	and, if there is enough VRAM, the depthMode will still be at millions of colors
//
//		-> displayCode			The connected display
//		<- displaymodeID		The default DisplayModeID for the connected display type
//		<- depthMode			The default depthMode 
//
//=====================================================================================================
GDXErr GraphicsHALGetDefaultDisplayModeID(DisplayCode displayCode, DisplayModeID *displayModeID,
		DepthMode *depthMode)
{
	// Define a new type which is a table of default 'DisplayModeIDs' for the 
	// connected monitor.
	
	typedef struct DefaultResolutionTable DefaultResolutionTable;
	struct DefaultResolutionTable
	{
		DisplayCode displayCode;
		DisplayModeID displayModeID;
	};
	
	enum {kMaxDefaultTableEntries = 8};

	DefaultResolutionTable defaultResolutionTable[kMaxDefaultTableEntries] =
	{
		{kDisplayCodeVGA, kDisplay640x480At60HzVGA},
		{kDisplayCodeStandard, kDisplay640x480At67Hz},
		{kDisplayCodePortrait, kDisplay640x870At75Hz},
		{kDisplayCodePortraitMono, kDisplay640x870At75Hz},
		{kDisplayCode16Inch, kDisplay832x624At75Hz},
		{kDisplayCodeMultiScanBand1, kDisplay640x480At67Hz},
		{kDisplayCodeMultiScanBand2, kDisplay832x624At75Hz},
		{kDisplayCodeMultiScanBand3, kDisplay832x624At75Hz}
	};
	
	UInt32 i;										// Loop iterator
	GDXErr err = kGDXErrMonitorUnsupported;			// Assume failure	
	
	
	// Scan the defaultResolutionTable to find the the connected monitor
	
	for (i = 0; i < kMaxDefaultTableEntries; i++)
	{
		if (defaultResolutionTable[i].displayCode == displayCode)
		{
			*displayModeID = defaultResolutionTable[i].displayModeID;
			err = kGDXErrNoError;
			break;					// Found displayCode
		}
	}

	if (err)											// Didn't find displayCode
		goto ErrorExit;
		
	err = GraphicsHALGetMaxDepthMode(*displayModeID, depthMode);

	
ErrorExit:

	return (err);
}



//=====================================================================================================
//
// GraphicsHALProgramHardware()
//	This routine attempts to program the graphics hardware to the desired 'displayModeID', 'depthMode',
//	and 'page'.  It is not required to specifically check to see if the inputs are valid, since it
//	can assume that the checking has been done elsewhere.
//
//		-> displaymodeID		The desired DisplayModeID. 
//		-> depthMode			The desired relative bit depth.
//		-> page					The desired page.
//
//		<- directColor
//		This is 'tru'e if the desired depthMode results in the hardware being in a direct color mode,
//		otherwise it is 'false'. In the event on an error, it is undefined.
//
//		<- baseAddress
//		The resulting base address of the frame buffers ram. In the event of an error, it is undefined.
//
//=====================================================================================================
GDXErr GraphicsHALProgramHardware(DisplayModeID displayModeID, DepthMode depthMode, SInt16 page,
		Boolean *directColor, char **baseAddress)
{
	// Define a new type which maps a 'DisplayModeID' to the value that will represent
	// the appropriate 'GraphicsMode' for the hardware.
	
	typedef struct DisplayModeIDToVideoModeMap DisplayModeIDToVideoModeMap;
	struct DisplayModeIDToVideoModeMap
	{
		DisplayModeID displayModeID;
		HWRegister8Bit graphicsMode;
	};
	
	enum {kDMIDMapSize = 4};
	
	DisplayModeIDToVideoModeMap displayModeIDMap[kDMIDMapSize] = 
	{
		{kDisplay640x480At60HzVGA, 	0x0b},
		{kDisplay640x480At67Hz, 	0x06},
		{kDisplay640x870At75Hz, 	0x01},
		{kDisplay832x624At75Hz, 	0x09}
	};
		
	
	// Define a new type which maps a 'DepthMode' to the value that will represent
	// the appropriate bit depth for the hardware.  
		
	typedef struct DepthModeToBitDepthMap DepthModeToBitDepthMap;
	struct DepthModeToBitDepthMap
	{
		DepthMode depthMode;
		HWRegister8Bit bitDepth;			// Data for Cosmo's 'bitDepth' register
		HWRegister8Bit control;				// Data for Irazu's 'control' register
	};
	
	enum {kDMMapSize = 3};
	
	DepthModeToBitDepthMap depthModeMap[kDMMapSize] =
	{
		{kDepthMode1, 0x00, 0x00},
		{kDepthMode2, 0x01, 0x01},
		{kDepthMode3, 0x02, 0x02},
	};

	GDXErr err = kGDXErrUnknownError;					// Assume failure
	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo;
	IrazuRegisters *irazu;

	HWRegister8Bit bitDepth;
	HWRegister8Bit graphicsMode;
	HWRegister8Bit control;
	
	UInt16 i;
	
	cosmo = &templateEZHALData->cosmo;
	irazu = &templateEZHALData->irazu;
	
	// Make sure that we know what to program for the requested 'displayModeID'
	
	err = kGDXErrDisplayModeIDUnsupported;			// Assume invalid displayModeID
	
	for (i = 0 ; i < kDMIDMapSize ; i++)
	{
		if (displayModeIDMap[i].displayModeID == displayModeID)
		{
			graphicsMode = displayModeIDMap[i].graphicsMode;
			err = kGDXErrNoError;
			break;
		}
	}

	if (err)
		goto ErrorExit;

	// Make sure that we know what to program for the requested 'depthMode'
	
	err = kGDXErrDepthModeUnsupported;				// Assume invalid depthmode
	
	for (i = 0 ; i < kDMMapSize ; i++)
	{
		if (depthModeMap[i].depthMode == depthMode)
		{
			bitDepth = depthModeMap[i].bitDepth;
			control = depthModeMap[i].control;
			err = kGDXErrNoError;
			break;
		}
	}

	if (err)
		goto ErrorExit;

		
	*cosmo->graphicsMode = graphicsMode;
	templateEZHALData->graphicsMode = graphicsMode;	// Save this, for it will be needed in SetSync()
	
	*cosmo->bitDepth = bitDepth;
	*irazu->control = control;
		
	if (depthMode > kDepthMode1)		
		*directColor = true;						// > 8 bpp, so directColor == true
	else
		*directColor = false;						// == 8 bpp, so using indexed (CLUT) color
		
	*baseAddress = templateEZHALData->vramBaseAddress;	// Frame buffer base for all DisplayModeIDs
	
ErrorExit:

	return (err);



}



//=====================================================================================================
//
// GraphicsHALDrawHardwareCursor()
//	This routine sets the cursor's X and Y coordinates and its visible state.  If the cursor was set
//	successfully by a previous call to GraphicsHALSetHardwareCursor(), then the HAL must program the
//	hardware with the given X, Y and visible state.  If the previous call to
//	GraphicsHALSetHardwareCursor() failed, then an error should be returned.
//
//		-> x		X coordinate
//		-> y		Y coordinate
//		-> visible	'true' if the cursor must be visible.
//
//=====================================================================================================
GDXErr  GraphicsHALDrawHardwareCursor(SInt32 x, SInt32 y, Boolean visible)
{

	// TemplateEZ does not support a hardware cursor, so return an error.
	
	return (kGDXErrUnsupportedFunctionality);
}



//=====================================================================================================
//
// GraphicsHALSetHardwareCursor()
//	This routine is called to setup the hardware cursor and determine if whether the hardware can
//	support it.  The HAL should remember whether this call was successful for subsequent
//	GetHardwareCursorDrawState() or DrawHardwareCursor() calls, but should NOT change the cursor's
//	X or Y coordinates, nor its visible state.
//	
//		-> gamma
//		Current gamma table to correct cursor colors with, if the HAL can apply gamma correction.
//
//		-> luminanceMapping
//		This will be true if the Core had luminance mapping enabled AND it was in an indexed color
//		mode.  If 'true', the HAL should luminance map the cursor CLUT EVEN if the hardware cursor is
//		a super-duper cursor capable of direct color. This is because the hardware cursor should look
//		like the software cursor it is replacing.
//
//		-> cursorRef
//		Opaque data to be handed to VSLPrepareCursorForHardwareCursor().
//
//=====================================================================================================
GDXErr GraphicsHALSetHardwareCursor(const GammaTbl *gamma, Boolean luminanceMapping, void *cursorRef)
{
	// TemplateEZ does not support a hardware cursor, so return an error.

	return (kGDXErrUnsupportedFunctionality);
}



//=====================================================================================================
//
// GraphicsHALSupportsHardwareCursor()
//	This call is used to determine if the HAL supports a hardware cursor.
//
//		<- supportsHardwareCursor	'true' if supports a hardware cursor, 'false' otherwise.
//
//=====================================================================================================
GDXErr GraphicsHALSupportsHardwareCursor(Boolean *supportsHardwareCursor)
{
	// TemplateEZ does not support a hardware cursor, so return an 'false'.

	*supportsHardwareCursor = false;

	return (kGDXErrNoError);
}



//=====================================================================================================
//
// GraphicsHALGetHardwareCursorDrawState()
//	This routine is used to determine the state of the hardware cursor.  After HAL initialization
//	the cursors visible state and set state should be false. After a mode change the cursor should be
//	made invisible but the set state should remain unchanged.
//
//		<- csCursorX		X coordinate from last DrawHardwareCursor call
//		<- csCursorY		Y coordinate from last DrawHardwareCursor call
//		<- csCursorVisible	'true' if the cursor is visible
//		<- csCursorSet		'true' if last GraphicsHALSetHardwareCursor() call was successful.
//
//=====================================================================================================
GDXErr GraphicsHALGetHardwareCursorDrawState(SInt32  * cursorX, SInt32  * cursorY,
		UInt32  *cursorVisible, UInt32  *cursorSet)
{
	// TemplateEZ does not support a hardware cursor, so return an error.

	return (kGDXErrUnsupportedFunctionality);
}


//=====================================================================================================
//
// GraphicsHALSetPowerState()
//	The graphics hw might have the ability to to go into some kind of power saving mode.  Just
//	pass the call to the HAL
//
//	For this routine, the relevant fields indicated by 'VDPowerStateRec' are:
//			->	powerState		desired power mode: kAVPowerOff, kAVPowerStandby, kAVPowerSuspend,
//								kAVPowerOn
//
//			<- powerFlags		kPowerStateNeedsRefresh bit set if hw needs to be refreshed after
//								coming out of the designated power state.
//
//=====================================================================================================
GDXErr GraphicsHALSetPowerState(VDPowerStateRec *powerState)
{
	
	// TemplateEZ does not support setting its ASICs in a low power mode.

	return (kGDXErrUnsupportedFunctionality);
}



//=====================================================================================================
//
// GraphicsHALGetPowerState()
//	The graphics hw might have the ability to to go into some kind of power saving mode.  Just
//	pass the call to the HAL
//
//	For this routine, the relevant fields indicated by 'VDPowerStateRec' are:
//			<- powerState		current power mode: kAVPowerOff, kAVPowerStandby, kAVPowerSuspend,
//								kAVPowerOn
//
//			<- powerFlags		kPowerStateNeedsRefresh bit set if hw needs to be refreshed after
//								coming out of power state
//
//=====================================================================================================
GDXErr GraphicsHALGetPowerState(VDPowerStateRec *avPowerState)
{	
	// TemplateEZ does not support setting its ASICs in a low power mode.

	return (kGDXErrUnsupportedFunctionality);
}



//=====================================================================================================
//
// GraphicsHALPrivateControl()
//	Routine accepts private control calls.  The core passes all control codes that it doesn't
//	understand to this routine.  If the HAL knows what to do with the code, it hands it off
//	accordingly.
//	
//		-> genericPtr			Points to the data structure that the HAL needs for this control
//								call.  Should be cast to appropriate data type if internal routine is
//								invoked.
//		-> privateControlCode	The private csCode that the HAL might know what to do with.
//
//
//=====================================================================================================
OSErr GraphicsHALPrivateControl(void * genericPtr, SInt16 privateControlCode)
{

	GDXErr err = kGDXErrUnknownError;				// Assume failure.
	OSErr returnErr = controlErr;					// Assume cs code is invalid

	switch (privateControlCode)
	{
		default:
			goto ErrorExit;							// csCode not supported
	}
	
ErrorExit:
	return (returnErr);
}



//=====================================================================================================
//
// GraphicsHALPrivateStatus()
//	Routine accepts private status calls.  The core passes all status codes that it doesn't
//	understand to this routine.  If the HAL knows what to do with the code, it hands it off
//	accordingly.
//	
//		-> genericPtr			Points to the data structure that the HAL needs for this status
//								call.  Should be cast to appropriate data type if internal routine is
//								invoked.
//		-> privateStatusCode	The private csCode that the HAL might know what to do with.
//
//
//=====================================================================================================
OSErr GraphicsHALPrivateStatus(void * genericPtr, SInt16 privateStatusCode)
{


	GDXErr err = kGDXErrUnknownError;				// Assume failure.
	OSErr returnErr = statusErr;					// Assume cs code is invalid

	switch (privateStatusCode)
	{
		default:
			goto ErrorExit;							// csCode not supported
	}
	
ErrorExit:
	return (returnErr);
}



//=====================================================================================================
//
// TemplateEZMapDepthModeToCLUTAttributes()
//	This simple routine maps a 'DepthMode' to the corresponding CLUT attritributes associate with it.
//	
//		-> depthMode	The 'DepthMode' for which the CLUT attributes are desired.
//		<- startAddress	The physical address that corresponds to logical address 0.
//		<- entryOffset	The physical offset between each logical address.
//
//	This routine is private to the TemplateEZ Graphics HAL.
//
//=====================================================================================================
static GDXErr TemplateEZMapDepthModeToCLUTAttributes(DepthMode depthMode, UInt32 *startAddress,
		UInt32 *entryOffset)
{
	// Define a new type which maps a 'DepthMode' to the corresponding physical attributes of a CLUT.
	// These attributes are needed because the logical addresses when in 1, 2, or 4 bpp do not
	// correspond to the physical address.  Instead, they are distributed evenly through out the CLUT
	// address space.
	
	typedef struct DepthModeToCLUTAttributesMap DepthModeToCLUTAttributesMap;
	struct DepthModeToCLUTAttributesMap
	{
		DepthMode depthMode;			// DepthMode for which these attributes apply
		UInt32 startAddress;			// Physical address of logical address 0 
		UInt32 entryOffset;				// Physical offset between each logical address
	};

	enum {kMapSize = 3};
	
	DepthModeToCLUTAttributesMap depthModeMap[kMapSize] =
	{
		{kDepthMode1, 0x00, 0x01},				// 8 bpp
		{kDepthMode2, 0x00, 0x08},				// 16 bpp
		{kDepthMode3, 0x00, 0x01},				// 32 bpp
	};

	UInt32 i;													// Loop control variable

	GDXErr err = kGDXErrDepthModeUnsupported;					// Assume invalid depthmode

	// Scan the 'DepthModeToCLUTAttributesMap' to find CLUT attributes for the given depth mode.	
	
	for (i = 0 ; i < kMapSize ; i++)
	{
		if (depthModeMap[i].depthMode == depthMode)
		{
			*startAddress = depthModeMap[i].startAddress;
			*entryOffset = depthModeMap[i].entryOffset;
			err = kGDXErrNoError;
			break;
		}
	}

ErrorExit:
	
	return (err);
}


//=====================================================================================================
//
// TemplateEZHandleVBLInterrupts()
//	This routines task is to clear the internal VBL interrupt source and re-prime it
// 	for the next occurrence.  However, TemplateEZ does not require any clearing or re-priming.
//
//=====================================================================================================
static void TemplateEZHandleVBLInterrupts(void *vblRefCon)
{
	
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo;
	
	cosmo = &templateEZHALData->cosmo;
	
	*cosmo->interruptControl &= 0x7F;		// Clear most signifant bit to clear/re-reprime VBL ints.
}



//=====================================================================================================
//
// TemplateEZResetSenseLines()
//	Before reading the Sense Lines to determine what monitor is connected, the sense lines
//	need to be reset.  Puts sense lines in a known state so the sensing process can start.
//
//=====================================================================================================
static void TemplateEZResetSenseLines(void)
{
	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();

	CosmoRegisters *cosmo  = &templateEZHALData->cosmo;
	*cosmo->monitorSense = 0x07;								//  No sense lines being driven. 
	
	SynchronizeIO();
	
	TemplateEZDelayForDuration(3*durationMillisecond);			// Wait for lines to stabilize
}
	


//=====================================================================================================
//
// TemplateEZDriveSenseLines()
//	Drive the approprite sense line so that the two other sense lines can be read.
//
//		-> senseLine		the sense line to drive.
//
//=====================================================================================================
static void TemplateEZDriveSenseLines(SenseLine senseLine)
{
	enum
	{
		kDriveAValue = 0x03,				
		kDriveBValue = 0x05,
		kDriveCValue = 0x06
	};

	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo  = &templateEZHALData->cosmo;

	UInt32 senseLineValue;

	
	switch (senseLine)
	{
		case kSenseLineA:
			senseLineValue = kDriveAValue;
			break;
		case kSenseLineB:
			senseLineValue = kDriveBValue;
			break;
		case kSenseLineC:
			senseLineValue = kDriveCValue;
			break;
	}
	
	*cosmo->monitorSense = senseLineValue;				// Drive appropriate sense line

	SynchronizeIO();
	
	TemplateEZDelayForDuration(3*durationMillisecond);			// Wait for lines to stabilize

}



//=====================================================================================================
//
// TemplateEZReadSenseLines()
//	Read the state of the sense lines.
//
//		<- senseLineValue		read value of the sense lines
//
//=====================================================================================================
static UInt32 TemplateEZReadSenseLines(void)
{

	enum { kTemplateEZSenseLineMask = 0x70 };

	TemplateEZHALData *templateEZHALData = GraphicsHALGetHALData();
	CosmoRegisters *cosmo  = &templateEZHALData->cosmo;

	UInt32 senseLineValue;
		
	senseLineValue = *cosmo->monitorSense;				// Read sense lines

	senseLineValue &= kTemplateEZSenseLineMask;				// Mask off the sense pins
	
	senseLineValue = (senseLineValue >> 4);				// Get sense bits into least significant bits

	TemplateEZResetSenseLines();							// Reset the sense lines
	
	return(senseLineValue);
}



//=====================================================================================================
//
//
// TemplateEZGetExtendedSenseCode()
//	This routine applies the 'standard' extended sense code algorithm to the sense lines to determine
//	their ExtendedSenseCode.
//
// 	For 'standard' sense line hardware, the extended sense code algorithm is as follows:
//	(Note:  as described here, sense line 'A' corresponds to '2', 'B' to '1', and 'C' to '0')
//		 Drive sense line 'A' low and read the values of 'B' and 'C'.  
//		 Drive sense line 'B' low and read the values of 'A' and 'C'.
//		 Drive sense line 'C' low and read the values of 'A' and 'B'.
//
//	In this way, a six-bit number of the form BC/AC/AB is generated. 
//
//=====================================================================================================
static ExtendedSenseCode TemplateEZGetExtendedSenseCode()
{
	ExtendedSenseCode extendedBC;								// Result from driving line A
	ExtendedSenseCode extendedAC;								// Result from driving line B
	ExtendedSenseCode extendedAB;								// Result from driving line C
	
	ExtendedSenseCode extendedSenseCode;						// Final value.
	
	// Obtain the extendedBC value by driving sense line A
	TemplateEZDriveSenseLines(kSenseLineA);
	extendedBC = TemplateEZReadSenseLines();						// xx xx BC
	extendedBC = (extendedBC << 4) & 0x30;						// BC 00 00
	
	// Obtain the extendedAC value by driving sense line B
	TemplateEZDriveSenseLines(kSenseLineB);
	extendedAC = TemplateEZReadSenseLines();						// xx xA xC
	extendedAC = ((extendedAC >> 2) << 3) |						// (00 A0 00) | (00 0C 00) = 00 AC 00
				 ((extendedAC & 0x1) << 2);
				 
	// Obtain the extendedAB value by driving sense line C
	TemplateEZDriveSenseLines(kSenseLineC);
	extendedAB = TemplateEZReadSenseLines();						// xx xA Bx
	extendedAB = (extendedAB >> 1) & 0x3;						// 00 00 AB
	
	extendedSenseCode = extendedBC | extendedAC | extendedAB;
	
	return(extendedSenseCode);
}



//=====================================================================================================
//
// TemplateEZDelayForDuration()
//	The DSL routine DelayFor() cannot be called at Primary Interrupt Level or Secondary Task Level.
//	This provides equivalent functionality by calling the special DelayForHardware().
//
//		-> duration		the number of microseconds/milliseconds for the delay
//						if the number is negative, it is the # of microseconds to delay
//						if the number is positive, it is the # of milliseconds to delay
//
//=====================================================================================================
static void TemplateEZDelayForDuration(Duration duration)
{
	AbsoluteTime delayTime;						// The duration of delay in # ticks

	delayTime = DurationToAbsolute(duration);
	DelayForHardware(delayTime);
	
}



//=====================================================================================================
//
// TemplateEZWaitForVBL()
//
//	This routine does not return until it senses that a VBL has occurred.  It is usually called before
//  updating the contents of the CLUT, with the assumption that the CLUT can be written during the 
//  Vertical Blanking Interval.  
//
//	This routine does not make any assumptions about whether VBL interrupts are
//	enabled or disabled prior to its calling.  However, it is recommended that the
//	VBL interrupts be disabled prior to calling this if any type of accurate waiting
//	takes place.
//
//
//=====================================================================================================
static void TemplateEZWaitForVBL( void )
{
	// Before waiting in a loop, one should make sure that interrupts are enabled and that any
	// clock sources that cause the interrupts to fire away are also active.  If that is not the
	// case, the routine can exit right away.  Otherwise, sit in a loop waiting for the interrupt.
	
	while (true)
	{
		// Read the interrupt status register and call SynchronizeIO to make sure that we are sending
		// the reads in order.
		SynchronizeIO();
		
		// If set, then break, otherwise continue looping.
	}
}