/*
	File:		GraphicsHALTemplate.c

	Contains:	This file contains the items needed to implement the Graphics HAL
				for an arbitrary Template.
				
	The 'Template' graphic hardware is characterized as follows:
	
		Toynbee - Graphics Controller
		This is the arbitrary name of the graphics controller ASIC.
		Toynbee can be placed in a low power mode in which VRAM decays.
		It can support the following resolutions:


			 512x384 at 60 Hz NTSC	(NTSC Safe Frame)
			 512x384 at 60 Hz		(12" RGB)
			 640x480 at 50 Hz PAL	(PAL Safe Frame)
			 640x480 at 60 Hz NTSC	(NTSC Full Frame)
			 640x480 at 60 Hz VGA	(Common variant of VGA timing)
			 640x480 at 67 Hz 		(Standard Macintosh 'high resolution' display)
			 640x870 at 75 Hz		(Portrait Display)
			 768x576 at 50 Hz PAL	(PAL Full Frame)
			 800x600 at 60 Hz VGA	(Common variant of VGA timing)
			 800x600 at 72 Hz VGA	(Common variant of VGA timing)
			 800x600 at 75 Hz VGA	(Common variant of VGA timing)
			 832x624 at 75 Hz		(16" RGB Display)
			 1024x768 at 60 Hz VGA	(Common variant of VGA timing)
			 1024x768 at 72 Hz VGA	(Common variant of VGA timing)
			 1024x768 at 75 Hz VGA	(Common variant of VGA timing)
			 1024x768 at 75 Hz		(19" RGB)
			 1152x870 at 75 Hz		(21" RGB)
			 1280x960 at 75 Hz		(Vesa timing)
			 1280x1024 at 75 Hz	(Vesa timing)

	
		There is support for hardware cursor, and putting the ASICs into a low power state.
					
		Spur - CLUT/DAC
		This is the arbitrary name of the CLUT/DAC, hereafter merely referred to as a CLUT.
		It is a simple 'triple 8x256' CLUT.  Spur has power saving features.  The PLL clock can
		be turned off and the CLUT can be turned off.  When the CLUT is off, the data decays.
		Spur implements the hardware cursor.
		
		MrSanAntonio - timing generator
		This is the arbitrary name of the timing generator.
		
		The conversion of this Template to a HAL for a specific hardware implementation can be
		started by doing a CASE SENSITIVE replacement of:
		
			"Template" with the "NameOfYourGraphicsArchitectureGoesHere"
			"template" with the "nameOfYourGraphicsArchitectureGoesHere"
			
			"Toynbee" with the "NameOfYourGraphicsControllerGoesHere"
			"toynbee" with the "nameOfYourGraphicsControllerGoesHere"
			
			"Spur" with the "NameOfYourCLUTGoesHere"
			"spur" with the "nameOfYourCLUTGoesHere"

			"MrSanAntonio" with the "NameOfYourTimingGeneratorGoesHere"
			"mrSanAntonio" with the "nameOfYourTimingGeneratorGoesHere"

	Written by:	Kevin Williams

	Copyright:	 1994-1995 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <2>	 7/17/95	SW		In GraphicsHAL GetModeTiming, now setting 'timingForamt =
		 							kDeclROMtables' prior to any error checking.  Also, in
									GraphicsHAL SetSync(), 'sitBitValidField' is now input only.
		 <1>	 4/15/95	SW		First Checked In
			
*/

#include "GraphicsHAL.h"
#include "GraphicsPriv.h"
#include "GraphicsCoreUtils.h"		// for GraphicsUtilMapSenseCodesToDisplayCode()
#include "GraphicsOSS.h"			// for GraphicsOSS Set/Get HALPref

#include <Devices.h>
#include <DriverServices.h>
#include <Errors.h>
#include <Files.h>					// for stages of development, alpha, beta, etc
#include <Kernel.h>					// for DelayForHardware()
#include <PCI.h>
#include <Types.h>
#include <Video.h>
#include <VideoServices.h>


enum
{
	kHardwareCursorOffset = 16,			// Offset to base address to account for the hw cursor
	kRowBytesOffset = 32				// Offset to add to standard rowbytes, to account for hw cursor
};


enum
{
	kHardwareCursorImageSize = (16 * 16) >> 1,	// Size of hardware cursor image in bytes
	kNumHardwareCursorColors = 8				// Number of colors for hardware cursor
};


//
// ToynbeeRegisters
//	This model reflects the version of the Toynbee cell that is in the Template ASIC.
//
typedef struct ToynbeeRegisters ToynbeeRegisters;
struct ToynbeeRegisters
{
	HWRegister32Bit *baseAddress;	
	HWRegister32Bit *rowWords;
	HWRegister32Bit *clockConfiguration;
	HWRegister32Bit *frameBufferConfiguration1;
	HWRegister32Bit *frameBufferConfiguration2;
	HWRegister32Bit *senseLineEnable;
	HWRegister32Bit *reset;
} ;



// following bit posiitions (in 68k ordering) describe the "logical registers"
// in Toynbee
enum
{
	kToynbeeFBCTwoMBitVRAMBit = 11,
	kToynbeeFBCFullVRAMBanksBit = 3,
	kToynbeeFBCFullVRAMBanksMask = 1 << kToynbeeFBCFullVRAMBanksBit
	
};

// following enums tell how much VRAM is in the system
typedef UInt16 VRAMSize;
enum
{
	k1MegVRAM,
	k2MegVRAM,
	k4MegVRAM
};


//
// MrSanAntonioRegisters
//
typedef struct MrSanAntonioRegisters MrSanAntonioRegisters;
struct MrSanAntonioRegisters
{
	HWRegister32Bit *mrSanAntonioMode;
	HWRegister32Bit *interruptMask;
	HWRegister32Bit *interruptStatus;
	HWRegister32Bit *clearCursorInterrupt;
	HWRegister32Bit *clearAnimationInterrupt;
	HWRegister32Bit *clearVBLInterrupt;
	HWRegister32Bit *cursorLine;
	HWRegister32Bit *animateLine;
	HWRegister32Bit *counterTest;
	HWRegister32Bit *horizontalSerration;
	HWRegister32Bit *halfLine;
	HWRegister32Bit *horizontalEqualization;
	HWRegister32Bit *horizontalSyncPulse;
	HWRegister32Bit *horizontalBreezeway;
	HWRegister32Bit *horizontalBurstGate;
	HWRegister32Bit *horizontalBackPorch;
	HWRegister32Bit *horizontalActiveLine;
	HWRegister32Bit *horizontalFrontPorch;
	HWRegister32Bit *horiztonalPixelCount;
	HWRegister32Bit *verticalHalfLine;
	HWRegister32Bit *verticalSync;
	HWRegister32Bit *verticalBackPorchEqualization;
	HWRegister32Bit *verticalBackPorch;
	HWRegister32Bit *verticalActiveLine;
	HWRegister32Bit *verticalFrontPorch;
	HWRegister32Bit *verticalFrontPorchEqualization;
	HWRegister32Bit *timingAdjust;
	HWRegister32Bit *currentLine;
};

// following bit posiitions (in 68k ordering) describe the "logical registers"
// in MrSanAntonio
enum
{
	// MrSanAntonioMode register
	kMrSanAntonioEnableBit = 4,		
	kMrSanAntonioEnableMask  = 1 << kMrSanAntonioEnableBit,

	kHActiveHighBit = 8,
	kHActiveHighMask  = 1 << kHActiveHighBit,
	kVActiveHighBit = 9,
	kVActiveHighMask  = 1 << kVActiveHighBit,
	kVHActiveHighMask = kVActiveHighMask | kHActiveHighMask,
	
	kCSyncEnableBit = 1,		
	kCSyncEnableMask  = 1 << kCSyncEnableBit,
	kHSyncEnableBit = 2,		
	kHSyncEnableMask  = 1 << kHSyncEnableBit,
	kVSyncEnableBit = 3,		
	kVSyncEnableMask  = 1 << kVSyncEnableBit,
	kMrSanAntonioSyncsMask = kCSyncEnableMask | kHSyncEnableMask | kVSyncEnableMask,

	kCBlankEnableBit = 9,
	kCBlankEnableMask = 1 << kCBlankEnableBit,
	
	kMrSanAntonioSyncsAndCBlankMask = kMrSanAntonioSyncsMask | kCBlankEnableMask,
	
	// interruptMask register
	kVBLInterruptEnableBit = 3,
	kVBLInterruptEnableMask = 1 << kVBLInterruptEnableBit,
	kAnimateInterruptEnableBit = 7,
	kAnimateInterruptEnableMask = 1 << kAnimateInterruptEnableBit,
	kCursorInterruptEnableBit = 9,
	kCursorInterruptEnableMask = 1 << kCursorInterruptEnableBit,

	// interruptStatus register
	kVBLInterruptStatusBit = 3,
	kVBLInterruptStatusMask = 1 << kVBLInterruptStatusBit,
	kAnimateInterruptStatusBit = 7,
	kAnimateInterruptStatusMask = 1 << kAnimateInterruptStatusBit,
	kCursorInterruptStatusBit = 9,
	kCursorInterruptStatusMask = 1 << kCursorInterruptStatusBit
	
};


//
// SpurRegisters
//	General description of how Spur is programmed:  Two address lines going into
//	Spur are looked at to determine which external register is accessed.  Call these
//	address lines A[9:8]
//	There are 4 external registers
//	1) address				A[9:8] = 00
//	2) cursorPalleteRAM		A[9:8] = 01		Pallette for hardware cursor
//	3) multiport			A[9:8] = 10
//							generic name for all of the internal registers that are accessed when
//							A[9:8] = 10
//	4) colorPalletteRAM		A[9:8] = 11		CLUT Pallette
//
//	Accessing the registers in the multiport:
//	1) write the 8 bit address of the internal register to address
//	2) read/write the multiport register.  When Spur sees A[9:8] = 10, it looks at the 
//		contents of the address register and accesses the internal register whose internal address
//		matches the contents of the address register
//
//	Accessing the CLUT:
//	Spur is referred to as a "3x256x8" CLUT for the following reasons:
//	3		The CLUT has three channels: red, green and blue.
//	256		The CLUT has 256 physical addresses (0..FF).
//	8		For each channel at a given address, an 8-bit value can be stored.
//	To access a specfic CLUT entry:
//	1) write the 8 bit physical addresses of the CLUT entry to address.  (The CLUT's internal counter
//		resets to grab the red channel whenever the address register is hit.)
//	2) read/write the colorPalletteRAM register.  When Spur sees A[1:0] = 11, it access the 
//		CLUT entry that matches the contents of the address register.  The first read/write to the
//		colorPalletteRAM register reads/writes the red channel, the 2nd read/write to the 
//		colorPalletteRAM reads/writes the blue channel, the 3rd read/write to the colorPalletteRAM
//		reads/writes the green channel.
//		After 3 successive reads/writes to the colorPalletteRAM register, Spur will
//		autoincrement the address register.  The internal counter resets to grab the red channel.
//		Note:  Each time the address register is changed or auto increments, it is necessary to delay
//		to allow the CLUT to fetch the RGB data.  That is why there are DelayForHardware calls around
//		CLUT accesses.
//		
typedef struct SpurRegisters SpurRegisters;
struct SpurRegisters
{
	HWRegister8Bit *address;
	HWRegister8Bit *cursorPaletteRAM;
	HWRegister8Bit *multiPort;					// access port for multiple registers
	HWRegister8Bit *colorPaletteRAM;
};

// following register addresses are for Spurs's internal registers that are accessed through
// the multiport.
enum
{
	kSpurCursorXPositionHigh = 0x20,
	kSpurCursorXPositionLow = 0x21,
	kSpurControl = 0x33,
	kSpurPhaseLockLoopControl = 0x66,
	kSpurMSetA = 0x43,
	kSpurPNSetA = 0x41,
	kSpurRISetA = 0x47,
	kSpurMSetB = 0x53,
	kSpurPNSetB = 0x51,
	kSpurRISetB = 0x57
};

// following bit posiitions (in 68k ordering) describe the "logical registers"
// in Spur
enum
{
//	kkSpurControl = 0x33,
	kSpurControlCursorEnableBit = 1,
	kSpurControlCursorEnableMask = 1 << kSpurControlCursorEnableBit,

//	kSpurPhaseLockLoopControl = 0x66,
	kSpurPhaseLockLoopControlSetSelectBit = 3,
	kSpurPhaseLockLoopControlSetSelectMask = 1 << kSpurPhaseLockLoopControlSetSelectBit,

	kSpurPhaseLockLoopControlCLKorPLLBit = 1,
	kSpurPhaseLockLoopControlCLKorPLLMask = 1 << kSpurPhaseLockLoopControlCLKorPLLBit,

	kSpurPhaseLockLoopControlCLKorPLLSetMask	= 	kSpurPhaseLockLoopControlCLKorPLLMask |
													kSpurPhaseLockLoopControlSetSelectMask,

	kSpurPhaseLockLoopControlPLLPowerBit = 0,
	kSpurPhaseLockLoopControlPLLPowerMask = 1 << kSpurPhaseLockLoopControlPLLPowerBit,

	kSpurPhaseLockLoopControlDACPowerBit = 2,
	kSpurPhaseLockLoopControlDACPowerMask = 1 << kSpurPhaseLockLoopControlDACPowerBit,

	kSpurPhaseLockLoopControlDACPLLPowerMask	= 	kSpurPhaseLockLoopControlDACPowerMask |
													kSpurPhaseLockLoopControlPLLPowerMask
};


// Default delay for writing the CLUT
enum {
	kDefaultCLUTDelayHigh = 0,
	kDefaultCLUTDelayLow = 800
};



//	An explanation of interrupts for Template:
//	Template contains three interrupt sources: vbl, cursor, and animation.
//	vbl interrupts when enabled, generate an interrupt at every vertical blanking period
//	cusor interrupts when enabled, generate an interrupt when the video line specified in the
//		cursorLine register in mrSanAntonio is drawn.  This is programable
//	animation interrupts when enabled, generate an interrupt when "the line of output video" specified
//		in the animateLine register is drawn.  This is programmable.
//
//	All 3 Template interrupts get funneled to the same bit in the IO controller.  If all 3 Template interrupt
//	sources were enabled, the HAL would need to examine each interrupt source in Template to determine
//	what caused the interrupt.  There is no need to enable all 3 interrupt sources.  The HAL cares
//	about "vbl interrupts" i.e. generate an interrupt when the entire screen has been drawn.  If
//	vbl interrupts are enabled, the interrupt will automatically be generated every time this occurs.
//	However, the cursor interrupts (and probably animation interrupts) can be programmed to go off
//	when the last scan line is drawn.  This occurs slighly before the vbl interrupt would occur and
//	allows more time for "vbl tasks" to run.
//	The HAL uses the cursor interrupt.  The cursorLine register is programmed to be the last scan line
//	for the given displayModeID (height).
//
//	The cursor interrupt is always enabled inside Template.  This allows the driver to determine when 
//	a "vbl" occurs and program the hw at that time.  When interrupts are supposed to be off, the HAL
//	turns of the external interrupts at the controller (using the GraphicOSS functions).
//	This allows the HAL to program hw at the correct time while the system believes interrupts are off.
//




//
// BitDepthIndependentData
//	For each DisplayModeID, certain parameters are constant, regardless of the
//	bit depth.  This structure is used for those parameters.
//
typedef struct BitDepthIndependentData BitDepthIndependentData;
struct BitDepthIndependentData
{	
	UInt32 horizontalSerration;				// Horizontal MrSanAntonio parameters (bpp invariant)
	UInt32 halfLine;						//      "        "        "           "
	UInt32 horizontalEqualization;			//      "        "        "           "
	UInt32 horizontalSyncPulse;				//      "        "        "           "
	UInt32 horizontalBreezeway;				//      "        "        "           "
	UInt32 horizontalBurstGate;				//      "        "        "           "
	UInt32 horizontalBackPorch;				//      "        "        "           "
	UInt32 horizontalActiveLine;			//      "        "        "           "
	UInt32 horizontalFrontPorch;			//      "        "        "           "
	UInt32 horiztonalPixelCount;			//      "        "        "           "


	UInt32 verticalHalfLine;				// Vertical MrSanAntonio parameters (bpp invariant)
	UInt32 verticalSync;					//    "       "       "              "
	UInt32 verticalBackPorchEqualization;	//    "       "       "              "
	UInt32 verticalBackPorch;				//    "       "       "              "
	UInt32 verticalActiveLine;				//    "       "       "              "
	UInt32 verticalFrontPorch;				//    "       "       "              "
	UInt32 verticalFrontPorchEqualization;	//    "       "       "              "

	UInt16 baseAddress;						// FB base Address of VRAM.  hw cursor starts here

	UInt8 frankM;							// PLL inside Spur
	UInt8 frankPN;							// Parameters P = 3 ms bits, N = 5 ls bits
	UInt8 frankRI;							// Parameters R = 4 ms bits, I = 4 ls bits
	
	UInt8 vhSyncPolarity;					// 0x0c if active high, 0 if active low

};



//
// BitDepthDependentData
//	For DisplayModeID, certain parameters change according to the bit depth.
//	This structure is used for those parameters.
//
typedef struct BitDepthDependentData BitDepthDependentData;
struct BitDepthDependentData
{	
	UInt32 rowWords;						// "       "          "
	UInt32 systemConfiguration;				// System Configuration Reg in Iridium, need set depthMode
};



//
// VRAMWidthDependentData
//	For DisplayModeID, certain parameters change according to the VRAM Width AND the bit depth.
//	This structure is used for those parameters.
//
typedef struct VRAMWidthDependentData VRAMWidthDependentData;
struct VRAMWidthDependentData
{	
	UInt32 fbConfiguration1;				// FB controller parameter
};



//
// WidthAndDepthDependentData
//	For DisplayModeID, certain parameters change according to the bit depth.
//	This structure is used for those parameters.
//
typedef struct WidthAndDepthDependentData WidthAndDepthDependentData;
struct WidthAndDepthDependentData
{	
	UInt32 clockConfiguration;				// FB controller parameter
	UInt32 timingAdjust;					// Timging Adjust register = adj1, adj2 and pipeDelay
	UInt8 spurControl;						// CLUT parameter
};


//
// DisplayInfo
//	This structure will describe the capabilities of the frame buffer given the configuration of the
//	system.  
//
typedef struct DisplayInfo DisplayInfo;
struct DisplayInfo
{	
	SInt16 width;
	SInt16 height;
	DepthMode maxDepthMode;			// filled in by the each TemplateGetXXX
	SInt16 rowBytes;					// filled in by the each TemplateGetXXX
};

enum								// index for the maxDepthMode for a displayModeID given available 
{									// VRAM
	k1MegVRAMIndex = 0,				// maxDepthMode when: 1 meg of VRAM
	k2MegVRAMIndex = 1,				// maxDepthMode when: 2 meg of VRAM
	k4MegVRAMIndex = 2				// maxDepthMode when: 4 meg of VRAM
};

enum
{
	kDepthMode1Index = 0,
	kDepthMode2Index = 1,
	kDepthMode3Index = 2
};

enum
{
	vw32Index = 0,
	vw64Index = 1
};


enum
{
	kvw32d8Index = 0,
	kvw32d16Index = 1,
	kvw32d32Index = 2,
	kvw64d8Index = 3,
	kvw64d16Index = 4,
	kvw64d32Index = 5
};

// The TemplateGetXXX functions contain the programming information for all registers based on the
// 1) the amount of VRAM 2) VRAM Width.  When information about a DisplayModeID, is desired
// the caller calls TemplateMapToIndicies which returns the proper index for each type of data based on the
// inputs.  The indices are then passed into the TemplateGetXXX function and the proper data is returned.
typedef struct InfoIndicies InfoIndicies;
struct InfoIndicies
{	
	UInt16 dIndex;
	UInt16 vwIndex;
	UInt16 vwdIndex;
	UInt16 maxDepthIndex;
};


// Template saves the clut when GraphicsHALSetPowerState is told to enter the low power mode kAVPowerOff.
// This stucture is 1/4 the size of a csTable
typedef struct RGB RGB;
struct RGB
{	
	UInt8 red;
	UInt8 green;
	UInt8 blue;
};

//
// HardwareCursorData
//  This structure contains the data needed to maintain a hardware cursor.
//

typedef UInt8 SpurCursorImage[kHardwareCursorImageSize];
typedef UInt32 SpurColorMap[kNumHardwareCursorColors];
typedef UInt32 SpurColorEncodings[kNumHardwareCursorColors];
typedef struct SpurHardwareCursorData SpurHardwareCursorData;
struct SpurHardwareCursorData
{
	HardwareCursorDescriptorRec cursorDescriptor;		// Record describing hardware cursor implementation
	SpurColorEncodings colorEncodings;					// Table of pixel encodings for each color
	SpurCursorImage cursorImage;						// Image data of hardware cursor
	SpurColorMap colorMap;								// Color map for hardware cursor
	SInt32 x, y;										// Current coordinates of cursor used to erase cursor
	Boolean cursorSet;									// Flag indicating that the cursor has been set up
	Boolean cursorRendered;								// Flag indicating that cursor is rendered in cursor buffer
														// Used to determine if erasing is necessary
	Boolean cursorCleared;								// True if HWCursor buffer has been cleared
	Boolean cursorVisible;								// True if cursor is visible on screen
	SInt32 deferredMove;								// True if we're deferring cursor movement to next VBL
														// Must be long and long aligned for CompareAndSwap
	SInt32 deferredX, deferredY;						// Coordinates to move cursor to on deferred move
};

//
// TemplateHALData
//	This structure contains the necessary items for the Template HAL to maintain its
//	state information.
//	
//	This is COMPLETELY private to the Template implementation of the Graphics HAL.
//
typedef struct TemplateHALData TemplateHALData;
struct TemplateHALData
{
	RegEntryID regEntryID;						// save our RegEntryID in case we need anything
	DisplayModeID displayModeID;				// current displayModeID
	Ptr baseAddress;							// current FrameBufferBaseAddress of VRAM reported QD
	UInt32 vramBaseAddress;						// VRAM base address read from Base Register 1
	SpurRegisters spur;							// Addresses of Spur registers
	ToynbeeRegisters toynbee;					// description of Toynbee registers
	MrSanAntonioRegisters mrSanAntonio;			// description of MrSanAntonio registers
	DisplayCode displayCode;					// class of the connected display
	AbsoluteTime nsCLUTAddrRegDelay;			// 800ns in absolute time used for hitting clut
	UInt32 vdPowerState;						// current state of hw...on, off..etc
	RGB savedClut[256];							// save the clut if DAC and PLL are powered down
	DepthMode depthMode;						// current depthMode
	UInt16 cvhSyncDisabled;						// c,v,h Bits if set, sync is disabled
	VRAMSize vramSize;							// Amount of VRAM 1, 2 or 4 megs
	SInt16 width;								// current width for displayModeId
	SInt16 height;								// current height displayModeId
	UInt16 rowBytes;							// current rowbytes for displayModeId and depthMode
	Boolean monoOnly;							// True if attached display only support Monochrome
	Boolean usingClockSetA;						// 'true' if using Set A of the Frankenstien clock
	Boolean vramWidth32;						// 'true' if vram width is 32 bits wide
	Boolean clutOff;							// 'true' if DAC and PLL are powered down
	Boolean toynbeeRunning;						// 'true' mrSanAntonio is on and interrupts are firing
	SpurHardwareCursorData hardwareCursorData;	// Record of data for hardware cursor
};



//
// HALReplacementDriverInfo
//	In the event that the driver is being superseded, the HAL will try and save this information so
//	that the driver's HAL which replaces it can attempt to come up in the same state, if possible.
//	This will help to minimize the visual artifacts that a users sees.
//
typedef struct HALReplacementDriverInfo HALReplacementDriverInfo;
struct HALReplacementDriverInfo
{
	DisplayModeID displayModeID;	// current displayModeID
	Ptr baseAddress;				// current FrameBufferBaseAddress...I hate using ptr
	DisplayCode displayCode;		// class of the connected display
	UInt32 vdPowerState;			// current state of hw...on, off..etc
	DepthMode depthMode;			// current depthMode
	UInt16 cvhSyncDisabled;			// c,v,h Bits if set, sync is disabled
	VRAMSize vramSize;				// Amount of VRAM 1, 2 or 4 megs
	SInt16 width;					// current width for displayModeId
	SInt16 height;					// current height displayModeId
	UInt16 rowBytes;				// current rowbytes for displayModeId and depthMode
	Boolean monoOnly;				// True if attached display only support Monochrome
	Boolean vramWidth32;			// 'true' if vram width is 32 bits wide
	Boolean clutOff;				// 'true' if DAC and PLL are powered down
	Boolean toynbeeRunning;			// 'true' swatch is on and interrupts are firing
};

enum { kDepthModeInvalid = kDepthMode1 -1 };	// invalid depthMode is ALWAYS less than all valid depthModes

TemplateHALData* GraphicsHALGetHALData(void);

// Naming conventions for functions:
//		 GraphicsHALxxx	- functions which all HALs must implement.  These have external scope
//		 {Template}xxx	- functions which are strictly private to a specific HAL



// Sometimes it is necessary to Delay for long periods of times....like when reading sense lines
static void MoreRealDelayFor(Duration duration);


static GDXErr TemplateMapDepthModeToCLUTAttributes(DepthMode depthMode,
				UInt32 *startAddress, UInt32 *entryOffset);

// This routine handles interrupts and is passed back as handler in GraphicsHAL GetVBLInterruptRoutines
static VBLHandler TemplateClearInternalVBLInterrupts;

static void TemplateWaitForVBL(void);
static GDXErr TemplateAssertVideoReset(const ToynbeeRegisters *toynbee);

static GDXErr TemplateSetupClockGenerator(const SpurRegisters *spur, const BitDepthIndependentData *bdiData,
		Boolean *usingClockSetA);

static GDXErr TemplateSetupCLUT(const SpurRegisters *spur, const WidthAndDepthDependentData *vwdData);

static GDXErr TemplateSetupFBController(Boolean resolutionChange, const ToynbeeRegisters *toynbee,
		const MrSanAntonioRegisters *mrSanAntonio, const BitDepthIndependentData *bdiData,
		const BitDepthDependentData *bddData, const VRAMWidthDependentData *vwData,
		const WidthAndDepthDependentData *vwdData);
		
static GDXErr TemplateReleaseVideoReset(const ToynbeeRegisters *toynbee);

static ExtendedSenseCode TemplateGetExtendedSenseCode(void);
static void TemplateResetSenseLines(void);
static void TemplateDriveSenseLines(SenseLine senseLine);
static UInt32 TemplateReadSenseLines(void);

GDXErr TemplateGetDisplayData(DisplayModeID displayModeID, DepthMode depthMode, VRAMSize vramSize,
	Boolean vramWidth32, BitDepthIndependentData *bdiData, BitDepthDependentData *bddData,
	VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, DisplayInfo *info);

static GDXErr TemplateMapToIndices(DepthMode depthMode, VRAMSize vramSize, Boolean vramWidth32,
	InfoIndicies *infoIndicies);

static void DeferredMoveHardwareCursor(void);

// Here is typedef for a function used to retrieve BitDepthIndependentData & BitDepthDependentData
typedef GDXErr GetBDIAndBDDDataFunction(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info);

// Forward declarations of various instantiations of the GetBDIAndBDDDataFunction type

static GetBDIAndBDDDataFunction TemplateGet512x384At60HzNTSC;
static GetBDIAndBDDDataFunction TemplateGet512x384At60Hz;
static GetBDIAndBDDDataFunction TemplateGet640x480At50HzPAL;
static GetBDIAndBDDDataFunction TemplateGet640x480At60HzNTSC;
static GetBDIAndBDDDataFunction TemplateGet640x480At60HzVGA;
static GetBDIAndBDDDataFunction TemplateGet640x480At67Hz;
static GetBDIAndBDDDataFunction TemplateGet640x870At75Hz;
static GetBDIAndBDDDataFunction TemplateGet768x576At50HzPAL;
static GetBDIAndBDDDataFunction TemplateGet800x600At60HzVGA;
static GetBDIAndBDDDataFunction TemplateGet800x600At72HzVGA;
static GetBDIAndBDDDataFunction TemplateGet800x600At75HzVGA;
static GetBDIAndBDDDataFunction TemplateGet832x624At75Hz;
static GetBDIAndBDDDataFunction TemplateGet1024x768At60HzVGA;
static GetBDIAndBDDDataFunction TemplateGet1024x768At72HzVGA;
static GetBDIAndBDDDataFunction TemplateGet1024x768At75HzVGA;
static GetBDIAndBDDDataFunction TemplateGet1024x768At75Hz;
static GetBDIAndBDDDataFunction TemplateGet1152x870At75Hz;
static GetBDIAndBDDDataFunction TemplateGet1280x960At75Hz;
static GetBDIAndBDDDataFunction TemplateGet1280x1024At75Hz;



// HAL global, persistant data is stored in gTemplateHALData
TemplateHALData	gTemplateHALData;	// Persistant Template specific data storage


// The version of Template Graphics driver
enum
{
	kmajorRev = 1, 
	kminorAndBugRev = 0,
	kstage = developStage,				//developStage, alphaStage, betaStage, finalStage
	knonRelRev = 0

};

// TheDriverDescription
// 	This structure describes the native driver.  Since This is a Graphics driver, the driver should
//	be loaded by the DisplayMgr's expert loader.  This is indicated by the driverOSRuntimeInfo
//	with the kdriverIsUnderExpertControl flag set.
DriverDescription TheDriverDescription =
{
	// driverDescSignature
		kTheDescriptionSignature,						// signature of DriverDescription

	// driverDescVersion
		kInitialDriverDescriptor,						// Version of this data structure

	// driverType
	{
		"\ptemplate",										// device name must match in Devicetree
		kmajorRev, kminorAndBugRev, kstage, knonRelRev,		// Major, Minor, Stage, Rev
	},

	
	// driverOSRuntimeInfo								// OS Runtime Requirements of Driver
		{
			kDriverIsOpenedUponLoad + kDriverIsUnderExpertControl,	// Runtime Options
			"\p.Display_Video_Template",
		},

		
	// driverServices									// Apple Service API Membership

		// nServices
		1,												// Number of Services Supported						

		// serviceCategory								// Service Category Name
		kServiceCategoryNdrvDriver,						// We support the 'ndrv' category

		// serviceType									// Type within Category
		kNdrvTypeIsVideo,								// Video type

		// serviceVersion								// Version of service 
		1,0,0,0											// majorRev, minorAndBugRev, stage, nonRelRev

};



//=====================================================================================================
//
// GraphicsHALGetHALData()
//	Return the pointer to the global Core data.  Yes...you guessed it...it just returning
//	the address of the global, but use the accessor function anyway to isolate yourself
//	in the event CFM coolness is ever lost.
//
//=====================================================================================================
TemplateHALData  *GraphicsHALGetHALData(void)
{
	return (&gTemplateHALData);
}



//=====================================================================================================
//
// GraphicsHALInitPrivateData()
//	Allocate and initialize the HAL private data.
//
//		-> regEntryID	The NameRegistry ID for this device.
//
//		<> replacingDriver
//		On input, this indicates whether the Core got a 'kInitializeCommand' or a 'kReplaceCommand'.
//		These commands are similar, but with subtle differences.  A 'kInitializeCommand' is issued
//		if no version of this driver has been previously loaded, whereas a 'kReplaceCommand' is
//		issued if a previous version of the driver has been loaded, but subsequently superseded.
//
//		If 'false', then a 'kInitializeCommand' had been recieved by the Core, and the HAL should
//		do a full hardware initialization.
//
//		If 'true', then a 'kReplaceCommand' had been received by the Core, and the HAL can attempt
//		to configure itself to its state prior to it being superceded.
//
//		On output, this allows the HAL to override the Core's default behavior if it chooses to do so.
//
//		If 'false' then the HAL is signinaling the Core that it is unable to re-configure itself to
//		its state prior to being superseded, and the core will continue as if a 'kInitializeCommand'
//		had occurred.
//		
//		If 'true', then the HAL was able to re-configure itself in the event of being replaced,
//		and the Core will proceeded accordingly.
//
//=====================================================================================================
GDXErr GraphicsHALInitPrivateData(const RegEntryID *regEntryID, Boolean *replacingDriver)
{

	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	HardwareCursorDescriptorPtr cursorDescriptor;
	UInt32 i;
	UInt32 halPreferences;								// look at HAL data in nvram
	Nanoseconds nanoseconds;							// convert 800ns into Absolute Time
	PCIAssignedAddress assignedAddresses[2];			// there should be two "phys-addr size" pairs 
	UInt32 applAddress[2];								// there should be two logical addresses
	UInt32 hwBaseAddress;								// get hwBaseAddress from Base Register 0
	UInt32 baseRegister0Index;							// which entry in APPL,address is for Base Register 0
	UInt32 baseRegister1Index;							// which entry in APPL,address is for Base Register 1
	UInt16 commandRegister;								// enable Template's memory space if everything ok
	OSErr osErr;
	GDXErr err = kGDXErrUnableToAllocateHALData;		// assume failure
	
	// save the regEntryID,
	RegistryEntryIDCopy(regEntryID, &templateHALData->regEntryID);

	// Template is on a PCI card.  Template contains the Toynbee framebuffer, the Spur CLUT,
	// and the MrSanAntonio timing generator. In this sample code, it is assumed that the
	// Template PCI card uses 2 Base Registers.  This illustrates how the "assigned-address" property
	// must be examined to determine which "AAPL,address" 32 bit entry corresponds with each Base Register.
	// Base Register 0 (offset at 0x10 from the configuration registers) is the base address for the hardware.  
	// Base Register 1 (offset at 0x14 from the configuration registers) is the base address for the VRAM.  
	// Template's Open Firmware startup code created a "reg" property that described
	// 1) the configuration registers
	// 2) Base Register 0
	// 3) Base Register 1
	// Open Firmware creates an "assigned-address" property that describes Base Register 0 and Base Register 1

	err = GraphicsOSSGetProperty(&templateHALData->regEntryID, kPCIAssignedAddressProperty, assignedAddresses,
		sizeof(PCIAssignedAddress)*2);		// times 2 since there should be 2 assigned-addresses

	// if there is an error, that means Open Firmware was unable to allocate the memory space that was
	// requested.  Hence, Template is unusable at this time and the driver should quit.
	if (err)
		goto ErrorExit;
		
	// Since the entries of the "assigned-address" property does not necessarily match the order of the
	// entries in the "reg" property, examine the rrrrrrrr bits in each physHi to determine which 
	// "phys-addr size" pair matches each Base Register.
	// Base Register 0 (offset at 0x10 from the configuration registers) has rrrrrrrr = 0x10 etc...
	if ( 0x10 == assignedAddresses[0].registerNumber )
	{
		baseRegister0Index = 0;
		baseRegister1Index = 1;
	}
	else
	{
		baseRegister0Index = 1;
		baseRegister1Index = 0;
	}
		
	// The order of the "phys-addr size" pairs in the "assigned-address" property match the order
	// of the logical addresses in the "AAPL,address" property.  Go find the logical addresses.
	err = GraphicsOSSGetProperty(&templateHALData->regEntryID, "AAPL,address", applAddress, 8);
	
	if (err)							// should NEVER be an error if gotten this far
		goto ErrorExit;					
	
	hwBaseAddress = applAddress[baseRegister0Index];
	templateHALData->vramBaseAddress = applAddress[baseRegister1Index];
	
	// have successfully found all base addresses.
	// Enable Template's memory space.  Always do a read-modify-write when hitting configuration registers
	osErr = ExpMgrConfigReadWord(&templateHALData->regEntryID, (LogicalAddress) 0x4, &commandRegister);

	if (osErr)							// shouldn't be an error
		goto ErrorExit;					

	commandRegister|= 1 << 1;			// enable the Memory space which is Bit 1

	osErr = ExpMgrConfigWriteWord(&templateHALData->regEntryID, (LogicalAddress) 0x4, commandRegister);
	if (err)							// shouldn't be an error
		goto ErrorExit;					

	
	// Note: More examples
	// 1) a PCI card only uses 1 BaseRegister to access hardware and VRAM.  If the card's
	// Open Firmware startup code created a "reg" property for just the 1 Base Register that it uses,
	// the "AAPL,address" property will contain the single logical address.  So if the "AAPL,address" property
	// exists and is the right size (4), then the PCI card knows Open Firmware was able to allocate the
	// requested memory.  Such a card would only have to do the following to find the logical address.
	//
	//	err = GraphicsOSSGetProperty(&templateHALData->regEntryID, "AAPL,address", &hwBaseAddress, 4);
	//	if (err)							// it is possible to error out if Open Firmware couldn't 
	//		goto ErrorExit;					// allocate the requested memory space.
	//
	// 2) a PCI card that doesn't have Open Firmware startup code (BECAUSE THE CARD IS IN DEVELOPMENT)
	// will have a "reg" property automatically created for it by Open Firmware. 
	// This is described in chapter 4 of Designing PCI Cards&Drivers.
	// The PCI Graphics card should implement the Expansion ROM Base Register (offset 0x30 from the configuration
	// registers)  because it will eventually be a startup device. Open Firmware will automatically add
	// a "phys-addr size" pair to the "assigned-address" property for the Expansion ROM Base Register.
	// Likewise, an additional entry will be created in the "AAPL,address" property.
	// The driver, during development, needs to allocate additional space in the buffers used to fetch each
	// property.  The entries of the "assigned-address" property need to be exaimined to determine the order
	// as shown above.
	// (All PCI graphics cards should be a startup device and need the Expansion ROM Base Register.  YOU ARE
	// LAME IF YOU DON'T HAVE ONE)
	
	// fill in the addresses for the toynbee framebuffer. 
	templateHALData->toynbee.baseAddress = (HWRegister32Bit*) (hwBaseAddress | 0x140);
	templateHALData->toynbee.rowWords = (HWRegister32Bit*) (hwBaseAddress | 0x150);
	templateHALData->toynbee.clockConfiguration = (HWRegister32Bit*) (hwBaseAddress | 0x130);
	templateHALData->toynbee.frameBufferConfiguration1 = (HWRegister32Bit*) (hwBaseAddress | 0x170);
	templateHALData->toynbee.frameBufferConfiguration2 = (HWRegister32Bit*) (hwBaseAddress | 0x190);
	templateHALData->toynbee.senseLineEnable = (HWRegister32Bit*) (hwBaseAddress | 0x110);
	templateHALData->toynbee.reset = (HWRegister32Bit*) (hwBaseAddress | 0x1a0);
	
	// fill in the addresses for mrSanAntonio
	templateHALData->mrSanAntonio.mrSanAntonioMode = (HWRegister32Bit*) (hwBaseAddress | 0x340);
	templateHALData->mrSanAntonio.interruptMask = (HWRegister32Bit*) (hwBaseAddress | 0x350);
	templateHALData->mrSanAntonio.interruptStatus = (HWRegister32Bit*) (hwBaseAddress | 0x330);
	templateHALData->mrSanAntonio.clearCursorInterrupt = (HWRegister32Bit*) (hwBaseAddress | 0x320);
	templateHALData->mrSanAntonio.clearAnimationInterrupt = (HWRegister32Bit*) (hwBaseAddress | 0x310);
	templateHALData->mrSanAntonio.clearVBLInterrupt = (HWRegister32Bit*) (hwBaseAddress | 0x300);
	templateHALData->mrSanAntonio.cursorLine = (HWRegister32Bit*) (hwBaseAddress | 0x390);
	templateHALData->mrSanAntonio.animateLine = (HWRegister32Bit*) (hwBaseAddress | 0x3a0);
	templateHALData->mrSanAntonio.counterTest = (HWRegister32Bit*) (hwBaseAddress | 0x3e0);
	templateHALData->mrSanAntonio.horizontalSerration = (HWRegister32Bit*) (hwBaseAddress | 0x3f0);
	templateHALData->mrSanAntonio.halfLine = (HWRegister32Bit*) (hwBaseAddress | 0x400);
	templateHALData->mrSanAntonio.horizontalEqualization = (HWRegister32Bit*) (hwBaseAddress | 0x440);
	templateHALData->mrSanAntonio.horizontalSyncPulse = (HWRegister32Bit*) (hwBaseAddress | 0x430);
	templateHALData->mrSanAntonio.horizontalBreezeway = (HWRegister32Bit*) (hwBaseAddress | 0x410);
	templateHALData->mrSanAntonio.horizontalBurstGate = (HWRegister32Bit*) (hwBaseAddress | 0x4e0);
	templateHALData->mrSanAntonio.horizontalBackPorch = (HWRegister32Bit*) (hwBaseAddress | 0x4f0);
	templateHALData->mrSanAntonio.horizontalActiveLine = (HWRegister32Bit*) (hwBaseAddress | 0x460);
	templateHALData->mrSanAntonio.horizontalFrontPorch = (HWRegister32Bit*) (hwBaseAddress | 0x3e0);
	templateHALData->mrSanAntonio.horiztonalPixelCount = (HWRegister32Bit*) (hwBaseAddress | 0x3f0);
	templateHALData->mrSanAntonio.verticalHalfLine = (HWRegister32Bit*) (hwBaseAddress | 0x530);
	templateHALData->mrSanAntonio.verticalSync = (HWRegister32Bit*) (hwBaseAddress | 0x520);
	templateHALData->mrSanAntonio.verticalBackPorchEqualization = (HWRegister32Bit*) (hwBaseAddress | 0x570);
	templateHALData->mrSanAntonio.verticalBackPorch = (HWRegister32Bit*) (hwBaseAddress | 0x580);
	templateHALData->mrSanAntonio.verticalActiveLine = (HWRegister32Bit*) (hwBaseAddress | 0x5e0);
	templateHALData->mrSanAntonio.verticalFrontPorch = (HWRegister32Bit*) (hwBaseAddress | 0x5d0);
	templateHALData->mrSanAntonio.verticalFrontPorchEqualization = (HWRegister32Bit*) (hwBaseAddress | 0x540);
	templateHALData->mrSanAntonio.timingAdjust = (HWRegister32Bit*) (hwBaseAddress | 0x600);
	templateHALData->mrSanAntonio.currentLine = (HWRegister32Bit*) (hwBaseAddress | 0x630);

	
	// spur is the CLUT used in Template.
	// 1) 32 or 64 bit pixel port for spur.
	// 2) spur has a programmable PLL clock generator
	
	templateHALData->spur.address = (HWRegister8Bit*) (hwBaseAddress | 0x9900);
	templateHALData->spur.cursorPaletteRAM = (HWRegister8Bit*) (hwBaseAddress | 0x9910);
	templateHALData->spur.multiPort = (HWRegister8Bit*) (hwBaseAddress | 0x9920);
	templateHALData->spur.colorPaletteRAM = (HWRegister8Bit*) (hwBaseAddress | 0x9930);


	if (*replacingDriver)
	{
		// Important Implementation Note:  The exact behavior of what should happen during driver
		// replacement is largely dependent of the implementation of the driver, and the reason for
		// it being replaced.
		// Under the ideal circumstances, a driver should be able to be replaced without having to
		// shut down the raster or reprogram the hardware.  This will prevent visible flashes on the
		// display during the replacement process.
		// 
		// When a driver is being updated to incorporate new features (as opposed to bug fixes), then
		// replacement can usually occur without flashes.  This is because by default, the
		// Driver Loader Library (DLL) does not let QuickDraw know that a new driver got loaded, so
		// QuickDraw doesn't issue commands to gray the screen, etc. which result in 'flashes.'
		//
		// However, this default behavior is only acceptable if the new driver comes up in the same
		// bit depth and has the same 'base address' and 'rowbytes' as the driver it replaced.  If 
		// this is not the case, then the DLL needs to be informed that a full initialization of all
		// the QuickDraw variables is required.  The DLL can be informed by setting uncommenting
		// the following lines:
		// 
		// UInt32 needFullInit  = 1;
		// err = GraphicsOSSSaveProperty(&sixty6HALData->regEntryID, "needFullInit",
		//			&needFullInit, sizeof(needFullInit), kOSSPropertyVolatile);

		HALReplacementDriverInfo replacementDriverInfo;
		
		err = GraphicsOSSGetProperty(&templateHALData->regEntryID, "HALReplacementInfo",
			&replacementDriverInfo, sizeof(HALReplacementDriverInfo));

		if (!err)
		{
			
			templateHALData->displayModeID = replacementDriverInfo.displayModeID;
			templateHALData->baseAddress = replacementDriverInfo.baseAddress;
			templateHALData->vdPowerState = replacementDriverInfo.vdPowerState;
			templateHALData->depthMode = replacementDriverInfo.depthMode;
			templateHALData->cvhSyncDisabled = replacementDriverInfo.cvhSyncDisabled;
			templateHALData->vramSize = replacementDriverInfo.vramSize;
			templateHALData->width = replacementDriverInfo.width;
			templateHALData->height = replacementDriverInfo.height;
			templateHALData->rowBytes = replacementDriverInfo.rowBytes;
			templateHALData->monoOnly = replacementDriverInfo.monoOnly;
			templateHALData->vramWidth32 = replacementDriverInfo.vramWidth32;
			templateHALData->clutOff = replacementDriverInfo.clutOff;
			templateHALData->toynbeeRunning = replacementDriverInfo.toynbeeRunning;
			
			// Have now grabbed all the old state information.  The new driver knows what bugs the old
			// driver had.  Examine the state information to see if it is necessary to do a full hw 
			// initialization.  For example, one displayModeID was known to be bad...the hw was programmed
			// incorrectly for that resolution.
			// if ( brokenDisplayModeID == templateHALData->displayModeID)
			//		*replacingDriver = false;
			
			// The Template PCI card has power saving features, the clut can be turned off, VRAM refresh
			// can be turned off.  While it is supremely unlikely that a driver would be replaced while it
			// was in a low power state, do a sanity check to make sure the vdPowerState == kAVPowerOn.
			// Force a full initialization if in a low power state
			// if ( kAVPowerOn != templateHALData->vdPowerState)
			//		*replacingDriver = false;		// report that a Full initialization is necessary

		}
		else
		{
			*replacingDriver = false;					// report that a Full initialization is necessary
			err = kGDXErrNoError;
		}
		
	
	}	// end if replacingDriver = true
	
	// always try to delete the HALReplacementDriverInfo in the nameRegistry
	(void) GraphicsOSSDeleteProperty(&templateHALData->regEntryID, "HALReplacementInfo");

	
	if (!*replacingDriver)
	{
		// If the driver is replacing an old driver, the raster was left on.  Don't really need to hit
		// hw since the old driver left all the necessary state information as shown above.
	
	
		// Template supports 3 VRAM configurations:
		// 1) 1 MB VRAM arranged as a 32-bit wide half bank (this is the base configuration on Template)
		// 2) 2 MB VRAM arranged as a 64-bit wide full bank
		// 4) 4 MB VRAM arranged as two 64-bit wide full banks
	
		// size your VRAM here....left as an exercise for the reader....odd numbered answers are in appendix C
	}

	// initialize hardwareCursorData
	cursorDescriptor = &(templateHALData->hardwareCursorData.cursorDescriptor);
	cursorDescriptor->majorVersion = 0;
	cursorDescriptor->minorVersion = 0;
	cursorDescriptor->height = 32;
	cursorDescriptor->width = 32;
	cursorDescriptor->bitDepth = 4;
	cursorDescriptor->maskBitDepth = 0;
	cursorDescriptor->numColors = kNumHardwareCursorColors;
	cursorDescriptor->colorEncodings = (UInt32 *) &(templateHALData->hardwareCursorData.colorEncodings);
	for (i = 0; i < kNumHardwareCursorColors; i++)
		cursorDescriptor->colorEncodings[i] = 8 + i;
	cursorDescriptor->flags = 0;
	cursorDescriptor->supportedSpecialEncodings =
		kTransparentEncodedPixel | kInvertingEncodedPixel;
	cursorDescriptor->specialEncodings[kTransparentEncoding] = 0;
	cursorDescriptor->specialEncodings[kInvertingEncoding] = 1;


	// Spur clut must sometimes wait for stuff to happen, e.g. when writing rgb to the clut,
	// it is necessary to wait 800 ns before writing the next rgb.  Since this is done many times,
	// convert 800 ns to absolute time and save in the hal data.
	nanoseconds.hi = kDefaultCLUTDelayHigh;
	nanoseconds.lo = kDefaultCLUTDelayLow;
	templateHALData->nsCLUTAddrRegDelay = NanosecondsToAbsolute(nanoseconds);


	// Template might want to look at the HAL specfic data stored in NVRAM.
	// The last 4 bytes (each device is allowed 8 bytes of nvram) is private to each HAL.
	// GraphicsOSSGetHALPref automatically fetches the last 4 bytes for the HAL
	err = GraphicsOSSGetHALPref(&templateHALData->regEntryID, &halPreferences);
	if (!err)
	{
		// examine the haldata stored in nvram.  The HAL portion of the nvram is private to the HAL
		// (the core never examines or messes with it) and can have any meaning defined by the HAL.
		// If the HAL wishes to save data, it can use the routine GraphicsOSSSetHALPref
		// NOTE: the HAL should do a read-modify-write if it wishes to change the data.  This allows disk 
		// based drivers that replace the ROM based drivers to define new HAL data.  Ensures that the ROM 
		// based driver (always run at startup before the disk based driver) doesn't smash the data that the
		// disk based driver saved.
	}

	
	
	err = kGDXErrNoError;								// everything ok

ErrorExit:

	return err;
	
}



//=====================================================================================================
//
// GraphicsHALKillPrivateData()
//
//	Disposes of the HAL's private data
//
//=====================================================================================================
void GraphicsHALKillPrivateData()
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();

	RegistryEntryIDDispose(&templateHALData->regEntryID);

}


//=====================================================================================================
//
// GraphicsHALOpen()
//	It is possible for the driver to be opened and closed many times.
//	The HAL private data has been setup, so just initialize the hardware into
//	the state that is should be in on startup.  This should be the complete initialization
//	to get the HW into the desired state for the amount of VRAM that is in the system and
//	any other hardware specfic stuff that the HAL cares about. The sense lines, for example, should
//	be able to be read and tweaked to determine the type of the connected monitor.
//	No programming to set up the raster for a givin 'DisplayModeID' or 'DepthMode' is necessary at
//	this point.
//
//		-> spaceID			The AddressSpaceID for this device.
//		-> replacingDriver	'true' if the HAL should behave as if the driver is being replaced,
//							'false' otherwise.
//
//=====================================================================================================
GDXErr GraphicsHALOpen(const AddressSpaceID spaceID, Boolean replacingDriver)
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	MrSanAntonioRegisters *mrSanAntonio = &templateHALData->mrSanAntonio;
	ToynbeeRegisters *toynbee = &templateHALData->toynbee;
	SpurRegisters *spur = &templateHALData->spur;
	UInt8 control;
	GDXErr err = kGDXErrNoError;
	
	if (!replacingDriver)
	{
		// a bogus displayModeID so that MrSanAntonio will always be hit on first call to ProgramHardware
		templateHALData->displayModeID = kDisplayModeIDInvalid;
		err = TemplateAssertVideoReset(toynbee);
		*mrSanAntonio->mrSanAntonioMode = 0x0CC0;		// default state
	}

	// internal interrupts in Template should be on so that HAL programs hw during vbl
	*mrSanAntonio->interruptMask = kCursorInterruptEnableMask;	// cursor interrupts on, vbl & animate off


	// turn off the hw cursor.  Other control routines will turn it on as necessary
	*spur->address = kSpurControl;			// Setup the address register for 'control'
	SynchronizeIO();

	control = *spur->multiPort;
	control &= ~kSpurControlCursorEnableMask;	// turn off hw cursor
	
	*spur->multiPort = control;

	// it is possible that the driver is replacing a ROM driver and the machine was
	// in low power mode, in which case the clut would be off.  Turn on the DAC and PLL
	
	*templateHALData->spur.address = kSpurPhaseLockLoopControl;	// DAC and PLL address	
	SynchronizeIO();

	*templateHALData->spur.multiPort = 1 << kSpurPhaseLockLoopControlCLKorPLLBit;

	//	The phaseLockLoopControl register determines which of the two frequency register sets the PLL
	//	is actively using.  Bit kSpurPhaseLockLoopControlSetSelectBit selects bewteen set A and B.
	// 0 = set A, 1 = set B.  On reset, set A is selected as the active set.
	//	Since the non active set is always used to reprogram the clock, mark that Set A is active.
		
	templateHALData->usingClockSetA = true;

	// when opened, hw is up and running....vdPowerState is always kAVPowerOn as shown in Initialize
	templateHALData->vdPowerState = kAVPowerOn;

	// Reset cursor state.
	templateHALData->hardwareCursorData.cursorSet = false;
	templateHALData->hardwareCursorData.cursorRendered = false;
	templateHALData->hardwareCursorData.cursorCleared = false;
	templateHALData->hardwareCursorData.cursorVisible = false;

	return (kGDXErrNoError);


ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALClose()
//	Upon close, there are no major requirements, since the majority of the work will be handled
//	elsewhere.
//	
//=====================================================================================================
GDXErr GraphicsHALClose(const AddressSpaceID spaceID)
{

	return kGDXErrNoError;
}



//=====================================================================================================
//
// GraphicsHALTerminate()
//
//		-> superseded
//		'true' if current driver is going to be superseded by another driver, 'false' otherwise.
//		If 'true', the current driver can choose to save any state that the replacement driver
//		may need..if it wants to keep raster going.
//
//		'false' no driver is going to replace it.  In that event, it should stop the raster and
//		leave hardware in a 'polite' state.
//
//=====================================================================================================
GDXErr GraphicsHALTerminate(Boolean superseded)
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	ToynbeeRegisters *toynbee = &templateHALData->toynbee;

	GDXErr err = kGDXErrNoError;
	
	if (superseded)
	{
		// Driver is being superseded, so save state information for the replacement driver.

		HALReplacementDriverInfo replacementDriverInfo;

		replacementDriverInfo.displayModeID = templateHALData->displayModeID;
		replacementDriverInfo.baseAddress = templateHALData->baseAddress;
		replacementDriverInfo.vdPowerState = templateHALData->vdPowerState;
		replacementDriverInfo.depthMode = templateHALData->depthMode;
		replacementDriverInfo.cvhSyncDisabled = templateHALData->cvhSyncDisabled;
		replacementDriverInfo.vramSize = templateHALData->vramSize;
		replacementDriverInfo.width = templateHALData->width;
		replacementDriverInfo.height = templateHALData->height;
		replacementDriverInfo.rowBytes = templateHALData->rowBytes;
		replacementDriverInfo.monoOnly = templateHALData->monoOnly;
		replacementDriverInfo.vramWidth32 = templateHALData->vramWidth32;
		replacementDriverInfo.clutOff = templateHALData->clutOff;
		replacementDriverInfo.toynbeeRunning = templateHALData->toynbeeRunning;

		err = GraphicsOSSSaveProperty(&templateHALData->regEntryID, "HALReplacementInfo",
				&replacementDriverInfo, sizeof(HALReplacementDriverInfo), kOSSPropertyVolatile);

		if (err)
			err = TemplateAssertVideoReset(toynbee);
	}
	else
	{
		// Driver is going away for good, so blank turn off state machines.
		err = TemplateAssertVideoReset(toynbee);
	}
	
	return err;
}



//=====================================================================================================
//
// GraphicsHALGetVBLInterruptRoutines()
//	The OSS encapsulates how interrupts are handled by the system.  This routine supplies
//	that OSS with the HAL's interrupt routines that follow the OSS conventions.  Hopefully,
//	if the OS changes, only the OSS will need to change.
//
//		<- installVBLInterrupts	
//		'true' 	if the HAL's interrupt scheme can match the OSS's scheme. i.e. the HAL lets the OSS
//		handle most of the interrupt functions.
//		'false'	if the HAL's interrupt scheme is radically different than the OSS's scheme.  The
//		HAL is responsible for knowing how the OS handles interrupts.  Obviously, this is the 
//		escape mechanism for a poor OSS design.  The HAL needs a radically different interrupt handling
//		design so the OSS will not handle any interrupt services.  If this is false, all other
//		paramters are ignored
//
//		<- chainDefault
//		If 'halVBLEnabler' or 'halVBLDisabler' = NULL, this is ignored by the OSS for the respective
//		function since the default enabler/disabler supplied by the OS is used.
//		If 'chainDefault = true', if the halVBLEnabler or halVBLDisabler != NULL, the OSS will call
//		the default OS enabler/disbler after the HAL's enabler/disabler is called
//		If 'chainDefault = false', if the halVBLEnabler or halVBLDisabler != NULL, the OSS will NOT call
//		the default OS enabler/disbler after the HAL's enabler/disabler is called.  The HAL assumes
//		the responsibility for enabling/disabling the interrupt source.  (Dangerous!)
//
//		<- halVBLHandler
//		The HAL's VBL handler which should clear the internal interrupt source.
//
//		<- halVBLEnabler
//		If 'halVBLEnabler = NULL', the default OS enabler will be called and the HAL can ignore things.
//		If 'halVBLEnabler != NULL' and 'hainDefault = true', the HAL needs to enable the internal
//		interrupt source and the OSS calls the default OS enable routine to enable external interrupts.
//		If 'halVBLEnabler != NULL' and 'chainDefault = false', the HAL needs to enable the internal
//		and external interrupt source.  (Dangerous!)
//
//		<- halVBLDisabler
//		If 'halVBLDisabler = NULL', the default OS enabler will be called and the HAL can ignore things.
//		If 'halVBLDisabler != NULL' and 'chainDefault = true', the HAL can choose to disable the internal
//		interrupt source and the OSS calls the default OS disable routine to disable external interrupts.
//		If 'halVBLDisabler != NULL' and 'chainDefault = false', the HAL can choose to disable the internal
//		and must disable the external interrupt source.  (Dangerous!)
//
//		<- vblRefCon
//		If the HAL needs some data for the interrupt routines, allocate some structure that
//		is pointed to by vblRefCon.
//
//
//	Template, always has the internal interrupt source enabled so it knows when vbls occur.  Hence,
//	halVBLEnabler and halVBLDisabler = nil.  The default OS enable/disable routines are used to
//	control what interrupts the OS sees.
//
//=====================================================================================================
GDXErr GraphicsHALGetVBLInterruptRoutines(Boolean *installVBLInterrupts, Boolean *chainDefault,
		VBLHandler **halVBLHandler, VBLEnabler **halVBLEnabler, VBLDisabler **halVBLDisabler,
		void **vblRefCon)
{
	// Template is always going to have its internal interrupt source enabled for VBLs.  Hence,
	// 'halVBLEnabler' and 'halVBLDisabler' are NULL.  The internal interrupt will be 
	// allowed/stopped to propagate by opening/closing the external interrupt 'gateway' via the
	// default OS enabler/disabler.
	
	*installVBLInterrupts = true;		// This HAL supports the OSS's interrupt scheme.
	*chainDefault = false;				// Ignored by OSS since HAL's enabler/disablers are NULL
	*halVBLHandler = TemplateClearInternalVBLInterrupts; 
	*halVBLEnabler = NULL;
	*halVBLDisabler = NULL; 
	*vblRefCon = NULL;					// No private refCon needed.

	return kGDXErrNoError;
}



//=====================================================================================================
//
// GraphicsHALGrayCLUT()
//
//	This routine sets all the CLUT entries to 50% gray (with gamma correction).
//	This is useful so that the pixel depth can be subsequently changed without 
//	introducing screen anonmalies.
//	The 50% gray value will be obtained by using the midpoint value of the supplied
//	gamma table.
//
//	NOTE:  this assumes that the gamma correction data size is 1 byte,
//	as stated in the core.
//
//=====================================================================================================
GDXErr GraphicsHALGrayCLUT(const GammaTbl *gamma)
{
	
	// gray all 256 entries...it doesn't matter what depthMode we are in
	enum { kClutSize = 256 };

	GDXErr err = kGDXErrUnknownError;					// Assume failure

	UInt8 *midPointRed;									// Midpoint of red correction data
	UInt8 *midPointGreen;								//     "    "  green    "       "
	UInt8 *midPointBlue;								//     "    "  blue     "       "
	
	SInt16 channelCount = gamma->gChanCnt;	
	SInt16 entriesPerChannel = gamma->gDataCnt;
	
	SpurRegisters *spur;

	Boolean vblInterruptsEnabled = false;				// default to not enabled;

	UInt32 i;											// utterly famous iterator


	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	AbsoluteTime nsCLUTAddrRegDelay = templateHALData->nsCLUTAddrRegDelay;				// 800ns in absolute time to hit clut

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	spur = &templateHALData->spur;
	
	// Get the midpoint of the red correction data.  This is found by starting at the begining of the
	// correction data, which can be found at '&gFormulaData[0]', adding the 'gFormulaSize'.
	// Then go halfway into the table as determined by 'entriesPerChannel / 2'
	
	midPointRed = (UInt8 *) ((UInt32) &gamma->gFormulaData[0] + gamma->gFormulaSize +
			(entriesPerChannel / 2));

	// If there is only 1 channel of correction data, it means the same correction is applied to the
	// red, green, and blue channels.  If there are 3 channels then each color has its own correction
	// data.
	
	if (1 == channelCount)
	{
		midPointGreen = midPointRed;
		midPointBlue = midPointRed;
	} else {
		midPointGreen = midPointRed + entriesPerChannel;
		midPointBlue = midPointRed + (entriesPerChannel * 2);
	}


	if (templateHALData->clutOff)				// save grayed clut to virtual clut
	{
		for (i = 0 ; i < kClutSize ; i++)
		{
			templateHALData->savedClut[i].red = *midPointRed;		// Red
			templateHALData->savedClut[i].green =*midPointGreen;	// Green
			templateHALData->savedClut[i].blue = *midPointBlue;		// Blue
		}
	
	}
	else
	{
	// Turn off VBL interrupts so we can adjust the CLUT during VBL time, without having to worry
	// about the VBL interrupt handler stealing time from us.
		
		vblInterruptsEnabled = GraphicsOSSVBLDefaultDisabler();						
		TemplateWaitForVBL();
		
		*spur->address = 0;								// Start at CLUT entry 0
		SynchronizeIO();

		// gray the clut.  Each time the address register is hit, delay 800 ns since the
		// clut is retrieving data at the new address.  The address register autoincrements
		// after blue is written, hence the delay
		for (i = 0 ; i < kClutSize ; i++)
		{
		
			(void) DelayForHardware(nsCLUTAddrRegDelay);

			*spur->colorPaletteRAM = *midPointRed;		// Red			
			SynchronizeIO();
			*spur->colorPaletteRAM = *midPointGreen;	// Green			
			SynchronizeIO();
			*spur->colorPaletteRAM = *midPointBlue;		// Blue
			SynchronizeIO();
			
		}
		
		if (vblInterruptsEnabled)
			GraphicsOSSVBLDefaultEnabler();
	}
	err = kGDXErrNoError;								// Everything okay
	
ErrorExit:

	return err;
	
}



//=====================================================================================================
//
// GraphicsHALSetCLUT()
//
//	This routine will program the CLUT with the specified array of 'ColorSpecs'.
//	Two such arrays are provided, the original, and a second that has been luminance mapped (if
//	appropriate) and gamma corrected. It is up to the HAL implementation to decide which array should
//	be applied to the hardware. Most hardware will use the corrected version.
//
//	It is important to note that the positions of the entries refers to logical positions, not physical
//	ones. In 4-bits-per-pixel mode, for example, the entry positions could range from 0, 1, 3,, 15,
//	even though the physical positions may not have this number sequence.
//
//	No range checking is required, because the caller has allready done so.	
//
//		-> originalCSTable		
//		This is a pointer to the array of ColorSpecs provided by the caller. This is only provided in
//		the event that the hardware should not use the orrectedCSTable. If any adjuments need to made
//		to it, then they should be done to a copy. Dont throw away the const!
//
//		-> correctedCSTable
//		This is essentially a copy of originalCSTable, except that it has been luminance mapped (if
//		appropriate) and gamma corrected. Most hardware will use this information to set the CLUT. 
//		Though it is unlikely that you will need to change this information, it is not marked as
//		'const' in case you need to build a special version from the originalCSTable.  In that event,
//		you can alter the array as you see fit.  (Note:  regardless of the size of the originalCSTable,
//		correctedCSTable points to an array of ColorSpecs with 256 entries.
//
//		-> startPosition
//		(0 based) Starting point in the array of ColorSpecs
//		
//		-> numberOfEntries
//		(0 based) This is the number of entries to be set.
//
//		-> sequential
//		If 'false', then the 'value' field of the ColorSpec should be inspected to see what logical
//		position should be set. If 'true', then the array index indicates what logical position should
//		be set.
//
//		-> depthMode
//		The relative bit depth. This is provided so that the HAL can decide how to map the logical
//		entry positions to the physcial entry postions.
//
//=====================================================================================================
GDXErr GraphicsHALSetCLUT(const ColorSpec *originalCSTable, ColorSpec *correctedCSTable,
		SInt16 startPosition, SInt16 numberOfEntries, Boolean sequential, DepthMode depthMode)
{
	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	AbsoluteTime nsCLUTAddrRegDelay = templateHALData->nsCLUTAddrRegDelay;				// 800ns in absolute time to hit clut
	SpurRegisters *spur;

	Boolean vblInterruptsEnabled = false;				// default to not enabled;
	UInt32 startAddress;
	UInt32 entryOffset;
	UInt32 physicalAddress;
	UInt32 logicalAddress;
	
	UInt32 i;									// Loop control variable
	
	GDXErr err = kGDXErrUnknownError;			// Assume failure.

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	spur = &templateHALData->spur;
	
	err = TemplateMapDepthModeToCLUTAttributes(depthMode, &startAddress, &entryOffset);
	if (err)
		goto ErrorExit;

	if (templateHALData->clutOff)		// save to the "virtual" CLUT if current power state is kAVPowerOff
	{
		for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
		{
		
			if (sequential)
				logicalAddress = i;
			else
				logicalAddress = correctedCSTable[i].value;
	
			physicalAddress = startAddress + (logicalAddress * entryOffset);
			
			// correctedCSTable had the ms byte of an original 16 bit red, blue or green in its ls byte
			templateHALData->savedClut[logicalAddress].red = correctedCSTable[i].rgb.red;
			templateHALData->savedClut[logicalAddress].green = correctedCSTable[i].rgb.green;
			templateHALData->savedClut[logicalAddress].blue = correctedCSTable[i].rgb.blue;

		}
	
	}
	
	else
	{
	// Turn off VBL interrupts so we can adjust the CLUT during VBL time, without having to worry
	// about the VBL interrupt handler stealing time from us.
	
		vblInterruptsEnabled = GraphicsOSSVBLDefaultDisabler();						
		TemplateWaitForVBL();
	
		// Program the CLUT entries.  For our hardware, use the correctedCSTable
		for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
		{

			if (sequential)
				logicalAddress = i;
			else
				logicalAddress = correctedCSTable[i].value;
	
			physicalAddress = startAddress + (logicalAddress * entryOffset);
			
			*spur->address = physicalAddress;
			SynchronizeIO();
			(void) DelayForHardware(nsCLUTAddrRegDelay);	// need 800ns after the address register is hit
															// it is fetching the clut data at the address
			
			*spur->colorPaletteRAM = correctedCSTable[i].rgb.red;				// Red			
			SynchronizeIO();
			*spur->colorPaletteRAM = correctedCSTable[i].rgb.green;				// Green			
			SynchronizeIO();
			*spur->colorPaletteRAM = correctedCSTable[i].rgb.blue;				// Blue
			SynchronizeIO();

			// address autoincrements after blue is written.  delay 800 ns since the
			// clut is retrieving data at the new address.  Need these two delays since the address++
			// might not be the desired physical address 
			
			(void) DelayForHardware(nsCLUTAddrRegDelay);
		}
		
		if (vblInterruptsEnabled)
			GraphicsOSSVBLDefaultEnabler();
	
	}
	err = kGDXErrNoError;							// Everything okay
	
ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetCLUT()
//
//	This routine will fill out the specified array of ColorSpecs with the contents of the CLUT.
//	The 'RGBColor' structure in each 'ColorSpec' uses 16-bits for each channel (red, green, and blue),
//	whereas most CLUTs only use 8-bits. Therefore, when filling in the 'RGBColor' structure, the most
//	significant byte for each channel should be filled with the 8-bits extracted from its respective
//	channel in the CLUT. Moreover, to maintain the same behavior as the reference model, the 8-bits
//	from the CLUT should also be written to the least significant byte for each RGBColor.
//	
//	It is important to note that the positions of the entries refer to logical positions, not physical
//	ones. At 4 bpp, for example, the entry positions could range from 0, 1, 2,, 15, even though the
//	physical positions may not have this number sequence.
//
//	No range checking is required, because the caller has already done so.
//
//		<> csTable
//		Pointer to array of 'ColorSpecs' provided by the caller to be filled with the CLUT contents.
//
//		-> startPosition	(0 based) Starting point in the array to fill.
//		-> numberOfEntries	(0 based) The number of entries to be get.
//
//		-> sequential
//		If 'false', then the value field of the 'ColorSpec' should be inspected to see what logical
//		position should be retrieved. If 'true', then the array index indicates what logical position
//		should be read.
//
//		-> depthMode
//		The relative bit depth. This is provided so that the HAL can decide how to map the logical
//		entry positions to the physical entry postions.
//
//=====================================================================================================
GDXErr GraphicsHALGetCLUT(ColorSpec *csTable, SInt16 startPosition, SInt16 numberOfEntries,
		Boolean sequential, DepthMode depthMode)
{	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	AbsoluteTime nsCLUTAddrRegDelay = templateHALData->nsCLUTAddrRegDelay;				// 800ns in absolute time to hit clut
	SpurRegisters *spur;

	UInt32 startAddress;
	UInt32 entryOffset;
	UInt32 physicalAddress;
	UInt32 logicalAddress;
	
	UInt32 i;											// Loop control variable
	
	GDXErr err = kGDXErrUnknownError;					// Assume failure.

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	spur = &templateHALData->spur;
	
	err = TemplateMapDepthModeToCLUTAttributes(depthMode, &startAddress, &entryOffset);
	if (err)
		goto ErrorExit;

	// Read the CLUT entries.
	if (templateHALData->clutOff)				// clut is off so return the "virtual" clut entries
	{
		for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
		{
			if (sequential)
				logicalAddress = i;
			else
				logicalAddress = csTable[i].value;
				
			csTable[i].rgb.red = templateHALData->savedClut[logicalAddress].red;		//		(xxrr)
			csTable[i].rgb.red |= (csTable[i].rgb.red << 8);	// Copy it to most sig. byte 	(rrrr)
			
			csTable[i].rgb.green = templateHALData->savedClut[logicalAddress].green;	//		(xxgg)
			csTable[i].rgb.green |= (csTable[i].rgb.green << 8);// Copy it to most sig. byte 	(gggg)
			
			csTable[i].rgb.blue = templateHALData->savedClut[logicalAddress].blue;		//		(xxbb)
			csTable[i].rgb.blue |= (csTable[i].rgb.blue << 8);	// Copy it to most sig. byte 	(bbbb)
		}
	
	}
	
	else	// read the real clut since it is powered up and running
	{
		for (i = startPosition ; i <= (startPosition + numberOfEntries); i++)
		{
			if (sequential)
				logicalAddress = i;
			else
				logicalAddress = csTable[i].value;
	
			physicalAddress = startAddress + (logicalAddress * entryOffset);
			
			*spur->address = physicalAddress;
			SynchronizeIO();
			(void) DelayForHardware(nsCLUTAddrRegDelay);	// need 800ns after the address register is hit
															// it is fetching the clut data at the address
			
			csTable[i].rgb.red = *spur->colorPaletteRAM;			// Get lower byte for Red		(xxrr)
			SynchronizeIO();
			csTable[i].rgb.red |= (csTable[i].rgb.red << 8);		// Copy it to most sig. byte 	(rrrr)
			
			csTable[i].rgb.green = *spur->colorPaletteRAM;			// Get lower byte for Green		(xxgg)
			SynchronizeIO();
			csTable[i].rgb.green |= (csTable[i].rgb.green << 8);	// Copy it to most sig. byte 	(gggg)
			
			csTable[i].rgb.blue = *spur->colorPaletteRAM;			// Get lower byte for Blue		(xxbb)
			SynchronizeIO();
			csTable[i].rgb.blue |= (csTable[i].rgb.blue << 8);		// Copy it to most sig. byte 	(bbbb)
	
			// address autoincrements after blue is read.  delay 800 ns since the
			// clut is retrieving data at the new address.  Need these two delays since the address++
			// might not be the desired physical address 
			
			(void) DelayForHardware(nsCLUTAddrRegDelay);
	
		}
	}
	
	err = kGDXErrNoError;										// Everything okay
	
ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetPages()
//
//	This routine reports the number of graphics pages supported for the specified 'DisplayModeID' at
//	the specified 'DepthMode'.
//	No attempt should be made to determine whether or not a display capable of being driven with
//	a raster of type 'DisplayModeID' is physically connected.
//
//		-> displaymodeID	The DisplayModeID for which the page count is desired.
//		-> depthMode		The relative bit depth for which the page count is desired.
//
//		<- pageCount
//		# of pages supported at the specified 'DisplayModeID' and 'DepthMode'.  In the event of an
//		error, 'pageCount' is undefined.  This is a counting number, so it is NOT zero based.
//
//=====================================================================================================
GDXErr GraphicsHALGetPages(DisplayModeID displayModeID, DepthMode depthMode, SInt16 *pageCount)
{
	
	GDXErr err = kGDXErrUnknownError;							// Assume failure.
	Boolean	modePossible = false;
	
	// For the Template graphics hardware, the 'pageCount' is always 1 for any supported 
	// 'DisplayModeID'  Therefore, the validity of the  the specified 'DisplayModeID' and 'DepthMode'
	// will be tested by calling GraphicsHALModePossible(), passing in '0' for the page #.
	
	err = GraphicsHALModePossible(displayModeID, depthMode, 0, &modePossible);
	if (!modePossible || err)
	{
		// Opps...caller specified an invalid 'DisplayModeID' or 'DepthMode'
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	// Template graphics architecture only supports 1 page.
	
	*pageCount = 1;
	
	err = kGDXErrNoError;									// Everything okay

ErrorExit:

	return err;
}




//=====================================================================================================
//
// GraphicsHALGetBaseAddress()
//
//	This returns the base address of a specified page in the current mode. 
//	This allows video pages to be written to even when not displayed
//
//			->	page			page number ( 0 based ).  Return the base address for this page
//			<-	baseAddress		base address of desired page
//
//=====================================================================================================
GDXErr GraphicsHALGetBaseAddress(SInt16 page, char **baseAddress)
{


	GDXErr err = kGDXErrUnknownError;			// Assume failure.
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	
	// Currently, Template Software only supports 1 page.  Hence, only page = 0 is valid.
	// All other values will return an error.

	if ( 0 != page )
	{
		// caller asked for baseAddress of invalid page
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	// the base address saved in the HAL data is the base address reported to Quickdraw
	// the base address in TemplateGetXXX routines is the base address that Toynbee is programmed with
	// Relationship: QD base address = Toynbee base address + kHardwareCursorOffset
	*baseAddress = (char *)(templateHALData->baseAddress);
		
	err = kGDXErrNoError;									// Everything okay
		

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetSync()
//
//	Multipurpose call to 
//	1) Report the capabilities of the frame buffer in controlling the sync lines and if HW can
//	"sync" on Red, Green or Blue
//	2) Report the current status of the sync lines and if HW is "syncing" on Red, Green or Blue
//	
//	If the display supported the VESA Device Power Management Standard (DPMS), it would respond
//	to HSync and VSync in the following manner:
//	The VESA Standards are:
//	
//	State       	Vert Sync		Hor Sync		Video
//	-----			--------		---------		------
//	Active         	Pulses      	Pulses      	Active
//	Standby			Pulses   		No Pulses      	Blanked
//	Idle			No Pulses      	Pulses			Blanked
//	Off			 	No Pulses		No Pulses		Blanked
//
//
//			->	getHardwareSyncCapability	true if reporting HW capability
//											false if reporting current status
//
//	For this routine, the relevant fields of the 'VDSyncInfoRec' structure are as follows:
//			<-	sync		report HW capability or current state.
//
//				if getHardwareSyncCapability then report the cabability of the HW
//				When reporting the capability of the HW, set the appropriate bits of csMode:
//				kDisableHorizontalSyncBit		set if HW can disable Horizontal Sync
//				kDisableVerticalSyncBit			set if HW can disable Vertical Sync
//				kDisableCompositeSyncBit		set if HW can disable Composite Sync
//				kSyncOnRedEnableBit				set if HW can sync on Red
//				kSyncOnGreenEnableBit			set if HW can sync on Green
//				kSyncOnBlueEnableBit			set if HW can sync on Blue
//				kNoSeparateSyncControlBit		set if HW CANNOT enable/disable H,V,C sync independently
//												Means that HW ONLY supports the VESA DPMS "OFF" state
//
//				if !getHardwareSyncCapability then report the current state of sync lines and  if HW is
//				"syncing" on Red, Green or Blue
//				Reporting the "current state of the sync lines" means: "Report the State of the Display"
//				To report the current state of the display, the kDisableHorizontalSyncBit and
//				the kDisableVerticalSyncBit have the following values:
//
//				State       	kDisableVerticalSyncBit		kDisableHorizontalSyncBit		Video
//				-----			-----------------------		-------------------------		------
//				Active       		  	0      						0      					Active
//				Standby					0   						1		   			   	Blanked
//				Idle					1     			 			0						Blanked
//				Off			 			1							1						Blanked
//
//				to report if HW is "syncing" on Red, Green or Blue:
//				kSyncOnRedEnableBit				set if HW is "syncing" on Blue
//				kSyncOnGreenEnableBit			set if HW is "syncing" on Green
//				kSyncOnBlueEnableBit			set if HW is "syncing" on Blue
//
//
//=====================================================================================================
GDXErr GraphicsHALGetSync(Boolean getHardwareSyncCapability, VDSyncInfoRec *sync)
{
	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	GDXErr err = kGDXErrUnknownError;								// Assume failure.


	if (getHardwareSyncCapability)
	{
		// report capability of the hardware.  Template CAN control H,V,C sync independently and
		// is UNABLE to sync on Red/Green/Blue
		sync->csMode = 0	|
						  1 << kDisableHorizontalSyncBit |
						  1 << kDisableVerticalSyncBit |
						  1 << kDisableCompositeSyncBit;
		sync->csFlags = 0;
	}
	else 
	{		
		// Report the current status.  Since Template does not support
		// the sync on Red/Green/Blue bits, just return the DPMS status
		sync->csMode = (UInt8)templateHALData->cvhSyncDisabled;
		sync->csFlags = 0;
	}

	err = kGDXErrNoError;											// Everything okay
		

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALSetSync()
//	If the display supported the VESA Device Power Management Standard (DPMS), it would respond
//	to HSync and VSync in the following manner:
//
//	The VESA Standards are:
//	
//	State       	Vert Sync		Hor Sync		Video
//	-----			--------		---------		------
//	Active         	Pulses      	Pulses      	Active
//	Standby			Pulses   		No Pulses      	Blanked
//	Idle			No Pulses      	Pulses			Blanked
//	Off			 	No Pulses		No Pulses		Blanked
//
//
//			->	syncBitField		bit field of the sync bits that need to be disabled/enabled
//
//				kDisableHorizontalSyncBit		set if HW should disable Horizontal Sync (No Pulses)
//				kDisableVerticalSyncBit			set if HW should disable Vertical Sync (No Pulses)
//				kDisableCompositeSyncBit		set if HW should disable Composite Sync (No Pulses)
//				kSyncOnRedEnableBit				set if HW should sync on Red
//				kSyncOnGreenEnableBit			set if HW should sync on Green
//				kSyncOnBlueEnableBit			set if HW should sync on Blue
//
//
//			->	syncBitFieldValid	mask of the bits that are valid in the csMode bit field
//
//
//=====================================================================================================
GDXErr GraphicsHALSetSync(UInt8 syncBitField, UInt8 syncBitFieldValid)
{

	UInt32 i;											// cluty loop iterator
	UInt32 numberOfEntries;						// clut will have 31 or 255 entries (0 based)
	UInt32 mrSanAntonioMode;									// program mrSanAntonio with correct c,v,h Syncs
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	MrSanAntonioRegisters *mrSanAntonio = &templateHALData->mrSanAntonio;
	UInt16 cvhSyncDisabled = templateHALData->cvhSyncDisabled;						

	GDXErr err = kGDXErrUnknownError;								// Assume failure
	
	// Template is UNABLE to sync on Red/Green/Blue and reported that on the GetSync status call.
	// if any of the kEnableSyncOnBlue, kEnableSyncOnGreen or kEnableSyncOnRed bits in csFlags
	// are set than return an error.
	
	if (kSyncOnMask & syncBitFieldValid)
	{
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	// if HW did not support the disabling of c,v,h sync independently, the HAL should check:
	// that bits kDisableCompositeSyncBit, kDisableVerticalSyncBit, kDisableHorizontalSyncBit in
	// 1) syncBitFieldValid are ALL marked as valid 
	// 2) syncBitField are ALL the same state
		
	
	if (kCompositeSyncMask & syncBitFieldValid)
	{
		// save state of compositeSync
		cvhSyncDisabled =	(cvhSyncDisabled & ~kCompositeSyncMask) |
							(syncBitField & kCompositeSyncMask);
	}
	if (kVerticalSyncMask & syncBitFieldValid)
	{
		// save state of verticalSync
		cvhSyncDisabled =	(cvhSyncDisabled & ~kVerticalSyncMask) |
							(syncBitField & kVerticalSyncMask);
	}
	if (kHorizontalSyncMask & syncBitFieldValid)
	{
		// save state of horizontalSync
		cvhSyncDisabled =	(cvhSyncDisabled & ~kHorizontalSyncMask) |
							(syncBitField & kHorizontalSyncMask);
	}
	
	
	
	mrSanAntonioMode = *mrSanAntonio->mrSanAntonioMode;
	mrSanAntonioMode &= ~kMrSanAntonioSyncsAndCBlankMask;			// clear c,h,v Syncs and CBlank

	// Template enables a sync by writing a 1 to the sync.  So the value of each sync is ~cvh bit in
	// in cvhSyncDisabled.  For each sync in MrSanAntonio, 1 = enabled, 0 = disabled
	if (!(cvhSyncDisabled & kCompositeSyncMask))
		mrSanAntonioMode |= kCSyncEnableMask; 				// start composite pulses
	if (!(cvhSyncDisabled & kVerticalSyncMask))
		mrSanAntonioMode |= kVSyncEnableMask; 				// start vertical pulses
	if (!(cvhSyncDisabled & kHorizontalSyncMask))
		mrSanAntonioMode |= kHSyncEnableMask; 				// start horizontal pulses

	if (0 == cvhSyncDisabled)							// if all syncs are on (display is active)
		mrSanAntonioMode |= kCBlankEnableMask;				// then enable CBlank


	*mrSanAntonio->mrSanAntonioMode = mrSanAntonioMode;
	
	templateHALData->cvhSyncDisabled = cvhSyncDisabled;
			
	err = kGDXErrNoError;								// Everything okay

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetModeTiming()
//
//	This is used to to gather scan timing information.  For the specified displayModeID, report
//	the timingFlags for the connected display.  If the HAL doesn't know if the display supports the
//	displayModeID, timingFlags should be 0 (kModeValid and kModeSafe bits should NOT be set).  That
//	indicates to the Display Mgr that the driver doesn't think the display can handle the resolution,
//	but it will let the Display Mgr make the final decision.
//
//			->	displayModeID	Describes the display resolution and scan timing
//
//			<-	timingFormat	Format of the info in 'timingData' field, only 'kDeclROMtables' is valid.
//
//			<-	timingFlags		Whether the display mode with these scan timings is required or optional.
//								If a 'displayModeID' is not thought to be valid for given display, set
//								timingFlags to 0 (invalid and unsafe)...the DisplayMgr will ask
//								display modules if the mode is valid.
//
//=====================================================================================================
GDXErr GraphicsHALGetModeTiming(DisplayModeID displayModeID, UInt32 *timingFormat,
			UInt32 *timingFlags)
{
	
	
	// Define a new type which is a table of supported 'DisplayModeIDs' and their corresponding
	// timingdata and their timing flags
	
	typedef struct DisplayModeTimingTable DisplayModeTimingTable;
	struct DisplayModeTimingTable
	{
		DisplayCode displayCode;
		DisplayModeID displayModeID;
		UInt32 timingFlags;
	};
	
	enum
	{	
		maxTimingTableEntries = 36,
		valid = (1 << kModeValid),
		validAndSafe = (1 << kModeValid | 1 << kModeSafe),
		validAndSafeAndDefault = ( (1 << kModeValid) | (1 << kModeSafe) | (1 << kModeDefault) )
	};
	
	DisplayModeTimingTable theDisplayModeTimingTable[maxTimingTableEntries] =
	{
		{kDisplayCode12Inch, kDisplay512x384At60Hz, validAndSafe},
		{kDisplayCodeStandard, kDisplay640x480At67Hz, validAndSafe},
		{kDisplayCodePortrait, kDisplay640x870At75Hz, validAndSafe},
		{kDisplayCodePortraitMono, kDisplay640x870At75Hz, validAndSafe},
		{kDisplayCode16Inch, kDisplay832x624At75Hz, validAndSafe},
		{kDisplayCode19Inch, kDisplay1024x768At75Hz, validAndSafe},
		{kDisplayCode21Inch, kDisplay1152x870At75Hz, validAndSafe},
		{kDisplayCode21InchMono, kDisplay1152x870At75Hz, validAndSafe},
	
		{kDisplayCodeVGA, kDisplay640x480At60HzVGA, validAndSafeAndDefault},
		{kDisplayCodeVGA, kDisplay800x600At60HzVGA, validAndSafe},
		{kDisplayCodeVGA, kDisplay800x600At72HzVGA, validAndSafe},
		{kDisplayCodeVGA, kDisplay800x600At75HzVGA, validAndSafe},
		{kDisplayCodeVGA, kDisplay1024x768At60HzVGA, valid},
		{kDisplayCodeVGA, kDisplay1024x768At72HzVGA, valid},
		{kDisplayCodeVGA, kDisplay1024x768At75HzVGA, valid},

		{kDisplayCodeNTSC, kDisplay512x384At60HzNTSC, validAndSafe},
		{kDisplayCodeNTSC, kDisplay640x480At60HzNTSC, validAndSafeAndDefault},

		{kDisplayCodePAL, kDisplay640x480At50HzPAL, validAndSafe},
		{kDisplayCodePAL, kDisplay768x576At50HzPAL, validAndSafeAndDefault},

		{kDisplayCodeMultiScanBand1, kDisplay640x480At67Hz, validAndSafeAndDefault},
		{kDisplayCodeMultiScanBand1, kDisplay800x600At60HzVGA, valid},
		{kDisplayCodeMultiScanBand1, kDisplay800x600At72HzVGA, valid},
		{kDisplayCodeMultiScanBand1, kDisplay832x624At75Hz, validAndSafe},

		{kDisplayCodeMultiScanBand2, kDisplay640x480At67Hz, validAndSafe},
		{kDisplayCodeMultiScanBand2, kDisplay832x624At75Hz, validAndSafeAndDefault},
		{kDisplayCodeMultiScanBand2, kDisplay1024x768At60HzVGA, valid},
		{kDisplayCodeMultiScanBand2, kDisplay1024x768At72HzVGA, valid},
		{kDisplayCodeMultiScanBand2, kDisplay1024x768At75Hz, validAndSafe},
		
		{kDisplayCodeMultiScanBand3, kDisplay640x480At67Hz, validAndSafe},
		{kDisplayCodeMultiScanBand3, kDisplay832x624At75Hz, validAndSafe},
		{kDisplayCodeMultiScanBand3, kDisplay1024x768At60HzVGA, valid},
		{kDisplayCodeMultiScanBand3, kDisplay1024x768At72HzVGA, valid},
		{kDisplayCodeMultiScanBand3, kDisplay1024x768At75Hz, validAndSafe},
		{kDisplayCodeMultiScanBand3, kDisplay1152x870At75Hz, validAndSafeAndDefault},
		{kDisplayCodeMultiScanBand3, kDisplay1280x960At75Hz, validAndSafe},
		{kDisplayCodeMultiScanBand3, kDisplay1280x1024At75Hz, validAndSafe}
	};
	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	
	DisplayCode displayCode = templateHALData->displayCode;	// class of connected display
	
	UInt32 i;			// omnipresent loop iterator
	
	GDXErr err = kGDXErrNoError;			// Never fails.  timingFlags = 0 if we don't think
											// monitor supports the displayModeID
	Boolean modePossible = false;
	*timingFlags = 0;				// Default to invalid and unsafe.  If HAL doesn't know if a display
									// supports a displayModeID, the OS will ask a display module

	*timingFormat = kDeclROMtables;	// Default to kDeclROMtables -- the only valid timingFormat

	// Prior to doing anything else, do some paranoid error checking, and make sure that Toynbee
	// can drive the indicated 'displayModeID.'  This can be accomplished by calling the
	// GraphicsHALModePossible() routine.

	err = GraphicsHALModePossible(displayModeID, kDepthMode1, 0, &modePossible);
	if (!modePossible || err)
	{
		// Opps...caller specified an invalid 'DisplayModeID' or 'DepthMode'
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}

	
	// Scan the theDisplayModeTimingTable to see if the connected monitor supports the displayModeID

	for (i = 0; i < maxTimingTableEntries; i++)
	{
		if ((theDisplayModeTimingTable[i].displayCode == displayCode) && 
				(theDisplayModeTimingTable[i].displayModeID == displayModeID))
		{
				*timingFlags = theDisplayModeTimingTable[i].timingFlags;
				break;
	
		}
	}

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetNextResolution()
//
//	This call will take a 'previousDisplayModeID' and return the next supported display mode.
//	The Core takes care of most of the work.  The HAL just needs to look at the 
//	'previousDisplayModeID' and return the next supported d'isplayModeID' and the max depthMode
//	that the HW supports.
//	The Core has already checked to see if previousDisplayModeID = -1, in which case it has
//	returned the current resolution's information.
//	The HAL returns all DisplayModeIDs that it supports.  Pay No attention to the monitor behind
//	the sense lines.
//
//			->	previousDisplayModeID		
//			'previousDiplayModeID = kDisplayModeIDCurrent' never happens.  Core will handle that case. 
//			If 'previousDiplayModeID = kDisplayModeIDFindFirstResolution', get the first supported 
//			resolution for the monitor.
//			Otherwise, 'previousDiplayModeID' contains the previous displayModeID (hence its name)
//			from the previous call.
//
//			<-	displayModeID	ID of the next display mode following 'csPreviousDiplayModeID'
//			Set to 'kDisplayModeIDNoMoreResolutions' once all supported display modes have been reported.
//
//			<-	maxDepthMode	Maximum relative bit depth for the 'displayModeID'
//
//=====================================================================================================
GDXErr GraphicsHALGetNextResolution(DisplayModeID previousDisplayModeID,
		DisplayModeID *displayModeID, DepthMode *maxDepthMode)
{
	
	GDXErr err = kGDXErrUnknownError;								// Assume failure	
			
	
enum { maxResolutions = 19 };

	DisplayModeID theResolutionTable[maxResolutions] =
	{
		kDisplay512x384At60HzNTSC,
		kDisplay512x384At60Hz,
		kDisplay640x480At50HzPAL,
		kDisplay640x480At60HzNTSC,
		kDisplay640x480At60HzVGA,
		kDisplay640x480At67Hz,

		kDisplay640x870At75Hz,

		kDisplay768x576At50HzPAL,
		kDisplay800x600At60HzVGA,
		kDisplay800x600At72HzVGA,
		kDisplay800x600At75HzVGA,
		kDisplay832x624At75Hz,

		kDisplay1024x768At60HzVGA,
		kDisplay1024x768At72HzVGA,
		kDisplay1024x768At75HzVGA,

		kDisplay1024x768At75Hz,
		
		kDisplay1152x870At75Hz,
		kDisplay1280x960At75Hz,
		kDisplay1280x1024At75Hz
	};
	
	
	
	UInt32 i;							// omnipresent loop iterator
	UInt32 displayModeIDIndex;			// index for the returned DisplayModeID
	
	// if previousDisplayModeID = kDisplayModeIDFindFirstResolution, return the first resolution 
	// that HW supports

	if (kDisplayModeIDFindFirstResolution == previousDisplayModeID)
		displayModeIDIndex = 0;		// index to first supported resolution
	
	else							// find the previousDisplayModeID
	{
		i = 0;
		while ( (i < maxResolutions) && 
			(theResolutionTable[i] != previousDisplayModeID) )
				i++;

		// Check if i has exceeded maxResolutions.  This means the
		// previousDisplayModeID was not valid.
	
		if ( maxResolutions == i )
		{
			err = kGDXErrInvalidParameters;
			goto ErrorExit;
		}
		
		displayModeIDIndex = ++i;		// index to next supported resolution. (or no more)
		

	}
	
	// displayModeIndex now points to the nextresolution.
	// Make sure 'displayModeIndex != kMaxResolutions'... if it does, all resolutions have been reported.
	if (displayModeIDIndex < maxResolutions)
	{
		*displayModeID = theResolutionTable[displayModeIDIndex];
		err = GraphicsHALGetMaxDepthMode(*displayModeID, maxDepthMode);

		if (err)		// system is configured such that displayModeID is not supported
						// recurse Dammit to get nextresolution
			err = GraphicsHALGetNextResolution(*displayModeID, displayModeID, maxDepthMode);
	}
	else
	{
		*displayModeID = kDisplayModeIDNoMoreResolutions;		// no more supported resolutions
		err = kGDXErrNoError;
	}
	
ErrorExit:
	return err;
}



//=====================================================================================================
//
// GraphicsHALGetVideoParams()
//
//	The HAL only needs to return the absolute bits per pixel and the rowBytes for a
//	given depthMode and displayModeID.  The displayModeID is passed in since rowBytes might depend
//	on the resolution.
//
//		-> displayModeID	the displayModeID for which the info is desired
//		-> depthMode		The relative bit depth for which the info is desired
//		<- bitsPerPixel		absolute bit depth for the given depthMode
//		<> rowBytes			on input, rowbytes contains the horizontal pixels for the dispalyModeID
//							on output, width of each row of video memory for the given depthMode
//
//=====================================================================================================
GDXErr GraphicsHALGetVideoParams(DisplayModeID displayModeID, DepthMode depthMode,
		UInt32 *bitsPerPixel, SInt16 *rowBytes)
{
	GDXErr err = kGDXErrUnknownError;				// Assume failure	

	err = GraphicsHALMapDepthModeToBPP(depthMode, bitsPerPixel);
	if (err)
		goto ErrorExit;
	
	switch (*bitsPerPixel)
	{
		case 8:
			 *rowBytes = *rowBytes + kRowBytesOffset;
			break;
		case 16:
			*rowBytes = (*rowBytes * 2) + kRowBytesOffset ;
			break;
		case 32:
			*rowBytes = (*rowBytes * 4) + kRowBytesOffset;
			break;
		default:
			err = kGDXErrInvalidParameters;				// invalid depth...shouldn't happen
	}

	// Template has nonstandard rowBytes when in 832 x 624 mode and there is only 1 meg of VRAM
	if ( kDisplay832x624At75Hz == displayModeID )
	{
		BitDepthIndependentData bdiData;			// dependent on resolution	(ignored)
		BitDepthDependentData bddData;				// dependent on depthMode	(ignored)
		VRAMWidthDependentData vwData;				// dependent on VRAM width	(ignored)
		WidthAndDepthDependentData vwdData;			// dependent on VRAM width and depthMode (ignored)
		DisplayInfo info;							// describes cabilities of display (get rowBytes from here)
		
		TemplateHALData *templateHALData = GraphicsHALGetHALData();
		VRAMSize vramSize = templateHALData->vramSize;
		Boolean vramWidth32 = templateHALData->vramWidth32;
		
		// get the rowBytes for the requested displayModeID and depthMode
		err = TemplateGetDisplayData(displayModeID, depthMode, vramSize, vramWidth32, 
				&bdiData, &bddData, &vwData, &vwdData, &info);
	
		if (err)
			goto ErrorExit;
			
		*rowBytes = info.rowBytes;
	}

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetMaxDepthMode()
//
// This takes a 'displayModeID' and returns the maximum depthMode that is supported by the
// hardware for that 'displayModeID'  NO check is made to determine if the 'displayModeID' is
// valid for the connected monitor.  The HAL should return an error if the 'displayModeID' is
// not supported or there is not enough VRAM to support the 'displayModeID'
//
// This function DETERMINES if the HW supports a given displayModeID
//
//			->	displayModeID		get the information for this display mode
//			<-	maxDepthMode		Maximum bit depth for the DisplayModeID (relative depth mode)
//
//=====================================================================================================
GDXErr GraphicsHALGetMaxDepthMode(DisplayModeID displayModeID, DepthMode *maxDepthMode)
{

	BitDepthIndependentData bdiData;			// dependent on resolution	(ignored)
	BitDepthDependentData bddData;				// dependent on depthMode	(ignored)
	VRAMWidthDependentData vwData;				// dependent on VRAM width	(ignored)
	WidthAndDepthDependentData vwdData;			// dependent on VRAM width and depthMode	(ignored)
	DisplayInfo info;							// describes cabilities of display (get maxDepthMode from here)
	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	VRAMSize vramSize = templateHALData->vramSize;
	Boolean vramWidth32 = templateHALData->vramWidth32;

	GDXErr err = kGDXErrUnknownError;								// Assume failure

	// get the maxDepthMode for the requested displayModeID
	err = TemplateGetDisplayData(displayModeID, kDepthMode1, vramSize, vramWidth32, &bdiData, &bddData,
			&vwData, &vwdData, &info);
	
	if (err)										// unknown displayModeID OR vramSize, vramWidth32 are
		goto ErrorExit;								// are really screwed up

	if (kDepthModeInvalid == info.maxDepthMode)
	{
		err = kGDXErrRequestedModeNotPossible;		// not enough vram to do depthmode
		goto ErrorExit;
	}
	
	*maxDepthMode = info.maxDepthMode;				

	if ( templateHALData->monoOnly )
		*maxDepthMode = kDepthMode1;	// if on a mono only display, never do more than 256 colors

	err = kGDXErrNoError;

ErrorExit:
	
	return err;

}



//=====================================================================================================
//
//  GraphicsHALMapDepthModeToBPP()
//	This routine maps a relative pixel depth (DepthMode) to an absolute pixel depth (bits per pixel)
//
//		-> depthMode		The relative pixel depth 
//
//		<- bitsPerPixel		Corresponding abosolute pixel depth.
//
//=====================================================================================================
GDXErr GraphicsHALMapDepthModeToBPP(DepthMode depthMode, UInt32 *bitsPerPixel)
{
	GDXErr err = kGDXErrUnknownError;			// Assume failure.
	
	switch (depthMode)
	{
		case kDepthMode1:
			*bitsPerPixel = 8;
			break;
		case kDepthMode2:
			*bitsPerPixel = 16;
			break;
		case kDepthMode3:
			*bitsPerPixel = 32;
			break;
		default:								// someone passed in bogus depthMode...say it is 8
			*bitsPerPixel = 8;
			err = kGDXErrUnableToMapDepthModeToBPP;
			goto ErrorExit;
	}
	

	err = kGDXErrNoError;

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALModePossible()
//
//	This routine checks to see if the frame buffer is cabable of driving the given 'displayModeID' at
//	the indicated 'depthMode' and 'page'. This DOES NOT check to see that the 'displayModeID' is valid
//	for the display type that is physically connected to the frame buffer.
//
//	IMPORTANT NOTE: The GDXErr return value DOES NOT indicate whether the mode is possible or not. 
//	It only signifies whether or not the value returned in 'modePossible' was correctly determined.  In
//	the event of an error, 'modePossible' does not contain valid information.
//
//		-> displaymodeID		The desired DisplayModeID. 
//		-> depthMode			The desired relative bit depth.
//		-> page					The desired page.
//
//		<- modePossible
//		This will be 'true' if the frame buffer can support the desired items, 'false' otherwise. In 
//		the event of an error, 'modePossible' is undefined.
//
//=====================================================================================================
GDXErr GraphicsHALModePossible(DisplayModeID displayModeID, DepthMode depthMode, SInt16 page,
		Boolean *modePossible)
{
	
	DepthMode maxDepthMode;						// the max depthMode for the DisplayModeID

	GDXErr err = kGDXErrNoError;				// if displayModeID, depthMode or page isn't supported
												// just return modePossible = false

	*modePossible = false;						// Assume mode not possible
	
	// Check to see if the requested page is valid.  For Template hardware, page 0
	// is the only valid page supported.
	
	if (0 != page)
		goto ErrorExit;
		
	// Find the maximum depthMode for the displayModeID
	// If the displayModeID is not supportted by hardware, GetMaxDepthMode() returns
	// kGDXErrDisplayModeIDUnsupported 

	err =  GraphicsHALGetMaxDepthMode(displayModeID, &maxDepthMode);

	if (err)
		goto ErrorExit;


	if (maxDepthMode >= depthMode)
	{
		*modePossible = true;
	}

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALDetermineDisplayCode()
//
//	This routine is called whenever it is necessary to determine the type of display that is
//	connected to the frame buffer controller.
//	When this routine is called, the following actions should occur:
//
//		 Perform required steps to determine what display is connected (e.g., read sense lines)
//		 Update the HAL's state information regarding the type of display connected (if HAL
//		  implementation maintains that state information)
//
//	In the event that the HAL is does not the specific type of display attached, it should set
//	'*displayCode = kDisplayCodeUnknown'
//
//			<- displayCode	DisplayCode for the attached display.
//
//=====================================================================================================
GDXErr GraphicsHALDetermineDisplayCode(DisplayCode *displayCode)
{
	RawSenseCode rawSenseCode;						// Result from reading sense lines after reset
	ExtendedSenseCode extendedSenseCode;			// Result from applying extended sense algorithm 

	GDXErr err = kGDXErrNoError;					// Assume sucess

	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	
	// The rawSenseCode is obtained by making sure the sense lines are reset (i.e., none of the lines
	// are being actively driven by the frame buffer controller) and then reading the state of the
	// lines.

	
	TemplateResetSenseLines();						
	rawSenseCode = TemplateReadSenseLines();	

	extendedSenseCode = TemplateGetExtendedSenseCode();
	
	// Since the Template architecture uses 'standard' sense codes, call the utility function
	// that maps the raw and extendended sense codes to a 'displayCode.'

	err = GraphicsUtilMapSenseCodesToDisplayCode(rawSenseCode, extendedSenseCode, displayCode);
	if (err)
		goto ErrorExit;

	templateHALData->displayCode = *displayCode;		// Update HAL information

	if ( (kDisplayCode21InchMono == *displayCode) || (kDisplayCodePortraitMono == *displayCode) )
		 templateHALData->monoOnly = true;
	else
		templateHALData->monoOnly = false;				

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetSenseCodes()
//
//	This routine is called whenever the state of the sense codes need to be reported.
//	NOTE:  This should only report the the sense code information.  No attempt should be made to
//	determine what type of display is attached here.  Moreover, the sense codes should be
//	determined EVERY time this call is made, and not make use of any previously saved values.
//
//		<- rawSenseCode
//		For 'standard' sense code hardware, this value is found by instructing the frame buffer
//		controller NOT to actively drive any of the monitor sense lines, and then reading the
//		state of the the monitor sense lines 2, 1, and 0.  (2 is the MSB, 0 the LSB)
//
//		<- extendedSenseCode
//		For 'standard' sense code hardware, the extended sense code algorithm is as follows:
//		(Note:  as described here, sense line 'A' corresponds to '2', 'B' to '1', and 'C' to '0')
//			 Drive sense line 'A' low and read the values of 'B' and 'C'.  
//			 Drive sense line 'B' low and read the values of 'A' and 'C'.
//			 Drive sense line 'C' low and read the values of 'A' and 'B'.
//
//		In this way, a six-bit number of the form BC/AC/AB is generated. 
//
//		<- standardInterpretation
//		If 'standard' sense code hardware is implemented (or the values are coerced to appear
//		'standard' then set this to TRUE.  Otherwise, set it to FALSE, and the interpretation for
//		'rawSenseCode' and 'extendedSenseCode' will be considered private.
//
//=====================================================================================================
GDXErr GraphicsHALGetSenseCodes(RawSenseCode *rawSenseCode, ExtendedSenseCode *extendedSenseCode,
		Boolean *standardInterpretation)
{
	GDXErr err = kGDXErrNoError;							// Assume sucess
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	
	// The rawSenseCode is obtained by making sure the sense lines are reset (i.e., none of the lines
	// are being actively driven by the frame buffer controller) and then reading the state of the
	// lines.
	
	TemplateResetSenseLines();						
	*rawSenseCode = TemplateReadSenseLines();	

	*extendedSenseCode = TemplateGetExtendedSenseCode();
		
	*standardInterpretation = true;						// Template has 'standard' sense lines
			
ErrorExit:
	return err;
}



//=====================================================================================================
//
// GraphicsHALGetDefaultDisplayModeID()
//
//	The 'displayCode' is passed in and the HAL returns the default 'displayModeID' and the
//	'depthMode'.  This routine gets called when a new monitor is connected to the computer.  The
//	HAL knows how much VRAM is available and whether it can handle a given 'displayModeID' for a 
//	monitor.
//	For example, the 'kIndexedMultiScanBand3' has a default 'displayModeID' of kDisplay1152x870At75Hz.
//	If there is not enough VRAM available, the HAL is unable to switch into that resolution, hence
//	the HAL will switch into the next best resolution.
//	The HAL will also return the max 'depthMode' that can be supported for the resolution.  This
//	enables a switch from a 13" monitor at millions of colors to be switched with a 21" monitor
//	and, if there is enough VRAM, the depthMode will still be at millions of colors
//
//		-> displayCode			the connected display type
//		<- displaymodeID		The default DisplayModeID for the connected monitor type
//		<- depthMode			The default depthMode 
//
//=====================================================================================================
GDXErr GraphicsHALGetDefaultDisplayModeID(DisplayCode displayCode, DisplayModeID *displayModeID,
		DepthMode *depthMode)
{
	
	// Define a new type which is a table of default 'DisplayModeIDs' for the 
	// connected monitor.
	
	typedef struct DefaultResolutionTable DefaultResolutionTable;
	struct DefaultResolutionTable
	{
		DisplayCode displayCode;
		DisplayModeID displayModeID;
	};
	
	enum { maxDefaultTableEntries = 15 };

	DefaultResolutionTable defaultResolutionTable[maxDefaultTableEntries] =
	{
		{kDisplayCodeUnknown, kDisplay640x480At67Hz},
		{kDisplayCode12Inch, kDisplay512x384At60Hz},
		{kDisplayCodeStandard, kDisplay640x480At67Hz},
		{kDisplayCodePortrait, kDisplay640x870At75Hz},
		{kDisplayCodePortraitMono, kDisplay640x870At75Hz},
		{kDisplayCode16Inch, kDisplay832x624At75Hz},
		{kDisplayCode19Inch, kDisplay1024x768At75Hz},
		{kDisplayCode21Inch, kDisplay1152x870At75Hz},
		{kDisplayCode21InchMono, kDisplay1152x870At75Hz},
		{kDisplayCodeVGA, kDisplay640x480At60HzVGA},
		{kDisplayCodeNTSC, kDisplay640x480At60HzNTSC},
		{kDisplayCodePAL, kDisplay768x576At50HzPAL},
		{kDisplayCodeMultiScanBand1, kDisplay640x480At67Hz},
		{kDisplayCodeMultiScanBand2, kDisplay832x624At75Hz},
		{kDisplayCodeMultiScanBand3, kDisplay1152x870At75Hz}
	};
	
	UInt32 i;								// loop iterator
	GDXErr err = kGDXErrMonitorUnsupported;			// Assume failure	
	
	
	// Scan the defaultResolutionTable to find the the connected monitor	
	
	for (i = 0; i < maxDefaultTableEntries; i++)
	{
		if (defaultResolutionTable[i].displayCode == displayCode)
		{
			*displayModeID = defaultResolutionTable[i].displayModeID;
			err = kGDXErrNoError;
			break;					// found displayCode
		}
	}

	if (err)											// didn't find displayCode
		goto ErrorExit;
		
	err = GraphicsHALGetMaxDepthMode(*displayModeID, depthMode);

	
ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALProgramHardware()
//
//	This routine attempts to program the graphics hardware to the desired 'displayModeID', 'depthMode',
//	and 'page'.  It is not required to specicifically check to see if the inputs are valid, since it
//	can assume that the checking has been done elsewhere.
//
//		-> displaymodeID		The desired DisplayModeID. 
//		-> depthMode			The desired relative bit depth.
//		-> page					The desired page.
//
//		<- directColor
//		This is 'true' if the desired depthMode results in the hardware being in a direct color mode,
//		otherwise it is 'false'. In the event on an error, it is undefined.
//
//		<- baseAddress
//		The resulting base address of the frame buffers ram. In the event of an error, it is undefined.
//
//=====================================================================================================
GDXErr GraphicsHALProgramHardware(DisplayModeID displayModeID, DepthMode depthMode, SInt16 page,
		Boolean *directColor, char **baseAddress)
{
	Boolean resolutionChange = false;			// don't hit timing registers if same resolution
	Boolean vblInterruptsEnabled = false;		// default to not enabled;
	BitDepthIndependentData bdiData;			// dependent on resolution
	BitDepthDependentData bddData;				// dependent on depthMode
	VRAMWidthDependentData vwData;				// dependent on VRAM width
	WidthAndDepthDependentData vwdData;			// dependent on VRAM width and depthMode
	DisplayInfo info;							// describes cabilities of display
	
	UInt32 *pBuffer;
	UInt32 rowBytes;
	UInt32 i, j;								// omnipresent loop iterators

	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	VRAMSize vramSize = templateHALData->vramSize;
	Boolean vramWidth32 = templateHALData->vramWidth32;

	GDXErr err = kGDXErrUnknownError;								// Assume failure

	// Flush out any deferred cursor moving without actually redrawing the cursor.
	if (false != templateHALData->hardwareCursorData.deferredMove)
	{
		templateHALData->hardwareCursorData.x =
			templateHALData->hardwareCursorData.deferredX;
		templateHALData->hardwareCursorData.y =
			templateHALData->hardwareCursorData.deferredY;

		templateHALData->hardwareCursorData.deferredMove = false;
	}

	// Get 'bdiData', 'bddData', 'vwData', 'vwdData' and 'info' for requested 'DisplayModeID' and 'DepthMode'
	err = TemplateGetDisplayData(displayModeID, depthMode, vramSize, vramWidth32, &bdiData, &bddData,
			&vwData, &vwdData, &info);
	
	if (err)
		goto ErrorExit;
	
	// Check if data is valid: check that the maxDepthMode is greater than or equal to the requested DepthMode
	if (info.maxDepthMode < depthMode)
	{
		err = kGDXErrRequestedModeNotPossible;		// not enough vram or can't do depthMode because dbl buffered
		goto ErrorExit;
	}
	
	// Start programing the hardware to the desired state.  For all these function calls, we are
	// explicitly ignoring any errors, since we have passed 'the point of no return'
	
	vblInterruptsEnabled = GraphicsOSSVBLDefaultDisabler();						
	TemplateWaitForVBL();

	if (displayModeID != templateHALData->displayModeID)
		resolutionChange = true;
	
	// turn hardware cursor off
	*templateHALData->spur.address = kSpurControl;
	SynchronizeIO ();
	*templateHALData->spur.multiPort &= ~kSpurControlCursorEnableMask;
	SynchronizeIO ();
	templateHALData->hardwareCursorData.cursorRendered = false;
	templateHALData->hardwareCursorData.cursorVisible = false;

	// When only the depthMode changes, there is no need to put Toynbee in reset.  That would
	// cause the screen to go black.  When the resolution changes, however, Toynbee must be reset
	// or the display won't work.  If the clut is off, however, never assert reset nor release
	// reset(since releasing reset would use more power).  Toynbee will
	// be taken out of reset when the monitor is turned back on.

	if (!templateHALData->clutOff && resolutionChange)
		(void) TemplateAssertVideoReset(&templateHALData->toynbee);
	
	if (resolutionChange)
		(void) TemplateSetupClockGenerator(&templateHALData->spur, &bdiData, &templateHALData->usingClockSetA);	
	
	(void) TemplateSetupCLUT(&templateHALData->spur, &vwdData);
	
	(void) TemplateSetupFBController(resolutionChange, &templateHALData->toynbee, 
			&templateHALData->mrSanAntonio, &bdiData, &bddData, &vwData, &vwdData);
	
	if (!templateHALData->clutOff && resolutionChange)
		(void) TemplateReleaseVideoReset(&templateHALData->toynbee);
	
	// interrupts occur when last line is drawn
	*templateHALData->mrSanAntonio.cursorLine = bdiData.verticalFrontPorch/2;		
	
	if (vblInterruptsEnabled)
		GraphicsOSSVBLDefaultEnabler();

	// Save any HAL data that might have changed
	templateHALData->displayModeID = displayModeID;
	templateHALData->depthMode = depthMode;

	// Base address that is reported to QD
	templateHALData->baseAddress = (Ptr)(bdiData.baseAddress + kHardwareCursorOffset
											+ templateHALData->vramBaseAddress);
	templateHALData->width = info.width;
	templateHALData->height = info.height;
	templateHALData->rowBytes = info.rowBytes;
	
	*baseAddress = ((char *)templateHALData->baseAddress);
	if (kDepthMode1 == depthMode)
		*directColor = false;								// 8 bpp
	else
		*directColor = true;								// 16 or 32 bpp
		
	// hardware cursor buffer is no longer cleared
	templateHALData->hardwareCursorData.cursorCleared = false;


	err = kGDXErrNoError;										// Everything Okay
	
ErrorExit:

	return err;
}


//=====================================================================================================
//
// GraphicsHALDrawHardwareCursor()
//	This routine sets the cursor's X and Y coordinates and its visible state.  If the cursor was set
//	successfully by a previous call to GraphicsHALSetHardwareCursor(), then the HAL must program the
//	hardware with the given X, Y and visible state.  If the previous call to
//	GraphicsHALSetHardwareCursor() failed, then an error should be returned.
//
//		-> x		X coordinate
//		-> y		Y coordinate
//		-> visible	'true' if the cursor must be visible.
//
//=====================================================================================================
GDXErr  GraphicsHALDrawHardwareCursor(SInt32 x, SInt32 y, Boolean visible)
{
	TemplateHALData				*templateHALData = GraphicsHALGetHALData();
	UInt32						screenWidth, screenHeight;
	SpurHardwareCursorData		*pHardwareCursorData;
	Ptr							hardwareCursorImage;
	Ptr							pHardwareCursorFrameBuffer;
	UInt32						screenRowBytes, cursorRowBytes;
	Ptr							pScreenRowBase, pCursorRowBase;
	UInt32						*pCursorImage, *pScreenImage;
	UInt32						screenLine, upperCursorLine, lowerCursorLine,
								endMask;
	UInt32						upperLineShift, lowerLineShift;
	SInt32						width, height;
	UInt32						cursorStartX, cursorStartY;
	SInt32						currentX, currentY;
	SInt32						rowLongs;
	SInt32						i, j;

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	// Get data for hardware cursor
	pHardwareCursorData = &(templateHALData->hardwareCursorData);

	// Check if the hardware cursor buffer must be cleared.
	if (false == pHardwareCursorData->cursorCleared)
	{
		// set hardware cursor buffer area to transparent
		pCursorImage = (UInt32 *) (((Ptr) (templateHALData->baseAddress)) - kHardwareCursorOffset);
		cursorRowBytes = templateHALData->rowBytes;
		for (i = 0; i < templateHALData->height; i++)
		{
			for (j = 0; j < 4; j++)
				*pCursorImage++ = 0;
			pCursorImage = (UInt32 *) ((UInt32) pCursorImage + cursorRowBytes - kHardwareCursorOffset);
		}
		pHardwareCursorData->cursorCleared = true;
	}

	if (false == pHardwareCursorData->cursorSet)
	{
		*templateHALData->spur.address = kSpurControl;
		SynchronizeIO ();
		*templateHALData->spur.multiPort &= ~kSpurControlCursorEnableMask;
		SynchronizeIO ();
		templateHALData->hardwareCursorData.cursorVisible = false;
		return (kGDXErrNoHardwareCursorSet);
	}

	if (false == visible)
	{
		*(templateHALData->spur.address) = kSpurControl;
		SynchronizeIO ();
		*(templateHALData->spur.multiPort) &= ~kSpurControlCursorEnableMask;
		SynchronizeIO ();
		templateHALData->hardwareCursorData.cursorVisible = false;
		return (kGDXErrNoError);
	}

	// Check if we should draw the cursor or defer movement.
	if (false == pHardwareCursorData->cursorVisible)
	{
		currentX = pHardwareCursorData->x;
		currentY = pHardwareCursorData->y;
	
		// Calculate proper currentY and height. 
		screenWidth = templateHALData->width;
		screenHeight = templateHALData->height;
		height = 16;
	
		if (currentY < 0)
		{
			height -= -currentY;
			currentY = 0;
		}
		if ((currentY + height) > screenHeight)
			height = screenHeight - currentY;
	
		hardwareCursorImage = (Ptr) &(pHardwareCursorData->cursorImage);
		pHardwareCursorFrameBuffer = templateHALData->baseAddress - kHardwareCursorOffset;
		screenRowBytes = templateHALData->rowBytes;
	
		pScreenRowBase = (Ptr) pHardwareCursorFrameBuffer + currentY*screenRowBytes;
		for (i = 0; i < height; i++)
		{
			pScreenImage = (UInt32 *) pScreenRowBase;
			for (j = 0; j < 2; j++)
				*pScreenImage++ = 0;
		pScreenRowBase += screenRowBytes;
		}
		pHardwareCursorData->x = x;
		pHardwareCursorData->y = y;
	
		// Calculate proper width, height, cursorStartX, and cursorStartY.
		width = 16;
		height = 16;
		cursorStartX = 0;
		cursorStartY = 0;
	
		if (x < 0)
		{
			width -= -x;
			cursorStartX = -x;
			x = 0;
		}
		if ((x + width) > screenWidth)
			width = screenWidth - x;
	
		if (y < 0)
		{
			height -= -y;
			cursorStartY = -y;
			y = 0;
		}
		if ((y + height) > screenHeight)
			height = screenHeight - y;
	
		pScreenRowBase = (Ptr) pHardwareCursorFrameBuffer + y*screenRowBytes;
		cursorRowBytes = 8;
		pCursorRowBase = (Ptr) hardwareCursorImage + cursorStartY*cursorRowBytes +
								((cursorStartX >> 1) & 0xFFFFFFFC);
		rowLongs = (width + (cursorStartX & 0x07) - 1) >> 3;
		upperLineShift = (cursorStartX & 0x07) << 2;
		lowerLineShift = (32 - upperLineShift);
		endMask = (0xFFFFFFFF << (32 - ((((width - 1) & 0x07) + 1) << 2))) >> upperLineShift;
		for (i = 0; i < height; i++)
		{
			pScreenImage = (UInt32 *) pScreenRowBase;
			pCursorImage = (UInt32 *) pCursorRowBase;
			upperCursorLine = *pCursorImage++;
			for (j = 0; j < rowLongs; j++)
			{
				lowerCursorLine = *pCursorImage++;
				screenLine = (upperCursorLine << upperLineShift) |
								(lowerCursorLine >> lowerLineShift);
				*pScreenImage++ = screenLine;
				upperCursorLine = lowerCursorLine;
			}
			*pScreenImage++ = (upperCursorLine & endMask) << upperLineShift;
			pScreenRowBase += screenRowBytes;
			pCursorRowBase += cursorRowBytes;
		}
	
		*(templateHALData->spur.address) = kSpurCursorXPositionLow;
		SynchronizeIO ();
		*(templateHALData->spur.multiPort) = x & 0xff;
		SynchronizeIO ();
		*(templateHALData->spur.address) = kSpurCursorXPositionHigh;
		SynchronizeIO ();
		*(templateHALData->spur.multiPort) = (x & 0x0f00) >> 8;
		SynchronizeIO ();
	
		// Make cursor visible
		*(templateHALData->spur.address) = kSpurControl;
		SynchronizeIO ();
		*(templateHALData->spur.multiPort) |= 1 << kSpurControlCursorEnableBit;
		SynchronizeIO ();
		templateHALData->hardwareCursorData.cursorVisible = true;
	}
	else
	{
		pHardwareCursorData->deferredX = x;
		pHardwareCursorData->deferredY = y;
		pHardwareCursorData->deferredMove = true;
	}

	return (kGDXErrNoError);
}




#pragma options align=mac68k

struct HWCursorColorTable {
	SInt32			ctSeed;
	SInt16			ctFlags;
	SInt16			ctSize;
	ColorSpec		ctTable[kNumHardwareCursorColors];
};
typedef struct HWCursorColorTable HWCursorColorTable;

#pragma options align=reset



//=====================================================================================================
//
// GraphicsHALSetHardwareCursor()
//	This routine is called to setup the hardware cursor and determine if whether the hardware can
//	support it.  The HAL should remember whether this call was successful for subsequent
//	GetHardwareCursorDrawState() or DrawHardwareCursor() calls, but should NOT change the cursor's
//	X or Y coordinates, nor its visible state.
//	
//		-> gamma
//		Current gamma table to correct cursor colors with, if the HAL can apply gamma correction.
//
//		-> luminanceMapping
//		This will be true if the Core had luminance mapping enabled AND it was in an indexed color
//		mode.  If 'true', the HAL should luminance map the cursor CLUT EVEN if the hardware cursor is
//		a super-duper cursor capable of direct color. This is because the hardware cursor should look
//		like the software cursor it is replacing.
//
//		-> cursorRef
//		Opaque data to be handed to VSLPrepareCursorForHardwareCursor().
//
//=====================================================================================================
GDXErr  GraphicsHALSetHardwareCursor(const GammaTbl *gamma, Boolean luminanceMapping, void *cursorRef)
{
	TemplateHALData				*templateHALData = GraphicsHALGetHALData();
	AbsoluteTime nsCLUTAddrRegDelay = templateHALData->nsCLUTAddrRegDelay;				// 800ns in absolute time to hit clut
	HardwareCursorInfoRec		hardwareCursorInfo;
	HWCursorColorTable			colorMap;
	SpurHardwareCursorData		*pHardwareCursorData;
	Ptr							hardwareCursorImage;
	UInt32						bitsPerPixel;
	Boolean						cursorVisible, cursorSet;
	SInt32						cursorX, cursorY;
	UInt32						i;

	SInt16 						dataWidth;
	UInt32 						redIndex;
	UInt32 						greenIndex;
	UInt32 						blueIndex;
	
	UInt8 						*redCorrection;			// corrected red data in gamma
	UInt8 						*greenCorrection;		// corrected green data in gamma
	UInt8 						*blueCorrection;		// corrected blue data in gamma

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	pHardwareCursorData = &(templateHALData->hardwareCursorData);
	hardwareCursorImage = (Ptr) &(pHardwareCursorData->cursorImage);

	// Get cursor state.
	cursorVisible = pHardwareCursorData->cursorVisible;
	cursorSet = pHardwareCursorData->cursorSet;
	cursorX = pHardwareCursorData->x;
	cursorY = pHardwareCursorData->y;

	hardwareCursorInfo.majorVersion = 0;
	hardwareCursorInfo.minorVersion = 0;
	hardwareCursorInfo.cursorHeight = 0;
	hardwareCursorInfo.cursorWidth = 0;
	hardwareCursorInfo.colorMap = (CTabPtr) &colorMap;
	hardwareCursorInfo.hardwareCursor = hardwareCursorImage;
	for (i = 0; i < 6; i++)
		hardwareCursorInfo.reserved[i] = 0;

	pHardwareCursorData->cursorSet =
		VSLPrepareCursorForHardwareCursor
		(cursorRef, &(pHardwareCursorData->cursorDescriptor), &hardwareCursorInfo);

	if (false == pHardwareCursorData->cursorSet)
	{
		GraphicsHALDrawHardwareCursor (cursorX, cursorY, false);
		return (kGDXErrCannotRenderCursorImage);
	}


	if (luminanceMapping)
	{

		// Convert the RGB colors into luminance mapped gray scale.
		// For those of us familiar with color space theory, 
		// 		
		//		Luminance = .299Red + .587Green + .114Blue
		//		("Video Demystified" by Keith Jack, page 28)
		//
		// Conveniently, on the PowerPC architechture, floating point math is FASTER
		// than integer math, so we will do outright floating point and not even
		// think about playing games with Fixed Point math.
		
		double redPortion;					// Luminance portion from red component
		double greenPortion;				// Luminance portion from green component
		double bluePortion;					// Luminance portion from blue component
		double luminance;					// Resulting luminosity
		
		for (i = 0 ; i < 8 ; i++)
		{
			
			redPortion = 0.299 * colorMap.ctTable[i].rgb.red;
			greenPortion = 0.587 * colorMap.ctTable[i].rgb.green;
			bluePortion = 0.114 * colorMap.ctTable[i].rgb.blue;
			
			luminance = redPortion + greenPortion + bluePortion;
			
			colorMap.ctTable[i].rgb.red = luminance;
			colorMap.ctTable[i].rgb.green = luminance;
			colorMap.ctTable[i].rgb.blue = luminance;
			
		}
	}
	
	
	// Apply gamma correction to the 'colorMap.ctTable'
	
	dataWidth = gamma->gDataWidth;
	redCorrection = (UInt8 *) &gamma->gFormulaData + gamma->gFormulaSize;
	
	if (1 == gamma->gChanCnt)
	{
		// Use same correction data for all three channels
		greenCorrection = redCorrection;
		blueCorrection = redCorrection;
	} else {
		// Each channel has its own correction data
		greenCorrection = redCorrection + gamma->gDataCnt;
		blueCorrection = redCorrection + ( 2 * gamma->gDataCnt);
	}
	
	for (i = 0 ; i < 8 ; i++)
	{                                                                         
		// Extract the most significant 'dataWidth' amount of bits from each color
		// to use as the index into the correction data
		
		redIndex = colorMap.ctTable[i].rgb.red >> (16 - dataWidth);
		greenIndex = colorMap.ctTable[i].rgb.green >> (16 - dataWidth);
		blueIndex = colorMap.ctTable[i].rgb.blue >> (16 - dataWidth);
		
		colorMap.ctTable[i].rgb.red = *(redCorrection + redIndex);
		colorMap.ctTable[i].rgb.green = *(greenCorrection + greenIndex);
		colorMap.ctTable[i].rgb.blue = *(blueCorrection + blueIndex);
	}


	// Make cursor invisible if it was previously visible and set.
	if (cursorSet && cursorVisible)
		GraphicsHALDrawHardwareCursor (cursorX, cursorY, false);

	*(templateHALData->spur.address) = 0x00;
	SynchronizeIO();
	for (i = 0; i < 8; i++)
	{
		(void) DelayForHardware(nsCLUTAddrRegDelay);
		*(templateHALData->spur.cursorPaletteRAM) = (colorMap.ctTable[i].rgb.red);
		SynchronizeIO ();
		*(templateHALData->spur.cursorPaletteRAM) = (colorMap.ctTable[i].rgb.green);
		SynchronizeIO ();
		*(templateHALData->spur.cursorPaletteRAM) = (colorMap.ctTable[i].rgb.blue);
		SynchronizeIO ();
	}

	// Make cursor visible if it was previously visible and set.
	if (cursorSet && cursorVisible)
		GraphicsHALDrawHardwareCursor (cursorX, cursorY, true);

	return (kGDXErrNoError);
}



//=====================================================================================================
//
// GraphicsHALSupportsHardwareCursor()
//	This call is used to determine if the HAL supports a hardware cursor.
//
//		<- supportsHardwareCursor	'true' if supports a hardware cursor, 'false' otherwise.
//
//=====================================================================================================
GDXErr GraphicsHALSupportsHardwareCursor(Boolean *supportsHardwareCursor)
{
  *supportsHardwareCursor = true;

  return (kGDXErrNoError);
}



//=====================================================================================================
//
// GraphicsHALGetHardwareCursorDrawState()
//	This routine is used to determine the state of the hardware cursor.  After HAL initialization
//	the cursors visible state and set state should be false. After a mode change the cursor should be
//	made invisible but the set state should remain unchanged.
//
//		<- csCursorX		X coordinate from last DrawHardwareCursor call
//		<- csCursorY		Y coordinate from last DrawHardwareCursor call
//		<- csCursorVisible	'true' if the cursor is visible
//		<- csCursorSet		'true' if last GraphicsHALSetHardwareCursor() call was successful.
//
//=====================================================================================================
GDXErr GraphicsHALGetHardwareCursorDrawState(SInt32  *pCursorX, SInt32  *pCursorY,
											 UInt32  *pCursorVisible, UInt32  *pCursorSet)
{
	TemplateHALData				*templateHALData = GraphicsHALGetHALData();
	SpurHardwareCursorData	*pHardwareCursorData;

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	// Get data for hardware cursor
	pHardwareCursorData = &(templateHALData->hardwareCursorData);

	// Check if cursor is set
	if (pHardwareCursorData->cursorSet)
	{
		*pCursorX = pHardwareCursorData->x;
		*pCursorY = pHardwareCursorData->y;
		*pCursorVisible = pHardwareCursorData->cursorVisible;
		*pCursorSet = true;
	}
	else
	{
		*pCursorX = 0;
		*pCursorY = 0;
		*pCursorVisible = false;
		*pCursorSet = false;
	}

	return (kGDXErrNoError);
}



//=====================================================================================================
//
// DeferredMoveHardwareCursor()
//	This is mainly called by the VBL interrupt handler to minimize the latency with moving
// the cursor.  This minimizes the effect of moving the cursor when it intersects the
// scan line.  Since this procedure writes to the Spur address register, this procedure
// must be called by any routine that wants to use the Spur address register so that the
// routine doesn't get interrupted by DeferredMoveHardwareCursor.  The only routine that
// defers drawing is GraphicsHALDrawHardwareCursor.
//
//		-> void
//
//=====================================================================================================
static void  DeferredMoveHardwareCursor(void)
{
	TemplateHALData				*templateHALData = GraphicsHALGetHALData();
	UInt32						screenWidth, screenHeight;
	SpurHardwareCursorData		*pHardwareCursorData;
	Ptr							hardwareCursorImage;
	Ptr							pHardwareCursorFrameBuffer;
	UInt32						screenRowBytes, cursorRowBytes;
	Ptr							pScreenRowBase, pCursorRowBase;
	UInt32						*pCursorImage, *pScreenImage;
	UInt32						screenLine, upperCursorLine, lowerCursorLine, endMask;
	UInt32						upperLineShift, lowerLineShift;
	SInt32						width, height;
	UInt32						cursorStartX, cursorStartY;
	SInt32						currentX, currentY;
	SInt32						x, y;
	SInt32						rowLongs;
	SInt32						deferredMove;
	SInt32						i, j;

	// Get data for hardware cursor.
	pHardwareCursorData = &(templateHALData->hardwareCursorData);

	// Check if we should move or not.
	deferredMove = pHardwareCursorData->deferredMove;
	while (!CompareAndSwap (deferredMove, false, (UInt32 *) &(pHardwareCursorData->deferredMove)))
		deferredMove = pHardwareCursorData->deferredMove;
	if (false == deferredMove)
		return;

	currentX = pHardwareCursorData->x;
	currentY = pHardwareCursorData->y;

	x = pHardwareCursorData->deferredX;
	y = pHardwareCursorData->deferredY;

	// Calculate proper currentY and height. 
	screenWidth = templateHALData->width;
	screenHeight = templateHALData->height;
	height = 16;

	if (currentY < 0)
	{
		height -= -currentY;
		currentY = 0;
	}
	if ((currentY + height) > screenHeight)
		height = screenHeight - currentY;

	hardwareCursorImage = (Ptr) &(pHardwareCursorData->cursorImage);
	pHardwareCursorFrameBuffer = templateHALData->baseAddress - kHardwareCursorOffset;
	screenRowBytes = templateHALData->rowBytes;

	pScreenRowBase = (Ptr) pHardwareCursorFrameBuffer + currentY*screenRowBytes;
	for (i = 0; i < height; i++)
	{
		pScreenImage = (UInt32 *) pScreenRowBase;
		for (j = 0; j < 2; j++)
			*pScreenImage++ = 0;
		pScreenRowBase += screenRowBytes;
	}
	pHardwareCursorData->x = x;
	pHardwareCursorData->y = y;

	// Calculate proper width, height, cursorStartX, and cursorStartY.
	width = 16;
	height = 16;
	cursorStartX = 0;
	cursorStartY = 0;

	if (x < 0)
	{
		width -= -x;
		cursorStartX = -x;
		x = 0;
	}
	if ((x + width) > screenWidth)
		width = screenWidth - x;

	if (y < 0)
	{
		height -= -y;
		cursorStartY = -y;
		y = 0;
	}
	if ((y + height) > screenHeight)
		height = screenHeight - y;

	pScreenRowBase = (Ptr) pHardwareCursorFrameBuffer + y*screenRowBytes;
	cursorRowBytes = 8;
	pCursorRowBase = (Ptr) hardwareCursorImage + cursorStartY*cursorRowBytes +
							((cursorStartX >> 1) & 0xFFFFFFFC);
	rowLongs = (width + (cursorStartX & 0x07) - 1) >> 3;
    upperLineShift = (cursorStartX & 0x07) << 2;
	lowerLineShift = (32 - upperLineShift);
	endMask = (0xFFFFFFFF << (32 - ((((width - 1) & 0x07) + 1) << 2))) >> upperLineShift;
	for (i = 0; i < height; i++)
	{
		pScreenImage = (UInt32 *) pScreenRowBase;
		pCursorImage = (UInt32 *) pCursorRowBase;
		upperCursorLine = *pCursorImage++;
		for (j = 0; j < rowLongs; j++)
		{
			lowerCursorLine = *pCursorImage++;
			screenLine = (upperCursorLine << upperLineShift) |
							(lowerCursorLine >> lowerLineShift);
			*pScreenImage++ = screenLine;
			upperCursorLine = lowerCursorLine;
		}
		*pScreenImage++ = (upperCursorLine & endMask) << upperLineShift;
		pScreenRowBase += screenRowBytes;
		pCursorRowBase += cursorRowBytes;
	}

	*(templateHALData->spur.address) = kSpurCursorXPositionLow;
	SynchronizeIO ();
	*(templateHALData->spur.multiPort) = x & 0xff;
	SynchronizeIO ();
	*(templateHALData->spur.address) = kSpurCursorXPositionHigh;
	SynchronizeIO ();
	*(templateHALData->spur.multiPort) = (x & 0x0f00) >> 8;
	SynchronizeIO ();
}



//=====================================================================================================
//
// GraphicsHALSetPowerState()
//	Turn off VRAM refresh in Toynbee and turn off the PLL and the CLUT in Spur if going to kAVPowerOff
//
//	For this routine, the relevant fields indicated by 'VDPowerStateRec' are:
//			->	powerState		desired power mode: kAVPowerOff, kAVPowerStandby, kAVPowerSuspend,
//								kAVPowerOn
//
//			<- powerFlags		kPowerStateNeedsRefresh bit set if hw needs to be refreshed after
//								coming out of the designated power state.
//
//=====================================================================================================
GDXErr GraphicsHALSetPowerState(VDPowerStateRec *vdPowerState)
{

	UInt32 i;									// cluty loop iterator
	UInt32 numberOfEntries;						// clut will have 31 or 255 entries (0 based)
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	MrSanAntonioRegisters *mrSanAntonio = &templateHALData->mrSanAntonio;
	SpurRegisters *spur = &templateHALData->spur;
	UInt16 cvhSyncDisabled = templateHALData->cvhSyncDisabled;						
	AbsoluteTime nsCLUTAddrRegDelay = templateHALData->nsCLUTAddrRegDelay;				// 800ns in absolute time to hit clut

	GDXErr err = kGDXErrUnknownError;								// Assume failure

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	vdPowerState->powerFlags = 0;									// clear the flags.

	if (kAVPowerOff == vdPowerState->powerState)
	{
		UInt8 phaseLockLoopControl;				// hit PhaseLockLoopControl to power down clut

		vdPowerState->powerFlags = 1 << kPowerStateNeedsRefresh; 	// power off, so refresh needed.
		if (templateHALData->clutOff)								// clut is already off
		{
			err = kGDXErrNoError;									// Everything okay
			goto ErrorExit;
		}
	
		// clut needs to be saved and turned off since we are about to shut down power

		if (kDepthMode2 == templateHALData->depthMode)
			numberOfEntries = 31;
		else
			numberOfEntries = 255;
			
		*spur->address = 0;
		SynchronizeIO();

		// read the clut.  Each time the address register is hit, delay 800 ns since the
		// clut is retrieving data at the new address.  The address register autoincrements
		// after blue is written, hence the delay
		for (i = 0 ; i <=  numberOfEntries; i++)
		{
			(void) DelayForHardware(nsCLUTAddrRegDelay);
			templateHALData->savedClut[i].red = *spur->colorPaletteRAM;
			SynchronizeIO();
			templateHALData->savedClut[i].green = *spur->colorPaletteRAM;
			SynchronizeIO();
			templateHALData->savedClut[i].blue = *spur->colorPaletteRAM;
			SynchronizeIO();
		}

		*templateHALData->toynbee.reset = 0x6;			// toynbee in reset, video refresh off
		
		*templateHALData->spur.address = kSpurPhaseLockLoopControl;	// DAC and PLL address
		SynchronizeIO();

		phaseLockLoopControl = *templateHALData->spur.multiPort;
		// set DAC,PLL Power bits to power down the systems
		phaseLockLoopControl |= kSpurPhaseLockLoopControlDACPLLPowerMask;
		*templateHALData->spur.multiPort = phaseLockLoopControl;
		
		// clut is now off
		templateHALData->clutOff = true;
		templateHALData->toynbeeRunning = false;
	}
	else if (templateHALData->clutOff)					// switching to NON off power state
	{													// so turn the hw back on
	
		UInt8 phaseLockLoopControl;						// hit PhaseLockLoopControl to power up clut
	
		*templateHALData->spur.address = kSpurPhaseLockLoopControl;	// DAC and PLL address	
		SynchronizeIO();

		phaseLockLoopControl = *templateHALData->spur.multiPort;
		// clear DAC,PLL Power bits to power up the systems
		// preserve ClKorPLL bit and SetSelect bit
		phaseLockLoopControl &= kSpurPhaseLockLoopControlCLKorPLLSetMask;	
		*templateHALData->spur.multiPort = phaseLockLoopControl;
		
		// clut is now on
		templateHALData->clutOff = false;
		
		// clut needs time to stablize once PLL is turned back on.
		// Spur is spec for a 10 ms delay before the clut can be used once PLL is turned on
		MoreRealDelayFor(10*durationMillisecond); 				

		// put Toynbee, VRAM refresh and VRAM state machine in reset
		(void) TemplateAssertVideoReset(&templateHALData->toynbee);	
		MoreRealDelayFor(1*durationMicrosecond); 				

		// turn on Toynbee, VRAM refresh and VRAM state machine
		(void) TemplateReleaseVideoReset(&templateHALData->toynbee);

					
		if (kDepthMode2 == templateHALData->depthMode)
			numberOfEntries = 31;
		else
			numberOfEntries = 255;
			
		*spur->address = 0;
		SynchronizeIO();
	
		// restore the clut.  Each time the address register is hit, delay 800 ns since the
		// clut is retrieving data at the new address.  The address register autoincrements
		// after blue is written, hence the delay
		for (i = 0 ; i <=  numberOfEntries; i++)
		{			
			(void) DelayForHardware(nsCLUTAddrRegDelay);
			*spur->colorPaletteRAM = templateHALData->savedClut[i].red;
			SynchronizeIO();
			*spur->colorPaletteRAM = templateHALData->savedClut[i].green;
			SynchronizeIO();
			*spur->colorPaletteRAM = templateHALData->savedClut[i].blue;
			SynchronizeIO();

		}
		
	}
		
	templateHALData->vdPowerState = vdPowerState->powerState;	

	err = kGDXErrNoError;								// Everything okay

ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALGetPowerState()
//	The graphics hw might have the ability to to go into some kind of power saving mode.  Just
//	pass the call to the HAL
//
//	For this routine, the relevant fields indicated by 'VDPowerStateRec' are:
//			<- powerState		current power mode: kAVPowerOff, kAVPowerStandby, kAVPowerSuspend,
//								kAVPowerOn
//
//			<- powerFlags		kPowerStateNeedsRefresh bit set if hw needs to be refreshed after
//								coming out of power state
//
//=====================================================================================================
GDXErr GraphicsHALGetPowerState(VDPowerStateRec *vdPowerState)
{

	TemplateHALData *templateHALData = GraphicsHALGetHALData();

	GDXErr err = kGDXErrNoError;							// Everything okay
	
	vdPowerState->powerFlags = 0;
	vdPowerState->powerState = templateHALData->vdPowerState;

	// VRAM has decayed if the current powerState is kAVPowerOff
	if ( kAVPowerOff == templateHALData->vdPowerState)
		vdPowerState->powerFlags = 1 << kPowerStateNeedsRefresh;


ErrorExit:

	return err;
}



//=====================================================================================================
//
// GraphicsHALPrivateControl()
//	Routine accepts private control calls.  The core passes all control codes that it doesn't
//	understand to this routine.  If the HAL knows what to do with the code, it hands it off
//	accordingly.
//	
//		-> genericPtr			Points to the data structure that the HAL needs for this control
//								call.  Should be cast to appropriate data type if internal routine is
//								invoked.
//		-> privateControlCode	The private csCode that the HAL might know what to do with.
//
//=====================================================================================================
OSErr GraphicsHALPrivateControl(void * genericPtr, SInt16 privateControlCode)
{


	GDXErr err = kGDXErrUnknownError;				// Assume failure.
	OSErr returnErr = controlErr;					// Assume cs code is invalid

	switch (privateControlCode)
	{
		default:
			goto ErrorExit;							// csCode not supported
	}
	
ErrorExit:
	return returnErr;
}



//=====================================================================================================
//
// GraphicsHALPrivateStatus()
//	Routine accepts private status calls.  The core passes all status codes that it doesn't
//	understand to this routine.  If the HAL knows what to do with the code, it hands it off
//	accordingly.
//	
//		-> genericPtr			Points to the data structure that the HAL needs for this status
//								call.  Should be cast to appropriate data type if internal routine is
//								invoked.
//		-> privateStatusCode	The private csCode that the HAL might know what to do with.
//
//=====================================================================================================
OSErr GraphicsHALPrivateStatus(void * genericPtr, SInt16 privateStatusCode)
{

			

	GDXErr err = kGDXErrUnknownError;				// Assume failure.
	OSErr returnErr = controlErr;					// Assume cs code is invalid

	switch (privateStatusCode)
	{
		default:
			goto ErrorExit;							// csCode not supported
	}
	
ErrorExit:
	return returnErr;
}



//=====================================================================================================
//
// TemplateMapDepthModeToCLUTAttributes()
//
//	This simple routine maps a 'DepthMode' to the corresponding CLUT attritributes associate with it.
//	
//		-> depthMode	The 'DepthMode' for which the CLUT attributes are desired.
//		<- startAddress	The physical address that corresponds to logical address 0.
//		<- entryOffset	The physical offset between each logical address.
//
//=====================================================================================================
static GDXErr TemplateMapDepthModeToCLUTAttributes(DepthMode depthMode,
				UInt32 *startAddress, UInt32 *entryOffset)
{
	// Define a new type which maps a 'DepthMode' to the corresponding physical attributes of a CLUT.
	// These attributes are needed because the logical addresses when in 1, 2, or 4 bpp do not
	// correspond to the physical address.  Instead, they are distributed evenly through out the CLUT
	// address space.
	
	typedef struct DepthModeToCLUTAttributesMap DepthModeToCLUTAttributesMap;
	struct DepthModeToCLUTAttributesMap
	{
		DepthMode depthMode;			// DepthMode for which these attributes apply
		UInt32 startAddress;		// Physical address of logical address 0 
		UInt32 entryOffset;		// physical offset between each logical address
	};

	#define kMapSize 3
	
	DepthModeToCLUTAttributesMap depthModeMap[kMapSize] =
	{
		{kDepthMode1, 0x00, 0x01},						// 8 bpp
		{kDepthMode2, 0x00, 0x01},						// 16 bpp
		{kDepthMode3, 0x00, 0x01}						// 32 bpp
	};

	UInt32 i;									// Loop control variable

	GDXErr err = kGDXErrDepthModeUnsupported;					// Assume invalid depthmode
	// Scan the 'DepthModeToCLUTAttributesMap' to find CLUT attributes for the given depth mode.
	
	
	for (i = 0 ; i < kMapSize ; i++)
	{
		if (depthModeMap[i].depthMode == depthMode)
		{
			*startAddress = depthModeMap[i].startAddress;
			*entryOffset = depthModeMap[i].entryOffset;
			err = kGDXErrNoError;
			break;
		}
	}

ErrorExit:
	
	return err;
}



//=====================================================================================================
//
// TemplateWaitForVBL()
//	This routine does not return until it senses that a VBL has occurred.
//	For the Template class of machines, VBL information is in 3 registers
//	interruptMask, interruptStatus, clearCursorInterrupt.
//	For the Template of machines the hardware operates in the follow way:
//	
//	Internal Template interrupts are always on.  Just watch the interruptStatus register 
//	and see when the kCursorInterruptStatusBit is set.
//
//	This routine does not make any assumptions about whether external VBL interrupts are
//	enabled or disabled prior to its calling.  However, it is recommended that the
//	external VBL interrupts be disabled prior to calling this if any type of accurate waiting
//	takes place.
//
//=====================================================================================================
static void TemplateWaitForVBL(void)
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	HWRegister32Bit *interruptStatus = templateHALData->mrSanAntonio.interruptStatus;
	UInt32 status;

	if (templateHALData->toynbeeRunning)
	{
		status = *templateHALData->mrSanAntonio.clearCursorInterrupt;
		SynchronizeIO();
	
		while (true)
		{
			status = *interruptStatus;
			if ( status & kCursorInterruptStatusMask)
			break;
	
		}
	}
}



//=====================================================================================================
//
// TemplateClearInternalVBLInterrupts()
//	This routine clears the external interrupt source in GrandCentral.  This routine
//	is the interrupt Handler that is passed to the OSS.
//	Register clearCursorInterrupt in Template is accessed to clear the interrupt.
//
//=====================================================================================================
static void TemplateClearInternalVBLInterrupts(void* vblRefCon)
{
	UInt32 interrupts;
	interrupts = *(HWRegister32Bit*)0x44443330;					// clear Template interrupt
	SynchronizeIO();
	
	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();
}



//=====================================================================================================
//
// TemplateAssertVideoReset()
//	In the Template graphics architecture, the frame buffer controller (Toynbee) needs to be put into reset
// prior to programing.
//
//		-> toynbee	pointer to structure of register address for Toynbee
//
//=====================================================================================================
static GDXErr TemplateAssertVideoReset(const ToynbeeRegisters *toynbee)
{	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();

	*toynbee->reset = 0x7;						// Put the puppy in reset.
	templateHALData->toynbeeRunning = false;
	
	return (kGDXErrNoError);
}



//=====================================================================================================
//
// TemplateSetupClockGenerator()
//	This will program the video clock generator for the Template graphics architecture.
//
//		-> spur			Pointer to structure of register addresses for Spur
//		-> bdiData		The hardware programming parameters that DO NOT vary with bit depth
//
//		<> usingClockSetA
//		As an input, this is 'true' if the current active set of the Frankenstein clock generator is
//		Set A, 'false' if Set B is in use.
//		As an output, this is changed to reflect the current set in use.
//
//=====================================================================================================
static GDXErr TemplateSetupClockGenerator(const SpurRegisters *spur,
		const BitDepthIndependentData *bdiData, Boolean *usingClockSetA)
{
	UInt8 phaseLockLoopControl;				// program the PhaseLockLoopControl reg to use new set
	UInt8 m;
	UInt8 pn;
	UInt8 ri;

	UInt8 mAddress;
	UInt8 pnAddress;
	UInt8 riAddress;
	
	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	m = bdiData->frankM;
	pn = bdiData->frankPN;
	ri = bdiData->frankRI;
	
	// Figure out if what set we are currently using, so we can program the other set.
	if (*usingClockSetA)
	{	// We are currently using Set A, so we want to get the addresses to program Set B
		mAddress = kSpurMSetB;
		pnAddress = kSpurPNSetB;
		riAddress = kSpurRISetB;
	}
	else
	{	// We are currently using Set B, so we want to get the addresses to program Set A
		mAddress = kSpurMSetA;
		pnAddress = kSpurPNSetA;
		riAddress = kSpurRISetA;
	}
	
	// Program the clock generator
	
	*spur->address = mAddress;			// Setup the address register to the appropriate M
	SynchronizeIO();
	*spur->multiPort = m;				// Program M
	SynchronizeIO();

	*spur->address = pnAddress;			// Setup the address register to the appropriate PN
	SynchronizeIO();
	*spur->multiPort = pn;				// Program PN
	SynchronizeIO();

	*spur->address = riAddress;		// Setup the address register to the appropriate RI
	SynchronizeIO();
	*spur->multiPort = ri;			// Program RI
	SynchronizeIO();

	*usingClockSetA = !(*usingClockSetA);	// Toggle the currently active set
	
	// based on the new set we are using, whack the PhaseLockLoopControl register to use new set
	// If SetA, SetSelect = 0, if SetB, SetSelect = 1

	*spur->address = kSpurPhaseLockLoopControl;
	SynchronizeIO();
	phaseLockLoopControl = *spur->multiPort;
	
	phaseLockLoopControl &= ~kSpurPhaseLockLoopControlSetSelectMask; 		// clear SetSelect bit
	
	if (*usingClockSetA)
		phaseLockLoopControl |= (0 << kSpurPhaseLockLoopControlSetSelectBit);
	else
		phaseLockLoopControl |= (1 << kSpurPhaseLockLoopControlSetSelectBit);

	*spur->multiPort = phaseLockLoopControl;			// Now using new Set
	
	return (kGDXErrNoError);
}



//=====================================================================================================
//
// TemplateSetupCLUT()
//	This will program the CLUT for the Template graphics architecture.
//
//		-> spur				Pointer to structure of register addresses for Spur
//		-> bddData			The hardware programming parameters that DO vary with bit depth
//
//=====================================================================================================
static GDXErr TemplateSetupCLUT(const SpurRegisters *spur, const WidthAndDepthDependentData *vwdData)
{
	
	UInt8 control;									// don't hit cursor enable
	UInt8 newControl = vwdData->spurControl;		// control for new depthMode. 

	// Flush out any deferred cursor moving.
	DeferredMoveHardwareCursor();

	newControl &= ~kSpurControlCursorEnableMask;	// make sure bddData doesn't hit cursor enable
	
	*spur->address = kSpurControl;					// Setup the address register for 'control'
	SynchronizeIO();

	control = *spur->multiPort;
	control &= kSpurControlCursorEnableMask;		// clear everything except cursor enable
	
	control |= newControl;						
	*spur->multiPort = control;


	return (kGDXErrNoError);
}



//=====================================================================================================
//
// TemplateSetupFBController()
//	This will program the frame buffer for the Template graphics architecture.
//
//		-> resolutionChange	true if bdi registers should be hit
//		-> toynbee			pointer to structure of register address for Toynbee
//		-> mrSanAntonio		pointer to structure of register address for MrSanAntonio
//		-> bdiData			The hardware programming parameters that DO NOT vary with bit depth
//		-> bddData			The hardware programming parameters that DO vary with bit depth
//		-> vwData			The hardware programming parameters that DO vary VRAM width
//		-> vwdData			The hardware programming parameters that DO vary VRAM width AND bit depth
//
//=====================================================================================================
static GDXErr TemplateSetupFBController(Boolean resolutionChange, const ToynbeeRegisters *toynbee,
		const MrSanAntonioRegisters *mrSanAntonio, const BitDepthIndependentData *bdiData,
		const BitDepthDependentData *bddData, const VRAMWidthDependentData *vwData,
		const WidthAndDepthDependentData *vwdData)
{	
	
	HWRegister32Bit systemConfiguration;	
	HWRegister32Bit mrSanAntonioMode;
	if (resolutionChange)
	{
		// Program MrSanAntonio with the 'BitDepthIndependentData' 
		mrSanAntonioMode = *mrSanAntonio->mrSanAntonioMode;
		mrSanAntonioMode &= ~kVHActiveHighMask;							// clear polarity of V & H Sync
		mrSanAntonioMode |= bdiData->vhSyncPolarity;
		*mrSanAntonio->mrSanAntonioMode = mrSanAntonioMode;


		*mrSanAntonio->verticalHalfLine = bdiData->verticalHalfLine;
		*mrSanAntonio->verticalSync = bdiData->verticalSync;
		*mrSanAntonio->verticalBackPorchEqualization = bdiData->verticalBackPorchEqualization;
		*mrSanAntonio->verticalBackPorch = bdiData->verticalBackPorch;
		*mrSanAntonio->verticalActiveLine = bdiData->verticalActiveLine;
		*mrSanAntonio->verticalFrontPorch = bdiData->verticalFrontPorch;
		*mrSanAntonio->verticalFrontPorchEqualization = bdiData->verticalFrontPorchEqualization;
	
		*mrSanAntonio->horizontalSerration = bdiData->horizontalSerration;
		*mrSanAntonio->halfLine = bdiData->halfLine;
		*mrSanAntonio->horizontalEqualization = bdiData->horizontalEqualization;
		*mrSanAntonio->horizontalSyncPulse = bdiData->horizontalSyncPulse;
		*mrSanAntonio->horizontalBreezeway = bdiData->horizontalBreezeway;
		*mrSanAntonio->horizontalBurstGate = bdiData->horizontalBurstGate;
		*mrSanAntonio->horizontalBackPorch = bdiData->horizontalBackPorch;
		*mrSanAntonio->horizontalActiveLine = bdiData->horizontalActiveLine;
		*mrSanAntonio->horizontalFrontPorch = bdiData->horizontalFrontPorch;
		*mrSanAntonio->horiztonalPixelCount = bdiData->horiztonalPixelCount;
	}
	// Program MrSanAntonio with the 'WidthAndDepthDependentData' 
	*mrSanAntonio->timingAdjust = vwdData->timingAdjust;

	// Program Toynbee with the 'BitDepthIndependentData'... (could change...always hit)
	*toynbee->baseAddress = bdiData->baseAddress;
	
	// Program Toynbee with the 'VRAMWidthDependentData' 
	*toynbee->frameBufferConfiguration1 = vwData->fbConfiguration1;
		
	// Program Toynbee with the 'BitDepthDependentData' 
	*toynbee->rowWords = bddData->rowWords;

	// Program Toynbee with the 'WidthAndDepthDependentData' 
	*toynbee->clockConfiguration = vwdData->clockConfiguration;
	
	return kGDXErrNoError;
}



//=====================================================================================================
// TemplateReleaseVideoReset()
//	In the Template graphics architecture, the frame buffer  controller (Toynbee) needs to be put through a
//	release-reset-release process.  Starting from a fully reset state, the steps are as follows:
//
//		release Toynbee from reset
//		reassert Toynbee reset
//		release Toynbee from reset
//		release VRAM state machine from reset
//		release video refresh state machine from reset
//
//		-> toynbee	pointer to structure of register address for Toynbee
//
//=====================================================================================================
static GDXErr TemplateReleaseVideoReset(const ToynbeeRegisters *toynbee)
{	
	TemplateHALData *templateHALData = GraphicsHALGetHALData();

	*toynbee->reset = 0x43;						// Release Toynbee from reset
	SynchronizeIO();
	
	*toynbee->reset = 0x68;						// Reassert Toynbee reset
	SynchronizeIO();

	*toynbee->reset = 0x43;						// Release Toynbee from reset
	SynchronizeIO();
	
	*toynbee->reset = 0x93;						// Release VRAM state machine from reset
	SynchronizeIO();
	
	*toynbee->reset = 0x00;						// Release video refresh state machine from reset
	SynchronizeIO();

	templateHALData->toynbeeRunning = true;

	return (kGDXErrNoError);
}



//=====================================================================================================
//
// TemplateGetDisplayData()
//	This routine returns all the data necessary to program the graphics hardware to the desired
//	'displayModeID', 'depthMode', and 'page', based on the the system configuration passed in.
//
//		-> displayModeID	get information for this resolution
//		-> depthMode		get information for this depthMode
//		-> vramSize			VRAM in the system (1, 2 or 4 megs)
//		-> vramWidth32		boolean set to true if VRAM width is 32 bits wide
//
//		<- bdiData		The hardware programming parameters that DO NOT vary with bit depth
//		<- bddData		The hardware programming parameters that DO vary with bit depth
//		<- vwData		The hardware programming parameters that DO vary VRAM width
//		<- vwdData		The hardware programming parameters that DO vary VRAM width AND bit depth
//		<- info			various tidbits of information:
//			the relevant fields of the 'DisplayInfo' structure are as follows:
//
//		<- width			horizontal pixels for the given displayModeID
//		<- height			vertical pixels for the given displayModeID
//		<- maxDepthMode		max depthMode based on inputs.  if inputs can't be supported,
//							maxDepthMode = kDepthModeInvalid
//		<- rowBytes			rowBytes reported in get video parameters
//
//=====================================================================================================
GDXErr TemplateGetDisplayData(DisplayModeID displayModeID, DepthMode depthMode, VRAMSize vramSize,
	Boolean vramWidth32, BitDepthIndependentData *bdiData, 
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData,
	DisplayInfo *info)
{

	InfoIndicies infoIndicies;
	
	GDXErr err = kGDXErrUnknownError;								// Assume failure
	
	err = TemplateMapToIndices(depthMode, vramSize, vramWidth32, &infoIndicies);
	if (err)
		goto ErrorExit;

	// Get the 'bdiData' and 'bddData' for the requested 'DisplayModeID' and 'DepthMode'
	switch (displayModeID)
	{

		case kDisplay512x384At60HzNTSC:
			err = TemplateGet512x384At60HzNTSC(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;

		case kDisplay512x384At60Hz:
			err = TemplateGet512x384At60Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;

		case kDisplay640x480At50HzPAL:
			err = TemplateGet640x480At50HzPAL(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay640x480At60HzNTSC:
			err = TemplateGet640x480At60HzNTSC(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;

		case kDisplay640x480At60HzVGA:
			err = TemplateGet640x480At60HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay640x480At67Hz:
			err = TemplateGet640x480At67Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay640x870At75Hz:
			err = TemplateGet640x870At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay768x576At50HzPAL:
			err = TemplateGet768x576At50HzPAL(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay800x600At60HzVGA:
			err = TemplateGet800x600At60HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay800x600At72HzVGA:
			err = TemplateGet800x600At72HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay800x600At75HzVGA:
			err = TemplateGet800x600At75HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay832x624At75Hz:
			err = TemplateGet832x624At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay1024x768At60HzVGA:
			err = TemplateGet1024x768At60HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay1024x768At72HzVGA:
			err = TemplateGet1024x768At72HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;

		case kDisplay1024x768At75HzVGA:	
			err = TemplateGet1024x768At75HzVGA(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay1024x768At75Hz:
			err = TemplateGet1024x768At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay1152x870At75Hz:
			err = TemplateGet1152x870At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		case kDisplay1280x960At75Hz:
			err = TemplateGet1280x960At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;

		case kDisplay1280x1024At75Hz:
			err = TemplateGet1280x1024At75Hz(&infoIndicies, bdiData, bddData, vwData, vwdData, info);
			break;
		
		default:
			err = kGDXErrDisplayModeIDUnsupported;
			goto ErrorExit;
	}
	err = kGDXErrNoError;										// Everything Okay
	
ErrorExit:

	return err;
}


//=====================================================================================================
//
// TemplateMapToIndices()
//	The TemplateGetxx routines return the correct programming and displayInfo status based on the
//	depthMode and the VRAM width and whether dbl buffering is enabled.  This routine maps the booleans into
//	the correct indicies
//	
//		-> depthMode		depthMode of interest
//		-> vramSize			how VRAM is in the system
//		-> vramWidth32		true if vram set to be 32 bits wide, false if vram set to be 64 bits wide
//		<- infoIndicies		the proper index for each type of data
//			the relevant fields of the 'InfoIndicies' structure are as follows:
//
//		<- dIndex			bit depth dependent data...only varies with bit depth
//		<- vwIndex			vram width dependent data...only varies with vram width
//		<- vwdIndex			vram width AND bit depth dependent data...varies with vram width & bit depth
//		<- maxDepthIndex	max DepthMode based on 1) vramSize 2) dbl buffering 3) vramWidth
//
//=====================================================================================================
static GDXErr TemplateMapToIndices(DepthMode depthMode, VRAMSize vramSize, Boolean vramWidth32,
	InfoIndicies *infoIndicies)
{	
	UInt32 vwbdd32Index;
	UInt32 vwbdd64Index;
	GDXErr err = kGDXErrUnknownError;						// Assume failure
	
	// First, some error checking.  It is not valid to:
	// 1) have 1 meg of VRAM and have a vramwidth of 64 bits
	
	if ( (k1MegVRAM == vramSize) && !vramWidth32 )
	{
		err = kGDXErrInvalidParameters;
		goto ErrorExit;
	}
	
	if (kDepthMode1 == depthMode)
	{
		infoIndicies->dIndex = kDepthMode1Index;			// 8 bpp
		vwbdd32Index = kvw32d8Index;
		vwbdd64Index = kvw64d8Index;
	} 
	else if (kDepthMode2 == depthMode)
	{
		infoIndicies->dIndex = kDepthMode2Index;			// 16 bpp
		vwbdd32Index = kvw32d16Index;
		vwbdd64Index = kvw64d16Index;
	} 
	else if (kDepthMode3 == depthMode)
	{
		infoIndicies->dIndex = kDepthMode3Index;			// 32 bpp
		vwbdd32Index = kvw32d32Index;
		vwbdd64Index = kvw64d32Index;
	} 
	else
	{
		err = kGDXErrDepthModeUnsupported;
		goto ErrorExit;
	}
	
	if (vramWidth32)
	{
		infoIndicies->vwIndex = vw32Index;
		infoIndicies->vwdIndex= vwbdd32Index;
	} 
	else 
	{
		infoIndicies->vwIndex = vw64Index;
		infoIndicies->vwdIndex = vwbdd64Index;
	}

	if (k1MegVRAM == vramSize)
		infoIndicies->maxDepthIndex = k1MegVRAMIndex;
	else if (k2MegVRAM == vramSize)
		infoIndicies->maxDepthIndex = k2MegVRAMIndex;
	else
		infoIndicies->maxDepthIndex = k4MegVRAMIndex;

	
	err = kGDXErrNoError;						// Everything okay

ErrorExit:
	return err;
}	
	
//=====================================================================================================
//
// TemplateGetXXX()
//	All of the TemplateGetXXX parameters are described here.  For the given indicies, the proper
//	data is returned.
//
//		-> InfoIndicies		the proper index for each type of data
//			the relevant fields of the 'InfoIndicies' structure are as follows:
//
//		-> dIndex			index for bddData in dRay		( kDepthMode1Index...kDepthMode3Index )
//		-> vwIndex			index for vwData in vwRay		( vw32Index, vw32Index )
//		-> vwdIndex			index for vwdData in vwdRay		( kvw32d8Index...kvw64d32Index )
//		-> maxDepthIndex	index for maxDepthRay			( k1MegVRAMIndex...k4MegVRAMIndex )
//
//		<- bdiData		The hardware programming parameters that DO NOT vary with bit depth
//		<- bddData		The hardware programming parameters that DO vary with bit depth
//		<- vwData		The hardware programming parameters that DO vary VRAM width
//		<- vwdData		The hardware programming parameters that DO vary VRAM width AND bit depth
//		<- info			various tidbits of information:
//			the relevant fields of the 'DisplayInfo' structure are as follows:
//
//		<- width			horizontal pixels for the given displayModeID
//		<- height			vertical pixels for the given displayModeID
//		<- maxDepthMode		max depthMode based on inputs.  if inputs can't be supported,
//							maxDepthMode = kDepthModeInvalid
//		<- rowBytes			rowBytes reported in get video parameters
//=====================================================================================================


	
//=====================================================================================================
//
//  TemplateGet512x384At60HzNTSC()
//
//=====================================================================================================
static GDXErr TemplateGet512x384At60HzNTSC(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {512, 384, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{(512 + kRowBytesOffset)*2, 0},		// kDepthMode1Index
										{(512*2 + kRowBytesOffset)*2, 2},		// kDepthMode2Index
										{(512*4 + kRowBytesOffset)*2, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords/2;							// slam rowBytes into DisplayInfo
	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet512x384At60Hz()
//
//=====================================================================================================
static GDXErr TemplateGet512x384At60Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {512, 384, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode3, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{512 + kRowBytesOffset, 0},		// kDepthMode1Index
										{512*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{512*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};
		
	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet640x480At50HzPAL()
//
//=====================================================================================================
static GDXErr TemplateGet640x480At50HzPAL(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {640, 480, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{(640 + kRowBytesOffset)*2, 0},		// kDepthMode1Index
										{(640*2 + kRowBytesOffset)*2, 2},		// kDepthMode2Index
										{(640*4 + kRowBytesOffset)*2, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords/2;							// slam rowBytes into DisplayInfo

	
	err = kGDXErrNoError;						// Everything okay
	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet640x480At60HzNTSC()
//
//=====================================================================================================
static GDXErr TemplateGet640x480At60HzNTSC(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {640, 480, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{(640 + kRowBytesOffset)*2, 0},		// kDepthMode1Index
										{(640*2 + kRowBytesOffset)*2, 2},		// kDepthMode2Index
										{(640*4 + kRowBytesOffset)*2, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords/2;							// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet640x480At60HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet640x480At60HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {640, 480, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{640 + kRowBytesOffset, 0},		// kDepthMode1Index
										{640*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{640*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet640x480At67Hz()
//
//=====================================================================================================
static GDXErr TemplateGet640x480At67Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {640, 480, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{640 + kRowBytesOffset, 0},		// kDepthMode1Index
										{640*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{640*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet640x870At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet640x870At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {640, 870, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x0b10, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{640 + kRowBytesOffset, 0},		// kDepthMode1Index
										{640*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{640*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet768x576At50HzPAL()
//
//=====================================================================================================
static GDXErr TemplateGet768x576At50HzPAL(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {768, 576, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{(768 + kRowBytesOffset)*2, 0},		// kDepthMode1Index
										{(768*2 + kRowBytesOffset)*2, 2},		// kDepthMode2Index
										{(768*4 + kRowBytesOffset)*2, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords/2;							// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet800x600At60HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet800x600At60HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {800, 600, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{800 + kRowBytesOffset, 0},		// kDepthMode1Index
										{800*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{800*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet800x600At72HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet800x600At72HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {800, 600, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{800 + kRowBytesOffset, 0},		// kDepthMode1Index
										{800*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{800*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet800x600At75HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet800x600At75HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {800, 600, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x1010, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{800 + kRowBytesOffset, 0},		// kDepthMode1Index
										{800*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{800*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

		
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet832x624At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet832x624At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {832, 624, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode2, kDepthMode3, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x0070, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{832 + kRowBytesOffset, 0},		// kDepthMode1Index
										{832*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{832*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	
	// if kDepthMode2 and there is only 1 meg of VRAM, the rowBytes is non standard.  Rowbytes has to
	// be 832*2 + 16 instead of 832*2 + kRowBytesOffset.  There is not enough vram otherwise.
	if ( ( kDepthMode2Index == infoIndicies->dIndex ) && ( kvw32d16Index == infoIndicies->vwdIndex ) )
		bddData->rowWords = 832*2 + 16;

	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1024x768At60HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet1024x768At60HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1024, 768, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x10b0, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1024 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1024*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1024*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1024x768At72HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet1024x768At72HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1024, 768, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x10b0, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1024 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1024*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1024*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1024x768At75HzVGA()
//
//=====================================================================================================
static GDXErr TemplateGet1024x768At75HzVGA(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1024, 768, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x10b0, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1024 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1024*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1024*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1024x768At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet1024x768At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1024, 768, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode3};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x10b0, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};

	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1024 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1024*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1024*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1152x870At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet1152x870At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1152, 870, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthMode1, kDepthMode2, kDepthMode2};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x10b0, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x00									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1152 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1152*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1152*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}


//=====================================================================================================
//
//  TemplateGet1280x960At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet1280x960At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1280, 960, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthModeInvalid, kDepthMode1, kDepthMode2};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x5c00, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1280 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1280*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1280*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
//  TemplateGet1280x1024At75Hz()
//
//=====================================================================================================
static GDXErr TemplateGet1280x1024At75Hz(const InfoIndicies *infoIndicies, BitDepthIndependentData *bdiData,
	BitDepthDependentData *bddData, VRAMWidthDependentData *vwData, WidthAndDepthDependentData *vwdData, 
	DisplayInfo *info)
{
	GDXErr err = kGDXErrNoError;				// infoIndicies are always correct
	DisplayInfo displayInfo = {1280, 1024, 0, 0};
	DepthMode maxDepthRay[3] = {kDepthModeInvalid, kDepthMode1, kDepthMode2};
	BitDepthIndependentData bdi = 
	{	// HSERR	HLFLN 	HEQ		HSP		HBWAY	HBRST	HBP 	HAL		HFP		HPIX
			123,	123, 	123, 	123,	123, 	123, 	123,	123,	123,	123, 
		
		// VHLINE	VSYNC 	VBPEQ	VBP		VAL		VFP		VFPEQ
			123,	123, 	123, 	123,	123, 	123, 	123,      

		0x5c00, 								// baseAddress points to start HW cursor
		0x77, 0x77, 0x77,						// Frank (PLL) M, PN, RI
		0x0c									// V sync and H Sync polarity
	};
	
	BitDepthDependentData dRay[3] = // RowWords,sysConfig
										{{1280 + kRowBytesOffset, 0},		// kDepthMode1Index
										{1280*2 + kRowBytesOffset, 2},		// kDepthMode2Index
										{1280*4 + kRowBytesOffset, 4}};		// kDepthMode3Index

	VRAMWidthDependentData vwRay[2] = {{0x0023}, {0x2219}};		// fbConfiguration1	vw32Index, vw64Index

	WidthAndDepthDependentData vwdRay[6] = 	//	kvw32d8Index		kvw32d16Index		kvw32d32Index
												{{11, 4425, 5},		{12,4332, 4}, 		{13, 7743, 7743},	
	// clockConfig, tAdj, dacCntl				kvw64d8Index		kvw64d16Index		kvw64d32Index	
												{3, 3432, 23},		{43, 3432, 0x15},	{32, 99, 23}};

	*bdiData = bdi;													// copy BitDepthIndependentData
	*bddData = dRay[infoIndicies->dIndex];							// copy BitDepthDependentData
	*vwData = vwRay[infoIndicies->vwIndex];							// copy VRAMWidthDependentData
	*vwdData = vwdRay[infoIndicies->vwdIndex];						// copy WidthAndDepthDependentData
	*info = displayInfo;											// copy DisplayInfo (width and height)
	info->maxDepthMode = maxDepthRay[infoIndicies->maxDepthIndex];	// slam maxDepthMode into DisplayInfo
	info->rowBytes = bddData->rowWords;								// slam rowBytes into DisplayInfo

	
ErrorExit:
	return err;
}



//=====================================================================================================
//
// TemplateResetSenseLines()
//	Before reading the Sense Lines to determine what monitor is connected, the sense lines
//	need to be reset.  Puts sense lines in a known state so the sensing process can start.
//
//=====================================================================================================
static void TemplateResetSenseLines(void)
{
	TemplateHALData *templateHALData = GraphicsHALGetHALData();

	ToynbeeRegisters *toynbee  = &templateHALData->toynbee;
	*toynbee->senseLineEnable = 0x07;						//  Tristate sense lines
	
	SynchronizeIO();
	MoreRealDelayFor(5*durationMillisecond);		// wait some amount of time for lines to stabilize
}
	


//=====================================================================================================
//
// TemplateDriveSenseLines()
//	Drive the approprite sense line so that the two other sense lines can be read.
//
//		-> senseLine		the sense line to drive.
//
//=====================================================================================================
static void TemplateDriveSenseLines(SenseLine senseLine)
{
enum
{
	kDriveAValue = 0x00000003,
	kDriveBValue = 0x00000005,
	kDriveCValue = 0x00000006
};

	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	ToynbeeRegisters *toynbee  = &templateHALData->toynbee;

	UInt32 senseLineValue;

	
	switch (senseLine)
	{
		case kSenseLineA:
			senseLineValue = kDriveAValue;
			break;
		case kSenseLineB:
			senseLineValue = kDriveBValue;
			break;
		case kSenseLineC:
			senseLineValue = kDriveCValue;
			break;
	}
	
	*toynbee->senseLineEnable = senseLineValue;				//  Drive appropriate sense line

	SynchronizeIO();
	
	MoreRealDelayFor(5*durationMillisecond);		// wait some amount of time for lines to stabilize

}



//=====================================================================================================
//
// TemplateReadSenseLines()
//	Drive the approprite sense line so that the two other sense lines can be read.
//
//		<- senseLineValue		read value of the sense lines
//
//=====================================================================================================
static UInt32 TemplateReadSenseLines(void)
{

enum { kTemplateSenseLineMask = 0x00000007 };

	TemplateHALData *templateHALData = GraphicsHALGetHALData();
	ToynbeeRegisters *toynbee  = &templateHALData->toynbee;

	UInt32 senseLineValue;
		
	senseLineValue = *toynbee->senseLineEnable;			//  read inverted sense line

	senseLineValue = ~senseLineValue;					// inverse the sense lines
	senseLineValue &= kTemplateSenseLineMask;			// Mask off the sense pins

	
	// Reset the lines:
	
	TemplateResetSenseLines();								// reset the sense lines	
	return(senseLineValue);
}



//=====================================================================================================
//
// TemplateGetExtendedSenseCode()
//	This routine applies the 'standard' extended sense code algorithm to the sense lines to determine
//	their ExtendedSenseCode.
//
// 	For 'standard' sense line hardware, the extended sense code algorithm is as follows:
//	(Note:  as described here, sense line 'A' corresponds to '2', 'B' to '1', and 'C' to '0')
//		 Drive sense line 'A' low and read the values of 'B' and 'C'.  
//		 Drive sense line 'B' low and read the values of 'A' and 'C'.
//		 Drive sense line 'C' low and read the values of 'A' and 'B'.
//
//	In this way, a six-bit number of the form BC/AC/AB is generated. 
//
//=====================================================================================================
static ExtendedSenseCode TemplateGetExtendedSenseCode()
{
	ExtendedSenseCode extendedBC;								// Result from driving line A
	ExtendedSenseCode extendedAC;								// Result from driving line B
	ExtendedSenseCode extendedAB;								// Result from driving line C
	
	ExtendedSenseCode extendedSenseCode;						// Final value.
	
	// Obtain the extendedBC value by driving sense line A
	TemplateDriveSenseLines(kSenseLineA);
	extendedBC = TemplateReadSenseLines();						// xx xx BC
	extendedBC = (extendedBC << 4) & 0x30;						// BC 00 00
	
	// Obtain the extendedAC value by driving sense line B
	TemplateDriveSenseLines(kSenseLineB);
	extendedAC = TemplateReadSenseLines();						// xx xA xC
	extendedAC = ((extendedAC >> 2) << 3) |						// (00 A0 00) | (00 0C 00) = 00 AC 00
				 ((extendedAC & 0x1) << 2);
				 
	// Obtain the extendedAB value by driving sense line C
	TemplateDriveSenseLines(kSenseLineC);
	extendedAB = TemplateReadSenseLines();						// xx xA Bx
	extendedAB = (extendedAB >> 1) & 0x3;						// 00 00 AB
	
	extendedSenseCode = extendedBC | extendedAC | extendedAB;
	
	return(extendedSenseCode);
}



//=====================================================================================================
// MoreRealDelayFor()
//	The DSL routine DelayFor cannot be called at Primary Interrupt Level or Secondary Task Level.
//	This provides equivalent functionality by calling UpTime to get the current AbsoluteTime
//	and then hanging around until the elapsed time period is over
//
//		-> duration		the number of microseconds/milliseconds for the delay
//						if the number is negative, it is the # of microseconds to delay
//						if the number is positive, it is the # of milliseconds to delay
//
//=====================================================================================================
static void MoreRealDelayFor(Duration duration)
{
	AbsoluteTime delayTime;						// the duration of delay in # ticks

	delayTime = DurationToAbsolute(duration);
	DelayForHardware(delayTime);
	
}
